; $Id: desmat_cubic_splines.pro,v 1.1 2008/03/07 05:27:27 do Exp $
;
; Copyright (c) 2002, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	desmat_cubic_splines.pro
; PURPOSE:
;	design matrix for smoothing cubic spline through data points
; CATEGORY:
;	Function.
; CALLING SEQUENCE:
;	design = desmat_cubic_splines( x, knots )
;
; INPUTS:
;	x	    x data. independent axis
;	knots	    array of knots, arbitrary spaced and increasing
;
;
; RETURN
;	design	    design matrix as double array
;
; RESTRICTIONS:
;	1. knots has to be of minimal length 2.
;	2. first one smaller than all x; last one larger than all x
;
; PROCEDURE:
;
;
;
;
;
; EXAMPLE:  x = findgen( 20 )
;	    y = exp( -0.3 * x ) + 0.02 * randomn( seed, 20 )
;	    knots = findgen( 5 ) * 4
;	    dm = desmat_cubic_splines( x, knots )
;	    yfit = fit_model( x, y, design=dm )
;	    fits a 5 knots cubic spline through these points
; REFERENCE:
;
; MODIFICATION HISTORY:
;	Nov. 2005		Written by Do Kester
;-

FUNCTION desmat_cubic_splines, time, knots

nk = n_elements( knots )
if ( nk EQ 1 ) then begin
    npt = n_elements( time )
    return, dblarr( 1, npt ) + 1
;    dm = dblarr( 2, npt )
;    dm[0,*] = 1
;    dm[1,*] = time
;    return, dm
endif

if ( ispresent( time LT knots[0] ) OR $
     ispresent( time GT knots[nk-1] ) ) then begin
    print, "Some points outside knots range"
    return, 0
endif

if ( where( knots LT shift( knots, 1 ) ) NE 0 ) then begin
    print, "Knots are not in increasing order"
    return, 0
endif

np = nk + 2
npt = n_elements( time )

dm = dblarr( np, npt )
x = dblarr( npt ) + 1
for i = 0, 2 do begin
    dm[i,*] = x
    x *= time
endfor
dm[3,*] = x
for i = 3, np - 2 do begin
    b = where( time GE knots[i-2], nb )
    if ( nb GT 0 ) then begin
	dm[i,b] = x[b]
	x[b] = ( time[b] - knots[i-2] ) ^ 3
	dm[i,b] -= x[b]
    endif
    dm[i+1,b] = x[b]
endfor

return, dm
end

