; $Id: fit_model.pro,v 1.1 2008/03/07 05:27:27 do Exp $
;
; Copyright (c) 2002, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	fit_model.pro
; PURPOSE:
;	fit a model to (x,y,[w]) data
; CATEGORY:
;	Function.
; CALLING SEQUENCE:
;	yfit = fit_model( x, y, weight=weight, design=design, chisq=chisq, $
;		param=param, stdev=stdev, scale=scale, $
;		prior=prior, logpr=logpr, robust=robust, status=status )
;
; INPUTS:
;	x	    x data. independent axis
;	y	    y data. dependent axis
;	weight	I   weights of the datapoints
;		    default : all 1.
;		O   weights as derived from robust fitting
;	design	I   design matrix
;	prior	    prior for evidence calculation
;		    default : 10000.0
;	robust	    number of iteration to robustify the solution
;		    default 0;
; OUT:
;	chisq	    chi-squared of the fit
;	param	    parameters of the fit
;	stdev	    standard deviations on the parameters
;	scale	    noise scale
;	logpr	    log of the evidence
;	status	    0-5 : error code from bayesfit
;		    6	: wrongly sized design matrix
;
; RETURN
;	yfit	    the (smooth) spline fit
;
; PROCEDURE:
;
;
;
;
;
; EXAMPLE:  x = findgen( 20 )
;	    y = exp( -0.3 * x ) + 0.02 * randomn( seed, 20 )
;	    yfit = fit_model( x, y, 5 )
;
; REFERENCE:
;
; MODIFICATION HISTORY:
;	Jul. 2007		Written by Do Kester
;-

FUNCTION fit_model, x0, y0, weight=weight, design=desmat, chisq=chisq, $
		    param=param, stdev=stdev, scale=scale, logpr=logpr, $
		    prior=prior, robust=robust, status=status

FORWARD_FUNCTION bayesfit

np = n_params()
if ( np EQ 0 OR np GT 2 ) then begin
    print, "Need at least one parameter and not more than 2"
    return, 0
endif else if ( np EQ 1 ) then begin
    y = x0
    x = dindgen( n_elements( x0 ) )
endif else begin
    x = x0
    y = y0
endelse

npt = n_elements( y )

dsz = size( desmat )
if ( dsz[0] NE 2 OR dsz[2] NE npt ) then begin
    print, "Design matrix of wrong size"
    status = 6
    return, 0
endif

np = dsz[1]
wgt = ( n_elements( weight ) NE npt ) ? dblarr( npt ) + 1 : weight
if ( n_elements( robust ) EQ 0 ) then robust = 0
if ( n_elements( prior ) EQ 0 ) then prior = 1.0

logpr = - !values.f_infinity
wt = wgt
result = dblarr( npt )
for k = 0, robust do begin

    dw = desmat * ( wt ## ( dblarr( np ) + 1 ) )	; weighted version
    fw = y * wt 					; weighted y

    mat = dw # transpose( dw )				; normal matrix
    vec = dw # fw					; vector

    if ( max( abs( mat ) ) EQ 0 ) then begin
	return, make_array( size=size( y ), value = 0 )
    endif

    stdev = dblarr( np + 1 ) + prior		    ; priors
    chisq = total( double( fw^2 ) )		    ; sum over (wgt*y)^2

    b = where( wt GT 0, nb )
    if ( nb EQ 0 ) then return, 1
    slsigma = -total( alog10( wt[b] ) )

    scale = 0					    ; automatic scaling

    status = bayesfit( mat, vec, slsigma, npt, chisq=chisq, sol=param, $
		stdev=stdev, scale=scale, logpr=logp )
    if ( status NE 0 ) then begin
	logpr = - !values.f_infinity
	break
    endif
    if ( k EQ 0 ) then logpr = logp	    ; I have to think more about this

    result = param # desmat

    if ( k GE robust ) then break

; recalculate the weigths based on the misfit (Tukey weighting)
    dif = y - result
    tukey = 1.0 / ( scale * 6 )
    wt = ( ( abs( dif * tukey ) < 1 ) ^ 2 - 1 ) ^ 2
    wt *= wgt

endfor

; return the weights
if ( robust GT 0 ) then weight = wt

RETURN, result
END
