;#> frin_search.dc3
; Identifier	frin_search
;
; Purpose	search fringes
;
; Synopsis	result = frin_cs_search( wn, xx, wt, freq, nfringes, tol=tol, $
;				      plot=plot, auto=auto, frid=frid, $
;				      logpr = logpr, status=status )
;
; Arguments	Name	I/O    Type:	Description
;		------------------------------------------------------------
;		wn	 I     fltarr	wave numbers
;		xx	 I     fltarr	fringe fluxes
;		wt	 I     fltarr	weights
;		freq	 I     fltarr	fringe cycles
;		nfringes I     int	max number of fringes to extract
;		tol	 I     float	tolerance, stopping criterion
;					default = 0 (use /auto or nfringes)
;		plot	 I     struct	for plotting
;		auto	 I     keyword	use automatic (=Baysian) stopping
;		frid	 I     struct	fringe-id
;		logpr	 0     float	evidence == log( probability )
;					(undefined when not auto)
;		status	 O     int	exit status
;					 0: succesful exit
;					 1: error
;
; Returns     fringelist O     struct
;		{ bandet       int	band or detector number
;		  reset        int	reset number
;		  frin_nr      int	fringe number
;		  cycle        float	fringe cycle
;		  cosamp       float	amplitude of cosine component
;		  sinamp       float	amplitude of sine component
;		  chisq        float	chisq
;		  chi_red }    float	accumulative chisq reduction
;
; Description
;
;
;
;
; Comment
;
;
; Example	--
;
; Category	UTIL ( module: fringes )
;
; Filename	frin_search.pro
;
; Author	Do Kester
;
; Version	1.7
;
; History	1.0  07-03-2001 DK, first implementation of frin_search.pro
;		1.1  05-04-2001 DK, remove compile_opt
;		1.2  16-05-2001 DK, repaired IDL5.2 bug
;		1.3  11-06-2001 DK, ... and another one
;		1.4  22-10-2001 FL, check exit status of frin_matrix
;				    define logprev in non-auto mode
;		1.5  24-10-2001 FL, more catches for memory problems
;		1.6  19-11-2001 DK, return (optionally) logpr.
;		1.7  03-07-2003 FL, check status of frin_model and bayesfit
;#<
;===========================================================================

function frin_cs_search, wn, xx0, wt, freq, nfringes, nrknots=nrknots, $
			 start=start, tol=tol, logpr=logpr, plot=plot, auto=auto, $
			 type=type, frid=frid, unit=unit, status=status

compile_opt idl2, hidden

forward_function bayesfit, append, frin_model, frin_matrix, frin_amplitude

common frin_search, ktry
common fringes, dmarr, dmpar, dmpnt

b = where( finite( wt ) EQ 0, nb )
if ( nb GT 0 ) then print, "NaNs in wt", nb

do_plot = ( size( plot, /type ) EQ 8 ) ? plot.plot : 0

xx = xx0
; frequencies to be tested
nfreq = n_elements( freq )

amprange  = 0.001			    ; prior for amplitudes
freqrange = freq[nfreq-1] - freq[0]	    ; prior for frequency
sigrange  = alog( 10.0 )		    ; prior for stdev

npt = n_elements( wn )			    ; length of the input vector
ones = fltarr( npt ) + 1.0
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; We need the evidence for a null model to compare real fringes with.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;print, "Null model"
if ( auto ) then begin
    b = where( wt GT 0, nb )
    if ( nb EQ 0 ) then return, 1
    slsigma = -total( alog10( wt[b] ) )
    chi = total( xx * wt * wt * xx )
    scale = 0.0
    stdev = alog( 10.0 )
    bferr = bayesfit( slsigma, npt, chisq=chi, logpr=logpr, $
		    scale=scale, stdev=stdev )
    if bferr gt 0 then return,1
    logpr0 = logpr
;   print, logpr, scale, chi, slsigma, npt
endif else logpr = 0.

lognxt = 0
wn0 = min( wn, max=wn1 )
eps = 0.001 * ( wn1 - wn0 )
npar = ( nrknots LE 1 ) ? 2 : 2 * ( nrknots + 2 )

bd = frid.det
kk = 0
klst = nfreq - 1
reduc = 2 * abs( tol ) > 1			; to get it started
bferr = 0
noerase=0
if ( plot.plot GE 2 ) then winset, 2
while( kk LT nfringes ) do begin

    knots = ( kk GT 0 OR nrknots LE 1 ) ? 1 : $
	      wn0 - eps + findgen( nrknots ) * ( wn1 - wn0 + 2 * eps ) / $
					    ( nrknots - 1 )

    at = -1
    chi_min = !values.d_infinity
    chim = fltarr( nfreq )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Search for minimum within the given set of frequencies
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    kpls = ktry
    kmin = ktry
    allk = ktry EQ 0
    for k = 0, nfreq - 1 do begin

       if ( npar EQ dmpar AND npt EQ dmpnt ) then $
	    desmat = reform( dmarr[*,*,ktry] ) $
       else desmat = desmat_cs_sines( wn, knots, freq[ktry], dmcs=dm )

       fit = fit_model( wn, xx, weight=wt, design=desmat, chisq=chisq, $
			 param=par, scale=scale )
       chim[ktry] = chisq

;print, "Fit Model ", k, ktry, kmin, kpls, chisq, at, chi_min, npt

       if ( chisq LT chi_min ) then begin
	  at = ktry
	  best = par
	  chi_min = chisq

; when success continue in the same direction otherwise reverse directions
	  ktry = ( kmin EQ 0 ) ? ++kpls : ( kpls EQ klst ) ? --kmin : $
		 ( ktry EQ kpls ) ? ++kpls : --kmin
       endif else begin
	  ktry = ( kmin EQ 0 ) ? ++kpls : ( kpls EQ klst ) ? --kmin : $
		 ( ktry EQ kpls ) ?  --kmin : ++kpls
       endelse

       if ( NOT allk AND kmin LT at - 2 AND kpls GT at + 2 ) then break

    endfor
    ktry = at				  ; for the next call start at best

;    print, "search ",n_elements(par), npt, freq[at], at, scale, chi_min, $
;	    format='(a,2i5,f10.4,i5,2g12.4,f10.2,i4,f10.2)'
    if ( at EQ -1 ) then print, chim

    optifreq = freq[at]

    if ( nfreq GT 1 ) then begin
      df = freq[1] - freq[0]

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Expand search range when miminum is at the edge
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      if ( at EQ 0 OR at EQ nfreq-1 ) then begin
	 fr = freq[at]
	 repeat begin
	    fr = fr + 0.2 * ( ( at EQ 0 ) ? -df : +df )
	    desmat = desmat_cs_sines( wn, knots, fr, dmcs=dm )
;print, "Ext model  ", fr
	    fit = fit_model( wn, xx, weight=wt, design=desmat, chisq=chisq, $
			     param=par, scale=scale )
	    if ( chisq LT chi_min ) then begin
	       at = k
	       best = par
	       chi_min = chisq
	       optifreq = fr
	    endif
	 endrep until ( chisq GT chi_min )
	 print, npt, optifreq, at, scale, chi_min

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Apply parabolic fit to find optimal freq at chisq minimum
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      endif else begin
	 y1 = chim[at-1]
	 y3 = chim[at+1]
	 dd = y1 + y3 - 2 * chi_min
	 if ( dd NE 0 ) then $
	    optifreq = optifreq + 0.5 * df * ( y1 - y3 ) / dd
      endelse
    endif

    desmat = desmat_cs_sines( wn, knots, optifreq, dmcs=dm )
    npar = ( size( desmat ) )[1]
    pri = fltarr( npar ) + amprange
    prp = [pri, sigrange]
;print, "BestModel  ", optifreq
    fit = fit_model( wn, xx, weight=wt, design=desmat, chisq=chisq, $
		param=best, scale=scale, prior=prp, logpr=lpo, status=status )


;    print, "best   ", npar, npt, optifreq, nrknots, scale, chisq, lpo, status, $
;	    format='(a,2i5,f10.4,i5,2g12.4,f10.2,i4)'

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Investigate what the minimum in nrknots is
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    nrk = ( kk EQ 0 ) ? nrknots : 1
nrk = 1
    while ( nrk GT 1 ) do begin
       nrk--

       if ( nrk EQ 1 ) then knots = 1 $
       else knots = wn0 - eps + findgen( nrk ) * ( wn1 - wn0 + 2 * eps ) / $
						( nrk - 1 )

       desmat = desmat_cs_sines( wn, knots, optifreq, dmcs=dm )
       npar = ( size( desmat ) )[1]
       prp = [fltarr( npar ) + amprange, sigrange]
;print, "KnotModel  ", nrk
       fit = fit_model( wn, xx, weight=wt, design=desmat, chisq=chisq, $
	      param=newp, scale=scale, prior=prp, logpr=lpn, status=status )

;	print, "knots  ", npar, npt, optifreq, nrk, scale, chisq, lpn, status, $
;	    format='(a, 2i5,f10.4,i5,2g12.4,f10.2,i4)'

       if ( lpn LT lpo ) then begin
	  nrk++
	  knots = wn0 - eps + findgen( nrk ) * ( wn1 - wn0 + 2 * eps ) / $
							    ( nrk - 1 )
	  desmat = desmat_cs_sines( wn, knots, optifreq, dmcs=dm )
	  npar = ( size( desmat ) )[1]

	  break
       endif
       lpo = lpn
       best = newp
    endwhile
nrk = nrknots
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Calculate the evidence for the complete model, including previous fringes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    design = append( design, desmat )
    df = total( dm, 1 )
    pw = 2 * !DPI * wn
    df *= pw * ( cos( pw * optifreq ) - sin( pw * optifreq ) )
    design = append( design, reform( df, 1, npt ) )
    pars = append( append( pars, [best] ), [optifreq] )
    prior = append( prior, [ fltarr( npar ) + amprange, freqrange ] )
    prip = [ prior, sigrange ]			  ; noise scale is last

    yfit = fit_model( wn, xx0, weight=wt, design=design, chisq=chisq, $
		      stdev=stdev, param=par, scale=scale, logpr=lognxt, $
		      prior=prip, status=status )

    if ( plot.verbose GE 2 ) then print, "fit    ", (size(design))[1], npt, $
	    optifreq, nrk, scale, chisq, lognxt, status, logpr, $
	    format='(a,2i5,f10.4,i5,2g12.4,f10.2,i4,f10.2)'

;    print, pars, format='(10g12.4)'
;    print, par, format='(10g12.4)'
;    print, stdev, format='(10g12.4)'


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Do some plotting
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    if ( plot.plot GE 2 ) then begin
      yr = [0,max(chim) > 1e-6]
      rate = ( type EQ 'FIS_SW' ) ? 24 : 16
      plot, freq, chim, pos=[0.1,0.7,0.95,0.9], ytitle="chisq", yr=yr, $
	    xr=[0,rate/2], xtickname=replicate(' ',10), noerase=noerase, /nodata

	  xyouts, 5.5, 0.9*yr[1], string( bd, ":", frid.reset, format="(i3,a,i2)" )
      axis, xaxis=1, xtitle="frequency"

      fftxx = xx0 * ( wt GT 0 )
      plot_fft, fftxx, 1.0/rate, /oplot, col=2
      fxfit = ( fftxx - yfit )
      plot_fft, fxfit, 1.0/rate, /oplot, col=3
      oplot, freq, chim

      if ( wn[npt-1] GT 10 ) then xr = [-5,5] + wn[npt-1]
      plot, wn, fftxx, pos=[0.1,0.4,0.95,0.7], /noerase, ytitle="fringes", $
	    xr=xr, xtickname=replicate(' ',10)
      oplot, wn, wt * !y.crange[1] * 0.9, color=3
      oplot, wn, yfit, color=2
      yr = !y.crange

      plot, wn, fxfit, pos=[0.1,0.1,0.95,0.4], /noerase, ytitle="residual", $
	    xtitle=string( "time + ", start ), xr=xr, yr=yr
;    oplot, wn, fit, color=2
;    noerase = 1
    endif
;print, size( fftxx )

    reduc = lognxt - logpr

    if ( reduc LE tol OR bferr GT 0 ) then begin
;	 kk = kk - 1
	goto, no_more_fringes
    endif else if ( auto ) then logpr = lognxt

;waitforcursor, cont

    logpr = lognxt

    hind = indgen( npar/2 )
    cosamp = best[hind] # dm
    sinamp = best[hind+npar/2] # dm

    ampli = sqrt( cosamp^2 + sinamp^2 ) 	; amplitude
;    phase = atan( sinamp, cosamp ) * 180 / !PI 	; phase

    str_logpr = ( auto ) ? string( logpr - logpr0, logpr, format='(2f10.2)' ) : ''
    if ( plot.verbose GT 0 ) then $
	printf, unit, bd, frid.reset,  kk+1, nrk, optifreq, $
		1000*mean( ampli ), $
		chi_min, str_logpr, format='(i4,i6,2i4,2f10.4,g10.4,a)'

    flist = { bandet	: bd, $
	      reset	: frid.reset, $
	      frin_nr	: kk+1, $
	      cycle	: double( optifreq ), $
;	      cosamp	: cosamp, $
;	      sinamp	: sinamp, $
	      cosamp	: ptr_new( cosamp ), $
	      sinamp	: ptr_new( sinamp ), $
	      chisq	: chi_min, $
	      chi_red	: chi_min }
    fringelist = append( fringelist, flist )


    kk = kk + 1
    xx = xx0 - yfit

endwhile
no_more_fringes :

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Print the last fringes, or none if such is the case
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
if ( kk EQ 0 ) then begin
  fringelist = 0
  str_logpr = ( auto ) ? ( ( lognxt NE 0 ) ? $
	 string( lognxt - logpr0, lognxt, format='(2f10.2)' ) : $
	 string( logpr0, format='(10x,f10.2)' ) ) : ''
  if ( plot.verbose GT 0 ) then $
      printf, unit, bd, frid.reset, '  --> No fringes found...', $
	 str_logpr, format='(i4,i6,a,29x,a)'

endif
;print, "fringelist ", fringelist
return, fringelist
end
