;#> frin_search.dc3
; Identifier	frin_search
;
; Purpose	search fringes
;
; Synopsis	result = frin_search( wn, xx, wt, freq, nfringes, tol=tol, $
;				      plot=plot, auto=auto, frid=frid, $
;				      logpr = logpr, status=status )
;
; Arguments	Name	I/O    Type:	Description
;		------------------------------------------------------------
;		wn	 I     fltarr	wave numbers
;		xx	 I     fltarr	fringe fluxes
;		wt	 I     fltarr	weights
;		freq	 I     fltarr	fringe cycles
;		nfringes I     int	max number of fringes to extract
;		tol	 I     float	tolerance, stopping criterion
;					default = 0 (use /auto or nfringes)
;		plot	 I     struct	for plotting
;		auto	 I     keyword	use automatic (=Baysian) stopping
;		frid	 I     struct	fringe-id
;		logpr	 0     float	evidence == log( probability )
;					(undefined when not auto)
;		status	 O     int	exit status
;					 0: succesful exit
;					 1: error
;
; Returns     fringelist O     struct
;		{ bandet       int	band or detector number
;		  reset        int	reset number
;		  frin_nr      int	fringe number
;		  cycle        float	fringe cycle
;		  cosamp       float	amplitude of cosine component
;		  sinamp       float	amplitude of sine component
;		  chisq        float	chisq
;		  chi_red }    float	accumulative chisq reduction
;
; Description
;
;
;
;
; Comment
;
;
; Example	--
;
; Category	UTIL ( module: fringes )
;
; Filename	frin_search.pro
;
; Author	Do Kester
;
; Version	1.7
;
; History	1.0  07-03-2001 DK, first implementation of frin_search.pro
;		1.1  05-04-2001 DK, remove compile_opt
;		1.2  16-05-2001 DK, repaired IDL5.2 bug
;		1.3  11-06-2001 DK, ... and another one
;		1.4  22-10-2001 FL, check exit status of frin_matrix
;				    define logprev in non-auto mode
;		1.5  24-10-2001 FL, more catches for memory problems
;		1.6  19-11-2001 DK, return (optionally) logpr.
;		1.7  03-07-2003 FL, check status of frin_model and bayesfit
;#<
;===========================================================================

function frin_search, wn, xx, wt, freq, nfringes, tol=tol, logpr=logpr, $
		      plot=plot, auto=auto, frid=frid, unit=unit, status=status

compile_opt idl2, hidden

forward_function bayesfit, append, frin_model, frin_matrix, frin_amplitude

do_plot = ( size( plot, /type ) EQ 8 ) ? plot.plot : 0

icym = intarr( nfringes )
chis = fltarr( nfringes )

; frequencies to be tested
ncycle = n_elements( freq )
cystep = freq[1] - freq[0]
nn = n_elements( wn )
ones = fltarr( nn ) + 1.0

design = frin_matrix( freq, wn, wt, status=status )
if ( status gt 0 ) then return,status

if ( auto ) then begin
  b = where( wt GT 0, nb )
  if ( nb EQ 0 ) then return, 1
  slsigma = -total( alog10( wt[b] ) )
  chi = total( xx * xx * wt )		    ; no model --> sum of data sq
  scale = 0.0
  ndata = nn
  stdev = alog( 10.0 )
  bferr = bayesfit( slsigma, ndata, chisq=chi, logpr=logpr, $
		    scale=scale, stdev=stdev )
  if bferr gt 0 then return,1
  logpr0 = logpr
; print, logpr, scale, chi, slsigma, nn
endif else logpr = 0.

logprnext = 0
wr = cliprange( wn )
distance = 0.4 / ( wr[1] - wr[0] )
user = { wn : wn, $
	 xx : xx, $
	 wt : wt, $
	 nn : nn, $
	 dis : distance, $
	 fringe : fltarr( nn ), $
	 ac : fltarr( nfringes ), $
	 as : fltarr( nfringes ), $
	 mat : ptr_new( 1 ) }

bd = ( frid.band GT 0 ) ? frid.band : frid.det
kk = 0
fx = fltarr( nn )
reduc = 2 * abs( tol ) > 1			; to get it started
bferr = 0
while( kk LT nfringes ) do begin
  xw = xx * wt
  x2 = total( xx * xw )
  ncyc = ncycle

  ch = frin_amplitude( design, xw, x2, ampcos=ac, ampsin=as, status=status )
  if ( status gt 0 ) then return,status
  minc = min( ch, kmin, max = maxc )		; kmin is best chi-sq
  optifreq = freq[kmin]

;  if ( kmin EQ 0 OR kmin EQ ncycle - 1 ) then begin
  if ( kmin EQ -1 OR kmin EQ ncycle ) then begin
; it is on the edge of the cycle range --> extend the range
    ncyc = long( distance * 2 )

    frep = optifreq + ( indgen( ncyc ) - ( ( kmin EQ 0 ) ? ncyc - 2 : 2 ) ) * $
		      cystep

    desp = frin_matrix( frep, wn, wt, status=status )
    if ( status gt 0 ) then return, status
    ch = frin_amplitude( desp, xw, x2, status=status )
    if ( status gt 0 ) then return, status

    minc = min( [ ch, maxc ], kmin, max = maxc )	   ; kmin is best chi-sq
;   print, optifreq, kmin, ncyc, frep[kmin]

    optifreq = frep[kmin]
  endif

  chis[kk] = ( maxc - minc ) / maxc	; reduction in chi-sq
  icym[kk] = kmin			; indices of minimums
  if ( kk EQ 0 ) then begin
    aac = ac					    ; for plotting at the end
    aas = as
    chisq = ch
    trc = chis[kk]
    tmc = maxc
  endif else trc = ( tmc - ch[kmin] ) / tmc


; find the optimum fringe cycle as a minimum in chi-sq (parabolic approx)
  if ( kmin GT 0 AND kmin LT ncyc - 1 ) then begin
    y1 = ch[kmin-1]
    y3 = ch[kmin+1]
    if ( y1 + y3 - 2 * minc NE 0 ) then $
	optifreq = optifreq + 0.5 * cystep * ( y1 - y3 ) / $
					     ( y1 + y3 - 2 * minc )
  endif else goto, no_more_fringes

  par = append( par, optifreq )

  chi = frin_model( user, par )
  if chi lt 0. then begin
    goto,no_more_fringes
  endif

  optifreq = par[kk]

  if ( auto ) then begin
    k2 = 3 * ( kk + 1 )
    stdev = [ fltarr( k2 ) + 0.001, alog( 10.0 ) ]
    stdev[indgen(kk+1) * 3] = ncycle * cystep
;print, stdev
    scale = 0.0
    ndata = nn
;   print, *(user.mat)

    logprnext = 0
    bferr = bayesfit( *(user.mat), slsigma, ndata, chisq=chi, logpr=logprnext, $
		      stdev=stdev, scale=scale )
    if bferr gt 0 then begin
	print, "Bayesfit error : ", bferr
	goto, no_more_fringes
    endif
    ptr_free, user.mat			; not needed any more

;    print, logprnext, scale, chi, slsigma, ndata, minc
;   print, par, scale
;    print, stdev

    reduc = logprnext - logpr
  endif else reduc = chis[kk]

  if ( reduc LT tol OR bferr GT 0 ) then goto, no_more_fringes $
  else if ( auto ) then logpr = logprnext

  ammin = sqrt( user.ac[kk]^2 + user.as[kk]^2 ) 		; amplitude
  phmin = atan( user.as[kk], user.ac[kk] ) * 180 / !PI		; phase

  str_logpr = ( auto ) ? string( logpr - logpr0, logpr, format='(2f10.2)' ) : ''
  if ( NOT plot.sil ) then $
      printf, unit, bd, frid.reset,  kk+1, optifreq, ammin * 1000, $
	phmin, chis[kk], trc, str_logpr, format='(i4,i6,i4,5f10.4,a)'

  flist = { fringelist, $
	    bandet    : bd, $
	    reset     : frid.reset, $
	    frin_nr   : kk+1, $
	    cycle     : optifreq, $			; freq[kmin], $
	    cosamp    : user.ac[kk], $
	    sinamp    : user.as[kk], $
	    chisq     : ch[kmin], $
	    chi_red   : chis[kk] }
  fringelist = append( fringelist, flist )
  frincycle = par

  fx = user.fringe
  xx = user.xx - user.fringe


  fw5 = [par] # ( 2 * !PI * plot.wn5 )
  on5 = fltarr( plot.nn5 ) + 1.0
  plot.fr5 = total( ( [user.ac[0:kk]] # on5 ) * cos( fw5 ) + $
		    ( [user.as[0:kk]] # on5 ) * sin( fw5 ), 1 )

  if ( do_plot GE 4 OR do_plot EQ -4 ) then begin
; plot the fringe just found
    frin_plot_fringes, freq, ac, as, ch, minc, optifreq, kk, nfringes, frid

    frin_plot_total, wn, user.xx, wt, fringelist.cycle, freq, aac, aas, chisq, $
		   plot, frid
  endif

  kk = kk + 1

endwhile
no_more_fringes :

kk = kk - 1

if ( do_plot GE 2 OR do_plot EQ -2 ) then begin
if ( n_elements( aac ) EQ 0 ) then stop

  frin_plot_total, wn, user.xx, wt, frincycle, freq, aac, aas, chisq, $
		   plot, frid
endif

if ( kk GE 0 ) then begin
  if ( do_plot EQ 3 OR do_plot EQ -3 ) then $
	frin_plot, fringelist, freq, aac, aas, chisq, ch, frid
  fringelist.cosamp = user.ac[0:kk]
  fringelist.sinamp = user.as[0:kk]
endif else begin
  fringelist = 0
  str_logpr = ( auto ) ? ( ( logprnext NE 0 ) ? $
	 string( logprnext - logpr0, logprnext, format='(2f10.2)' ) : $
	 string( logpr0, format='(10x,f10.2)' ) ) : ''
  if ( NOT plot.sil ) then $
      printf, unit, bd, frid.reset, '  --> No fringes found...', $
	 str_logpr, format='(i4,i6,a,29x,a)'

endelse

return, fringelist
end
