;#> fringes.dc1
; Identifier   fringes
;
; Purpose      remove fringes in FIS data stream
;
; Synopsis     struct = fringes( struct, detlist, [smooth=smooth,
;		     /auto, nfringes=nfringes, tol=tol, $
;		     midcycle=midcylce, cycle=cycle, ncycle=ncycle, cystep=cystep, $
;		     /weight, fringelist=fringelist, $
;		     smowid=smowid, fill=fill, plot=plot, /noplot, $
;		     suppress=suppress, supwid=supwid, $
;		     evidence=evidence, status=status] )
;
; Arguments    Name	 I/O   Type:	 Description:
;	       -------------------------------------------------------------
;	       struct	  I    struct	 SPD, AAR or CAL25_xx
;	       detlist	  I    intarr	 list of detectors to be processed
;	       smooth	  I    fltarr	 smoothed background
;					 provided bij gb_set_glitch_status_gp
;	       auto	  I    keyword	 automatic determination of the number
;					 of fringes to extract
;					 default (unless TOL is set)
;	       nfringes   I    int	 number of fringes to be removed
;					 default = 50
;	       tol	  I    real	 reduce the chisq until reduction is
;					 less than tol ( 0.01 == 1 percent )
;					 default = 0
;	       midcycle   I    real	 cycle frequency used for smoothing
;					 [default: cycle+0.5*ncycle]
;	       cycle	  I    real	 starting cycle to search for fringes
;					 default band1 all    : 4000
;					 default band2 aper 2 : 4000
;					 default band2 aper 3 : 5500
;					 default band3 all    : 3000
;	       ncycle	  I    int	 number of cycles to check
;					 default band1 all    : 1000
;					 default band2 all    : 1000
;					 default band3 aper 1 : 3000
;					 default band3 other  :  600
;	       cystep	  I    real	 step in the cycles
;					 default = 1
;	       wrange	  I    fltarr[2] limit operations to wavelengths
;					 within wrange.
;					 default: defined by aperture & band
;	       weight	  I    keyword	 set all weights to 1
;					 default: less weights to outliers
;	       fmask	  I    long	 select only data with
;					 ( flag AND fmask ) EQ 0
;					 default = 16L OR 2L^30
;					   (== MASK and NODATA)
;	       plot	  I    int	 0 : plot the result ( default )
;					 -1: no plotting
;					 1-3: various plotting options
;	       noplot	  I    keyword	 no plotting (same as plot=-1)
;	       smowid	  I    int	 nr of fringes over which to smooth to
;					 get a background
;					 default : 1
;	       separate   I    keyword	 make a smooth background for each
;					 det individually
;	       fill	  I    float	 add fill times the smoothed envelope
;					 of the fringes to the data
;					 default : 0. range from -1 to +1
;	       suppress   I    fltarr	 set of wavelengths to disregard during
;					 fringe fitting (eg. at emission lines)
;	       supwid	  I    float	 width of the suppression
;					 default : 0.01
;	       fringelist O    struct	 cumulative list of extracted fringes
;		{ bandet	 int	 band or detector number
;		  updown	 string  up or down or both
;		  aotband	 string  aot band
;		  frin_nr	 int	 fringe number
;		  cycle 	 float	 fringe cycle
;		  cosamp	 float	 amplitude of cosine component
;		  sinamp	 float	 amplitude of sine component
;		  chisq 	 float	 chisq
;		  chi_red }	 float	 accumulative chisq reduction
;	       evidence   O	flt(arr) evidence == log( probability)
;					 (undefined if not auto)
;	       status	  O	int	 exit status
;					  0: succesfully processed
;					  1: error, input struct is returned
;
; Returns      struct	  O    struct	 SPD, AAR or CAL25 with the fringes
;					 in the requested bands/dets removed.
;					 All other info is passed undisturbed.
;
; Description  This procedure should be run AFTER respcal, otherwise you are
;	       dividing new fringes in with respcal. So eg. on an AAR.
;	       It actually works best on AARs as these structs are cleaner in
;	       contents.
;
;	       Fringes are detected in wavenumber (1/micron) space, where
;	       they have a constant `wavelength', in cycles per wavenumber
;	       (C/Wn [==micron]).
;
;	       To find fringes we fit a cosine function to the data
;		   f(n) = alfa.cos( p.n + phi ) == A.cos(p.n) + B.sin(p.n)
;	       A, B and p are parameters to be estimated. p the the 'wavelength'
;	       of the fringe. For band 3 the main fringe is located at
;	       3250 cycle/wavenumber. For a given p f(n) is a linear function
;	       which can easily be fitted to the data. So we search all
;	       wavelengths from CYCLE (=3000) to CYCLE+NCYCLE*CYSTEP (=3600),
;	       fit A and B and find the minimum in chi-sq.
;
;	       In automatic mode we calculate the (bayesian) evidence for
;	       this (new) fringe. We keep the fringe if this evidence is
;	       larger than the evidence carried by the previous (set of)
;	       fringe(s). The process of identifying new fringes continues
;	       as long as the evidence increases with each new fringe. For
;	       all fringes the amplitudes are deterimined and the total
;	       fringe-set is subtracted. This is the advised mode.
;
;	       In non-automatic mode fringes are extracted until the reduction
;	       in chi-sq is less than TOL or the number of fringes is larger
;	       than NFRINGES
;
;	       Before the fitting is started a smoothed version of the data
;	       is subtracted from itself, so that all the data are around zero.
;	       This smoothed version is at the end added in again.
;	       The smoothing is done with a cubic spline function with a
;	       width of about SMOWID cycles. The smoothing is done on a per
;	       detector basis and by default doing UPDOWN separately.
;	       The keyword /SEPARATE determines whether a smooth background
;	       is determined for each detector individually, or all detectors
;	       are taken together. This smoothing is somewhat of an art;
;	       different options have to be tried to get the best result.
;
;	       To decide which solution is better we can get the evidence
;	       from the last value in the column `LogPr'. The evidence of
;	       solution 1 w.r.t solution 2 equals 10 * ( LogPr1 - LogPr2 ) dB.
;	       When e.g. trial 1 had the keyword /UPDOWN set, it made 2
;	       calculations, one for UP and one for DOWN, with each a final
;	       LogPr. Lets say the numbers were 1040.9 (UP) and 1027.2 (DOWN).
;	       We can add these together to get the total 2068.1. The 2nd trial
;	       was exactly the same but without /UPDOWN. Now we get 1 final
;	       number for LogPr (2143.6). We see that the second trial is to be
;	       preferred over the first by more than 750 dB. An overwhelming
;	       case for 2.
;
;	       Large outliers in the background subtracted data are given less
;	       weight, unless /WEIGHT is set. When there are still points that
;	       should not be taken along in the calculations they can be
;	       given zero weight with keywords SUPPRESS and SPUWID.
;
;	       The process is best done per (AOT)band. It appears that the
;	       fringes of the different detectors in a band are so much
;	       similar that real gains are made by combining them.
;
;	       The keyword FILL is present to (partly) add or subtract the
;	       envelope of the fringe complex back into the data. In principle
;	       fringes remove light, more light or less light. We treat it
;	       here as if light is moved from the valleys of the fringes to
;	       the tops. However after application of the RSRF it is not clear
;	       whether the residual fringes are still only removing light.
;	       It could be both ways. Anyway wiggles have been reported in
;	       (definged) spectra which suspiciously look like the envelope
;	       of the fringe complex.
;
;	       Depending on the value of PLOT several sets of plots pop up.
;	       -1 or /NOPLOT: no plotting.
;	       0. The result of the fringe removal is plotted. Original data
;		  in white, smoothed background in red. The defringed data
;		  are displayed in green, offsetted somewhat for clarity. And
;		  in blue the (scaled) RSRF is plotted. In the lower panel
;		  the background-removed data are plotted in white (wgt=1),
;		  red (wgt=0) or yellow (0<wgt<1). Overplotted in blue are
;		  the fringe complex which is found. And in red (when FILL was
;		  present) the extra fill-factor.
;	       1. Three panels with the background subtracted data in white/
;		  yellow/red depending on its weights. The total of the
;		  fringe complex is overplotted in blue. All these are
;		  plotted against wavenumber.
;		  The top panel repeats the next plot, with all extracted
;		  fringes in green
;	       2. The parameters A (white) and B (yellow) as a function
;		  of p (cycles per wavenumber).
;		  Chi-sq (scaled and red) is also plotted as function of p. In
;		  green is the wavelength of the selected fringe.
;		  Also a preliminary version of 1 is popping up.
;	       3. Several plots of the background smoothing algoritm
;		  (for debugging only)
;
;	       See also the Examples.
;
; Comment      See Kester etal. 2001 "SWS Fringes and Models"
;		   in ISO Legacy Conference Proceedings.
;
;	       For 3 reasons this procedure is superior to the use of FFTs.
;	       1. The wavelength of the fringe can be estimated much better
;	       than in an FFT where it is limited to the Nyquist frequency,
;	       which in turn is determined by the number of datapoints.
;	       2. It works on unevenly spaced data just as well as on a
;	       regularly gridded data set. Even when there are large gaps
;	       in the data like eg. in an AOT2, it works fine.
;	       3. It automatically determines how many fringe components are
;	       present, avoiding possible overfitting and underfitting.
;
;	       Due to the fact that the RSRF was measured in the lab with an
;	       extended source while most of our objects are pointlike, the
;	       fringes in the RSRF are much less pronounced than in actual
;	       observations. Especially at full resolution, in AOT6 and
;	       AOT2 we see fringes in all bands while there are hardly any
;	       in eg. band 1 of the RSRF. Also amplitudes are larger. In AOT1s
;	       some of these fringes disappear in the lower resolution.
;	       Although I have seen `aliased' fringes (the beat between the
;	       fringe and the sampling) in single detector data of band 1d
;	       in an AOT1 speed 4. This aliased fringe could also be removed
;	       with this program by combining all detectors of the band.
;
;	       In CAL25_3d I saw after removal of fringes of the
;	       `fundamental' frequency of 3250 C/wn, residues of 2 times
;	       this i.e. 6500 C/wn, which also could succesfully be removed.
;
;	       In case of memory problems (possible in band 3a) it is advised
;	       to add the keyword cystep=2 (or 3).
;
; Example      To defringe an AAR in AOTband 3A only:
;		    aarf = fringes( aar, band='3a' )
;
;	       When the AAR is an AOT1 speed 4 or an AOT2 or 6, there is lots
;	       of data in each detector. It might be advantageous to do
;		    aarf = fringes( aar, band='3a', /separate )
;
;	       If the pointing was drifting only a little bit during the
;	       observation, resulting in a slight wavelength shift of much less
;	       than 1 resolution element, it might be better to do the up scan
;	       and the down scan separately:
;		    aarf = fringes( aar, band='3a', /updown )
;	       The fringe pattern is extremely sensitive to small changes in
;	       the wavelength, changes so small that they are well within the
;	       calibration of SWS.
;
;	       To defringe the same AAR in all of band 2 too.
;		    aarf = fringes( aarf, band=2 )
;	       (it will do each AOTBAND separately)
;
;	       To extract upto 4 fringe components from an SPD without plotting
;	       and doing all detectors of band 3 separately
;		    dlist = indgen( 12 ) + 25
;		    spdf = fringes( spd, dlist, nfringes=4, plot=-1 )
;
; Category     UTIL ( module: fringes )
;
; Filename     fringes.pro
;
; Author       Do Kester
;
; Version      2.12
;
; History      1.0  14-05-96 DK, first implementation
;	       1.1  21-05-96 DK, fmask introduced and silly error removed
;				 band option added.
;	       1.2  24 10 96 DK, it is now done for each aperture separately.
;	       1.3  11-03-97 DK, defaults for band 2, works on cal25_xn
;				 added keywords: TOL, WRANGE, APODIZE, WEIGHT,
;						 UPDOWN
;	       1.4  18-02-98 DK, better smoothing, increased nclycle for 3A
;				 (actually the smoothing done before left
;				  some fringes in the spectrum which could
;				  not be removed)
;	       1.5  19-02-98 DK, defringing by AOT-band i.e. there is an extra
;				 loop over the orders.
;				 added keyword: ORDERLIST
;	       1.6  24-03-99 FL, use AAR, added band 4A, 4C and 4D
;				 two small bug fixes
;	       2.0  15-03-01 DK, Serious rewrite adding bayesian fringe
;				 detection, improved background smoothing,
;				 more interesting plotting etc.
;				 keywords added: SMOWID, FRINGELIST, AUTO,
;						 CYSTEP, FILL, SUPPRESS, SUPWID
;				 keywords changed: UPDOWN, PLOT, TOL, BAND
;				 keywords removed: APODIZE, NOWINDOWS
;	       2.1  27-03-01 DK, make Tolerance option workable
;	       2.2  05-04-01 DK, remove compile_opt
;	       2.3  10-05-01 DK, various stability issues (thanks to FL)
;				 Added keyword SEPARATE, NOPLOT
;				 Expanded documentation on new issues of V2.n
;	       2.4  16-05-01 DK, mistake in logics
;	       2.5  02-08-01 FL, SPR_S0402 check NODATA and MASK for bad data
;	       2.6  21-06-01 FL, dont't plot when plot = -1 or /noplot
;				 use mask derived in frin_smooth
;	       2.7  27-09-01 DK  Data are correctly placed back into AAR/SPD
;				 ( was wrong since v2.0 )
;	       2.8  22-10-01 FL  Check exit status from frin_search
;	       2.9  24-10-01 FL  added comment on memory catch to the user
;	       2.10 19-11-01 DK  return (optionally) the evidence.
;	       2.11 19-02-03 FL  enable a non-SWS aar to be processed
;				 return fitted continuum
;				 pass suppressed lines to frin_smooth
;	       2.12 02-02-04 WF  quality output
;#<

function fringes, tsd, detlist, nfringes=nfringes, $
	tol=tol, smooth=smoflux, resetstep=resetstep, $
	detstep=detstep, range=range, unit=unit, $
	cycle=cyc, ncycle=ncy, cystep=cystep, silent=silent, $
	weight=weight, plot=plot, fringelist=fringelist, auto=auto, $
	evidence=evidence, status=status, _extra=_extra

compile_opt idl2

common fringes_save, splash

status = 0B

type = tsd->get_datatype()
if ( n_elements( detlist ) EQ 0 ) then $
    detlist = ( ( type EQ 'FIS_SW' ) ? indgen( 100 ) : indgen( 75 ) ) + 1
ndet = n_elements( detlist )
bnd = ( type EQ 'FIS_SW' ) ? 0 : 1
sbd = ( type EQ 'FIS_SW' ) ? 'FIS_SW' : 'FIS_LW'

if ( n_elements( plot ) EQ 0 ) then plot = 0
if ( n_elements( unit ) EQ 0 ) then unit = -1		; stdout

if ( n_elements( tol ) EQ 0 ) then tol = 0.0
auto = keyword_set( auto ) OR ( tol LE 0.0 AND n_elements( nfringes ) LE 0 )
if ( n_elements( nfringes ) LE 0 ) then nfringes = ( tol GT 0 OR auto ) ? 50 : 1
if ( n_elements( resetstep ) EQ 0 ) then resetstep = 1
if ( n_elements( detstep ) EQ 0 ) then detstep = 1

; set default for cycle and ncycle for bands 1, 2, 3, 4
bcycle	= [ 0.30, 0.50 ]
bncycle = [ 20, 20 ]
bsmooth = [ 1.0, 1.0, 1.0, 1.0, 1.0, 1.0 ]		;; TBD
if ( n_elements( cystep ) EQ 0 ) then cystep = 0.005

flux = tsd->get_value_tsd( 'fis_obs', 'flux' )
rstf = tsd->get_pix_flag_tsd( 'fis_obs', 'reset' )
smoo = smoflux
infl =	tsd->get_pix_flag_tsd( 'fis_obs', 'saturate' ) OR $
	tsd->get_pix_flag_tsd( 'fis_obs', 'no_diff' ) OR $
	tsd->get_pix_flag_tsd( 'fis_obs', 'gpgl_type1' ) OR $
	tsd->get_pix_flag_tsd( 'fis_obs', 'gpgl_type2' ) OR $
	tsd->get_pix_flag_tsd( 'fis_obs', 'gpgl_type3' ) OR $
	tsd->get_pix_flag_tsd( 'fis_obs', 'gpgl_tail' ) OR $
	rstf

if ( type EQ 'FIS_LW' ) then $
    calstat = tsd->get_status_tsd( 'fis_obs', 'calalon' )
if ( type EQ 'FIS_SW' ) then $
    calstat = tsd->get_status_tsd( 'fis_obs', 'calason' )
calstat = calstat OR shift( calstat, 2 )
infl = infl OR ( make_array( n_elements( infl[*,0] ), /byte, value=1 ) # $
		 calstat )

dead = where( tsd->get_pix_flag_tsd( 'fis_obs', 'dead' ) )
infl[dead] = rstf[dead]

if ( n_elements( range ) EQ 2 ) then begin
    b = lindgen( range[1] - range[0] ) + range[0]
    flux = flux[*,b]
    rstf = rstf[*,b]
    infl = infl[*,b]
    smoo = smoo[*,b]
endif

modl = make_array( size=size( flux ) )

for kk = 0, ndet-detstep, detstep do begin
    det = detlist[kk]
    kd	= det - 1

    nq = uniqer( rstf[kd,*], start=qs, length=ql )
    bq = where( rstf[kd,qs] EQ 0, nreset )
    treset = fltarr( nreset ) - 1
    qs = qs[bq]
    ql = ql[bq]
    for kr = 0, nreset - resetstep - 1, resetstep do begin
	nn = qs[kr+resetstep] - qs[kr] - 1
	km = lindgen( nn ) + qs[kr]
	tm = lindgen( nn ) + ( qs[kr] MOD ( 2 * !DPI ) )

	nnn = n_elements( tm ) * detstep
	dlist = detlist[kk:kk+detstep-1]
	jd = dlist - 1
	ss = (smoo[jd,*])[*,km]
	xx = (flux[jd,*])[*,km] - ss
	wt = (infl[jd,*])[*,km] EQ 0
;	xm = 10000 * total( abs( xx ) * wt, 2 ) / total( wt, 2 )

	wt = reform( wt, nnn )
	xx = reform( xx, nnn )	    ; / reform( xm # ( fltarr( nn ) + 1 ), nnn )
	ss = reform( ss, nnn )
	wn = reform( ( fltarr( detstep ) + 1 ) # tm, nnn )
	dd = reform( dlist # ( intarr( nn ) + 1 ), nnn )

	ool = 0.3
	bool = where( abs( xx ) GT ool, nbool )
	if ( nbool GT 0 ) then wt[bool] = 0


	if ( kr EQ 0 AND NOT keyword_set( silent ) ) then $
	  printf, unit, sbd + '      frin   cycle ' + $
		'amplitude     phase  chisq-red total-red', $
		( auto ) ? '  logOdds	logPr' : ''

	if ( n_elements( cyc ) LE 0 ) then cycle = bcycle[bnd] $
				      else cycle = cyc
	if ( n_elements( ncy ) LE 0 ) then ncycle = bncycle[bnd] $
				      else ncycle = ncy

	frid = { det	 : det, $
		 band	 : bnd, $
		 detlist : dd, $
		 sband	 : sbd, $
		 reset	 : kr }

; make a more finely spaced struct for plotting to better show the fringes
	five = 5
	nn5 = nn * five
	wn0 = min( wn, max=wn1 )
	wn5 = wn0 + findgen( nn5 ) * ( wn1 - wn0 ) / nn5
	ffr = { wn5  : wn5, $
		fr5  : fltarr( nn5 ), $
		nn5  : nn5, $
		plot : plot, $
		sil  : keyword_set( silent ) }
;	print, nn, nn5, wn0, wn1, min(wn5), max(wn5)

; these are the cycles to be searched for fringe components
	freq = cystep * findgen( ncycle ) + cycle

	flist = frin_search( wn, xx, wt, freq, nfringes, tol=tol, plot=ffr, $
		auto=auto, unit=unit, frid=frid, logpr=logpr, status=status )

	if ( status gt 0 ) then continue

; save the accumulating fringe list and evidence
	nflist = n_elements( flist )
	if ( size( flist, /type ) EQ 8 ) then begin
	    treset[kr] = km[nn/2]
	    model = fltarr( nn )
	    for kf = 0, nflist-1 do model += $
		( flist[kf].cosamp * cos( 2 * !pi * flist[kf].cycle * tm ) + $
		  flist[kf].sinamp * sin( 2 * !pi * flist[kf].cycle * tm ) )
	    for jd = 0, detstep - 1 do $
		modl[dlist[jd]-1,km] = model		    ; * xm[jd]

	    fringelist = append( fringelist, flist )
	endif else begin
	    nflist = 0
	endelse
	print_new_head = 1
	evidence = append( evidence, logpr )

	total = 0				; reset, done with this dataset
    endfor					; loop on resets

    if ( plot GT 0 ) then begin
      winset, 4

      for jj = kk, kk + detstep - 1 do begin
	jdet = detlist[jj]
	jd = jdet - 1
	time = lindgen( n_elements( flux[jd,*] ) )
	offt = ( n_elements( range ) EQ 2 ) ? range[0] : 0
	time += offt
	bb = where( infl[jd,*] EQ 0, nbb )
	if ( nbb EQ 0 ) then continue
	plot, time[bb], flux[jd,bb], position=[0.1,0.3,0.9,0.9], $
	    xtickname=replicate( ' ', 10 ), _extra=_extra, $
	    title=string( sbd, " Detector =", jdet, format='(2a,i3)' )
	oplot, time[bb], flux[jd,bb] - modl[jd,bb], color=3

	plot, cliprange( time ), [-1,1], position=[0.1,0.1,0.9,0.3], $
	    /nodata, /noerase, yr=[-1.0,1.0], _extra=_extra

	csz = 1.0 & ps0 = -1
	fnr = 1
	b = where( fringelist.frin_nr EQ fnr, nb )
	while ( nb GT 0 ) do begin
	    ac = fringelist[b].cosamp
	    as = fringelist[b].sinamp
	    cy = fringelist[b].cycle

	    amp = sqrt( ac^2 + as^2 )				; amplitude
	    phs = atan( as, ac ) / !PI				; phase

	    tr = fringelist[b].reset


	    oplot, offt+[treset[tr]], [amp*1000], ps=ps0, color=3, symsize=csz
	    oplot, offt+[treset[tr]], [cy], ps=ps0, symsize=csz
	    plot_phase, offt+treset[tr], ac, as, color=[2,8], /oplot, symsize=csz
	    csz *= 0.6 & ps0 = 1
	    fnr++
	    b = where( fringelist.frin_nr EQ fnr, nb )

	endwhile
      endfor
    endif

endfor						; loop on detectors

return, flux - modl		    ; return corrected flux

end

