;#> fringes.dc1
; Identifier   fringes
;
; Purpose      remove fringes in FIS data stream
;
; Synopsis     flux = fringes_cs( tsd, detlist, nrknots=nrknots, $
;		    nfringes=nfringes, tol=tol, smooth=smoo, range=range, $
;		    unit=unit, stepsize=stepsize, cycle=cycle, ncycle=ncycle, $
;		    cystep=cystep, verbose=verbose, weight=wgt, plot=plot, $
;		    fringelist=fringelist, auto=auto, evidence=evidence,
;		    status=status, _extra=_extra )
;
; Arguments    Name	 I/O   Type:	 Description:
;	       -------------------------------------------------------------
;	       tsd	  I    struct	 TSD
;	       detlist	  I    intarr	 list of detectors to be processed
;	       nrknots	  I    int	 number of knots in cuboc spline
;					 default: 10
;	       stepsize   I    int	 length of batches to be processed
;					 default: 100 * samplerate
;	       nfringes   I    int	 number of fringes to be extracted
;					 default: 1
;	       smooth	  O    fltarr	 smoothed background
;	       auto	  I    keyword	 automatic determination of the number
;					 of fringes to extract
;					 default (unless TOL is set)
;	       nfringes   I    int	 number of fringes to be removed
;					 default = 50
;	       tol	  I    real	 reduce the chisq until reduction is
;					 less than tol ( 0.01 == 1 percent )
;					 default = 0
;	       cycle	  I    real	 starting cycle to search for fringes
;					 default band1 all    : 4000
;					 default band2 aper 2 : 4000
;					 default band2 aper 3 : 5500
;					 default band3 all    : 3000
;	       ncycle	  I    int	 number of cycles to check
;					 default band1 all    : 1000
;					 default band2 all    : 1000
;					 default band3 aper 1 : 3000
;					 default band3 other  :  600
;	       cystep	  I    real	 step in the cycles
;					 default = 1
;	       range	  I    fltarr[2] limit operations to samples
;					 within range.
;					 default: all samples
;	       weight	  O    fltarr	 weights as calculated	smoothin
;	       verbose	  I    int	 amount of blabbering
;					 0 means silent
;					 default: 1
;	       plot	  I    int	 0 : plot the result ( default )
;					 -1: no plotting
;					 1-3: various plotting options
;	       fringelist O    struct	 cumulative list of extracted fringes
;		{ bandet	 int	 band or detector number
;		  updown	 string  up or down or both
;		  aotband	 string  aot band
;		  frin_nr	 int	 fringe number
;		  cycle 	 float	 fringe cycle
;		  cosamp	 float	 amplitude of cosine component
;		  sinamp	 float	 amplitude of sine component
;		  chisq 	 float	 chisq
;		  chi_red }	 float	 accumulative chisq reduction
;	       evidence   O	flt(arr) evidence == log( probability)
;					 (undefined if not auto)
;	       status	  O	int	 exit status
;					  0: succesfully processed
;					  1: error, input struct is returned
;
; Returns      flux	  O	 dblarr  corrected flux
;
;
; Description
;	       To find fringes we fit a cosine function to the data
;		   f(n) = alfa.cos( p.n + phi ) == A.cos(p.n) + B.sin(p.n)
;	       A, B and p are parameters to be estimated. p the the 'wavelength'
;	       of the fringe. To allow for the varying phase we replace
;	       A and B by cubic splines fits. As these are linear fits by
;	       themselves, the whole system is linear and can be (relatvely)
;	       easily fited to the data.
;
;	       In automatic mode we calculate the (bayesian) evidence for
;	       this (new) fringe. We keep the fringe if this evidence is
;	       larger than the evidence carried by the previous (set of)
;	       fringe(s). The process of identifying new fringes continues
;	       as long as the evidence increases with each new fringe. For
;	       all fringes the amplitudes are deterimined and the total
;	       fringe-set is subtracted. This is the advised mode.
;
;	       In non-automatic mode fringes are extracted until the reduction
;	       in chi-sq is less than TOL or the number of fringes is larger
;	       than NFRINGES
;
;	       Before the fitting is started a smoothed version of the data
;	       is subtracted from itself, so that all the data are around zero.
;	       This smoothed version is at the end added in again.
;	       Smoothing is somewhat of an art;
;	       different options have to be tried to get the best result.
;
;	       Depending on the value of PLOT several sets of plots pop up.
;	       -1 or /NOPLOT: no plotting.
;	       0. The result of the fringe removal is plotted. Original data
;		  in white, smoothed background in red. The defringed data
;		  are displayed in green, offsetted somewhat for clarity. And
;		  in blue the (scaled) RSRF is plotted. In the lower panel
;		  the background-removed data are plotted in white (wgt=1),
;		  red (wgt=0) or yellow (0<wgt<1). Overplotted in blue are
;		  the fringe complex which is found. And in red (when FILL was
;		  present) the extra fill-factor.
;	       1. Three panels with the background subtracted data in white/
;		  yellow/red depending on its weights. The total of the
;		  fringe complex is overplotted in blue. All these are
;		  plotted against wavenumber.
;		  The top panel repeats the next plot, with all extracted
;		  fringes in green
;	       2. The parameters A (white) and B (yellow) as a function
;		  of p (cycles per wavenumber).
;		  Chi-sq (scaled and red) is also plotted as function of p. In
;		  green is the wavelength of the selected fringe.
;		  Also a preliminary version of 1 is popping up.
;	       3. Several plots of the background smoothing algoritm
;		  (for debugging only)
;
;	       See also the Examples.
;
; Comment      See Kester etal. 2001 "SWS Fringes and Models"
;		   in ISO Legacy Conference Proceedings.
;
;	       For 3 reasons this procedure is superior to the use of FFTs.
;	       1. The wavelength of the fringe can be estimated much better
;	       than in an FFT where it is limited to the Nyquist frequency,
;	       which in turn is determined by the number of datapoints.
;	       2. It works on unevenly spaced data just as well as on a
;	       regularly gridded data set. Even when there are large gaps
;	       in the data it works fine.
;	       3. It automatically determines how many fringe components are
;	       present, avoiding possible overfitting and underfitting.
;
; Example      To defringe an TSD
;		    flux = fringes( tsd, nfringes=1, /plot )
;
; Filename     fringes.pro
;
; Author       Do Kester
;
; Version      2.12
;
; History      1.0  14-05-96 DK, first implementation
;		.....
;	       2.12 02-02-04 WF  quality output
;	       3.0  07-03-08 DK, rewritten for AKARI
;#<

function fringes_cs, tsd, detlist, nrknots=nrknots, nfringes=nfringes, $
	tol=tol, smooth=smoo, range=range, unit=unit, stepsize=stepsize, $
	cycle=cycle, ncycle=ncycle, cystep=cystep, verbose=verbose, $
	weight=wgt, plot=plot, fringelist=fringelist, auto=auto, $
	evidence=evidence, status=status, _extra=_extra

compile_opt idl2

common fringes, dmarr, dmpar, dmpnt

cpu = systime(1)

status = 0B

type = tsd->get_datatype()
if ( n_elements( detlist ) EQ 0 ) then $
    detlist = ( ( type EQ 'FIS_SW' ) ? indgen( 100 ) : indgen( 75 ) ) + 1
ndet = n_elements( detlist )
bnd = ( type EQ 'FIS_SW' ) ? 0 : 1
sbd = ( type EQ 'FIS_SW' ) ? 'FIS_SW' : 'FIS_LW'

if ( n_elements( plot ) EQ 0 ) then plot = 0
if ( n_elements( unit ) EQ 0 ) then unit = -1		; stdout
if ( n_elements( verbose ) EQ 0 ) then verbose = 1

if ( n_elements( tol ) EQ 0 ) then tol = 0.0
auto = keyword_set( auto ) OR ( tol LE 0.0 AND n_elements( nfringes ) LE 0 )
;if ( n_elements( nfringes ) LE 0 ) then nfringes = ( tol GT 0 OR auto ) ? 50 : 1
if ( n_elements( nfringes ) LE 0 ) then nfringes = 1

flux = tsd->get_value_tsd( 'fis_obs', 'flux' )
rstf = tsd->get_pix_flag_tsd( 'fis_obs', 'reset' )
infl = tsd->get_pix_flag_tsd( 'fis_obs', 'saturate' ) OR $
       tsd->get_pix_flag_tsd( 'fis_obs', 'no_diff' ) OR rstf

if ( type EQ 'FIS_LW' ) then samplerate = 16
if ( type EQ 'FIS_SW' ) then samplerate = 24

dead = where( tsd->get_pix_flag_tsd( 'fis_obs', 'dead' ) )
infl[dead] = rstf[dead] 	     ; dont eliminate dead dets

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Make a smooth version of the flux, and calculate (Tukey) weights
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
smoo = frin_cs_smooth( flux, infl, detlist=detlist, wgt=wgt )

if ( n_elements( range ) EQ 2 ) then begin
    b = lindgen( range[1] - range[0] ) + range[0]
    flux = flux[*,b]
    rstf = rstf[*,b]
    infl = infl[*,b]
    smoo = smoo[*,b]
    wgt  = wgt[*,b]
endif

modl = make_array( size=size( flux ) )

nrpoint = ( size( flux ) )[2]

if ( n_elements( nrknots ) EQ 0 ) then nrknots = 10
if ( n_elements( stepsize ) EQ 0 ) then stepsize = 100 * samplerate
samperknot = stepsize / nrknots

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; divide the timeline into reasonable chunks (~stepsize)
; avoid mixing of modes: normal-scan mode = 66, cds mode = 65
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
mode = tsd->get_value_tsd( 'fis_obs', 'PACKETID' )
cuts = where( mode - shift( mode, 1 ) NE 0 )
cuts = append( cuts, nrpoint )
if ( cuts[0] LE 0 ) then cuts = cuts[1:*]    ; remove starting 0 or -1

cutst = 0
for n = 0, n_elements( cuts) - 1 do begin
   cutlen = cuts[n] - cutst
   stp = stepsize < cutlen
   knt = stp / samperknot
   repeat begin
      step = append( step, stp )
      knot = append( knot, knt )
      cutlen -= stepsize
   endrep until ( cutlen - stepsize LE 0 )
   if ( cutlen GT 0 ) then begin
      lst = n_elements( step) - 1
      if ( cutlen LT 2 * samperknot ) then begin
	 step[lst] += cutlen
	 knot[lst] = step[lst] / samperknot
      endif else begin
	 step[lst] = ( step[lst] + cutlen ) / 2
	 step = append( step, step[lst] )
	 knot[lst] = step[lst] / samperknot
	 knot = append( knot, knot[lst] )
      endelse
   endif
   cutst = cuts[n]
endfor
stst = total( step, /cumulative ) - step[0]    ; start of each step

;print, cuts
;print, step
;print, knot
;steps = nrpoint / stepsize
;stepsize = ceil( float( nrpoint ) / ( steps + 1 ) )

; set default for cycle and ncycle for bands SW and LW
bcycle	= [ 7.5, 4.75 ]
bncycle = [ 40, 10 ]
bstep	= [0.05, 0.005]
if ( n_elements( cycle	) EQ 0 ) then cycle  = bcycle[bnd]
if ( n_elements( ncycle ) EQ 0 ) then ncycle = bncycle[bnd]
if ( n_elements( cystep ) EQ 0 ) then cystep = bstep[bnd]

; these are the cycles to be searched for fringe components
freq = cystep * findgen( ncycle ) + cycle

eps = 0.001 * stepsize / samplerate
tm = dindgen( stepsize ) / samplerate
if ( nrknots LE 1 ) then knots = 1 $
else knots = -eps + findgen( nrknots ) * ( stepsize / samplerate + 2 * eps ) / $
			( nrknots - 1 )

desmat = desmat_cs_sines( tm, knots, freq[0], dmcs=dm )
dmasiz = size( desmat )
dmpar = dmasiz[1]
dmpnt = dmasiz[2]
dmarr = dblarr( dmpar, dmpnt, ncycle )
for kf = 0, ncycle-1 do begin
   desmat = desmat_cs_sines( tm, knots, freq[kf], dmcs=dm )
   dmarr[*,*,kf] = reform( desmat, dmpar, dmpnt, 1 )
endfor

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; loop over the detectors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
for kk = 0, ndet - 1 do begin
  det = detlist[kk]
  kd  = det - 1

  frin_cs_reset, 0				 ; reset the smart searching
  start = 0
  starttime = 0
  for kr = 0, n_elements( step ) - 1 do begin	 ; loop over the chunks

	nn = step[kr]
	km = lindgen( nn ) + start
;	tm = ( dindgen( nn ) + ( start MOD ( 2 * !DPI ) ) ) / samplerate
	tm = dindgen( nn ) / samplerate
	start += step[kr]

	xx = reform( flux[kd,km] - smoo[kd,km], nn, /overwrite )
	wt = reform( wgt[kd,km], nn, /overwrite )

	if ( kr EQ 0 AND verbose GT 0 ) then $
	  printf, unit, sbd + "     frin nk     cycle " + $
		"amplitude    chisq  ", ( auto ) ? "  logOdds    logPr" : ""

	frid = { det	 : det, $
		 band	 : bnd, $
		 sband	 : sbd, $
		 reset	 : kr }

	plotstruct = {	plot	: plot, $
			verbose : verbose }

; search for fringes on this chunk of data
	flist = frin_cs_search( tm, xx, wt, freq, nfringes, nrknots=knot[kr], $
		tol=tol, plot=plotstruct, start=starttime, type=type, $
		auto=auto, unit=unit, frid=frid, logpr=logpr, status=status )

;	if ( status gt 0 ) then continue

; save the accumulating fringe list and evidence
	nflist = n_elements( flist )

	if ( size( flist, /type ) EQ 8 ) then begin
	    model = fltarr( nn )
	    for kf = 0, nflist-1 do model += $
		( *(flist[kf].cosamp) * cos( 2 * !pi * flist[kf].cycle * tm ) + $
		  *(flist[kf].sinamp) * sin( 2 * !pi * flist[kf].cycle * tm ) )
	    modl[kd,km] = model

	    detflist = append( detflist, flist )
	endif else begin
	    nflist = 0
	endelse
	print_new_head = 1
	evidence = append( evidence, logpr )

	total = 0				; reset, done with this dataset
    starttime += step[kr] / samplerate
  endfor					; loop on resets

  if ( plot GE 2 ) then begin
     winset, 4

     offt = ( n_elements( range ) EQ 2 ) ? range[0] : 0
     time = dindgen( nrpoint ) / 24.0 + offt
     bb = where( infl[kd,*] EQ 0, nbb )
     if ( nbb EQ 0 ) then continue

     yr = cliprange( flux[kd,bb], [0.001,0.999] )
     plot, time[bb], flux[kd,bb], position=[0.1,0.3,0.9,0.9], yr=yr, $
		xtickname=replicate( ' ', 10 ), xst=3, yst=3, _extra=_extra, $
		title=string( sbd, " Detector =", det, format='(2a,i3)' )
     oplot, time[bb], smoo[kd,bb], color=5
     oplot, time[bb], smoo[kd,bb] + modl[kd,bb], color=2
     oplot, time[bb], flux[kd,bb] - modl[kd,bb], color=3

     plot, cliprange( time[bb] ), [-1.0,1.0], position=[0.1,0.1,0.9,0.3], $
		/nodata, /noerase, yr=[-1.3,1.3], xst=3

     nb = n_elements( detflist )

     for j = 0, nb-1 do begin
	    kr = detflist[j].reset
	    nn = step[kr]
	    time = ( dindgen( nn ) + stst[kr] + offt ) / 24

	    ac = *(detflist[j].cosamp)
	    as = *(detflist[j].sinamp)
	    amp = sqrt( ac^2 + as^2 )*10000		; amplitude
	    phs = atan( as, ac ) / !PI				; phase
	    nn = n_elements( phs )
;	    oplot, time, ac * 10000
;	    oplot, time, as * 10000
	    plot_phase, time, ac, as, /oplot, ninter=0, $
			color=[2,8] + detflist[j].frin_nr - 1, ps=[3,3]
	    oplot, time, amp, color=2 + detflist[j].frin_nr

     endfor
;	waitforcursor, cont

  endif

  if ( size( detflist, /type ) EQ 8 ) then $
			     fringelist = append( fringelist, detflist )
  if ( n_elements( detflist ) GT 0 ) then dummy = temporary( detflist )
endfor						; loop on detectors

if ( plot GE 1 AND ndet GT 1 ) then begin
    winset, 1

    xx = flux - smoo
    yy = xx - modl
    b = where( infl NE 0, nb )
    if ( nb GT 0 ) then begin
       xx[b] = 0
       yy[b] = 0
    endif
    before = fltarr( ndet + 2 )       ; +2: to make the hisogram look better
    after  = fltarr( ndet + 2 )
    detl2  = fltarr( ndet + 2 )
    for kk = 1, ndet do begin
       kd = detlist[kk-1]
       detl2[kk]  = kd
       before[kk] = median( abs( xx[kd-1,*] ) )
       after[kk]  = median( abs( yy[kd-1,*] ) )
    endfor
    detl2[0] = detl2[1] - 1
    detl2[ndet+1] = detl2[ndet] + 1
    yr = cliprange( [before,after], [0,1] )
    yr[0] = 0

    plot, detl2, before, ps=10, yr=yr, title="CS fit Reduction", $
	    ytitle="Noise level", xstyle=3, $
	    xtickname=replicate( " ", 10), position=[0.1,0.3,0.95,0.9]
    oplot, detl2, after, ps=10, color=3
    before[0] = 1
    before[ndet+1] = 1		     ; to make the division look beter
    plot, detl2, after/before, ytitle="amount", ps=10, yminor=0, /noerase, $
	    xtitle="detectors", xstyle=3, position=[0.1,0.1,0.95,0.3], yr=[0,1]

endif

;print, systime(1) - cpu

return, flux - modl		    ; return corrected flux

end

