; $Id: gb_periodic_noise.pro,v 1.3 2008/03/07 05:27:27 do Exp $
;
; Copyright (c) 2002, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	GB_periodic_noise.pro
; PURPOSE:
;	detect and correct for periodic noise
; CATEGORY:
;	Function. Green Box procedure in ASTRO-F pipeline.
; CALLING SEQUENCE:
;	Result = GB_periodic_noise( tsd, [Calib_Const_File = ccf, $
;	    periodiclist=periodiclist, pixel=pixel, plot=plot, $
;	    _extra=_extra] )
; INPUTS:
;	tsd		TSD object reference.
;	cal_const_file	Calibration constants file;
;			not yet implemented
;	pixel		operate on these pixels only
;			default: all
;	sinefit 	keyword: use sinefit method
;			default: use noise averaging method
;	SW		keyword: apply to SW only
;			default: both SW and LW
;	plot		make plots
;	_extra		for plotting.
; OUTPUTS:
;	tsd		TSD object reference. Deglitched TSD.
;	periodiclist	make list of the results, 1 line per detector
;			How to manage the summaries is under considerations.
;RETURN
;	!SUCCEED if function ended successfully.
;	!FAIL in failure.
; RESTRICTIONS:
;	Only works for SW. Needs to be tested more. Run before the deglitchers.
; PROCEDURE:
;
; EXAMPLE:
;	See README
; REFERENCE:
;	Baba et al. 2002 Nov. 21, ``ASTRO-F/FIS Survey Data Structure"
;
; MODIFICATION HISTORY:
;	Oct  2006   DK	Creation
;	Mar  2008   DK	add defringe option
;-

FUNCTION GB_periodic_noise, tsd, Calib_Const_File = ccf, $
	    periodiclist=periodiclist, pixel=pixel, plot=plot, $
	    sinefit=sinefit, SW=SW, _extra=_extra

;   ON_ERROR 2

;;; Starting GB_periodic_noise ;;;
    lmessage, 7,'Executing GB_periodic_noise...'

    type = tsd->get_datatype()
    if ( keyword_set( SW ) AND type NE 'FIS_SW' ) then return, 0B

    if ( n_elements( periodiclist ) GT 0 ) then begin
	if ( size( periodiclist, /type ) EQ 7 ) then listfile = periodiclist $
	else begin
	    infile = tsd->get_filename()
	    str_len = strpos( infile, ".fits" )
	    listfile = strmid( infile, 0, str_len ) + '_periodic_list.dat'
	endelse
	openw, glunit, listfile, /get_lun
    endif

    pltype = size( plot, /type )
    if ( pltype GT 0 ) then begin
	do_plot = 1
	if ( pltype EQ 7 ) then start_hc, plot $
	else do_plot = plot
    endif

    if ( type EQ 'FIS_SW' ) then begin
	detstep = 20
	if ( n_elements( pixel ) EQ 0 ) then pixel = indgen( 100 ) + 1
	nrknots = 10
	stepsize = 2400L
    endif
    if ( type EQ 'FIS_LW' ) then begin
	detstep = 15
	if ( n_elements( pixel ) EQ 0 ) then pixel = indgen( 75 ) + 1
	nrknots = 1
	stepsize = 160000L
    endif

    npixel = n_elements( pixel )

    if ( keyword_set( sinefit ) ) then begin
	flux = fringes_cs( tsd, pixel, nfringes=1, nrknots=nrknots, $
	    stepsize=stepsize, unit=glunit, plot=do_plot, status=status, $
	    verbose=0, _extra=_extra )
    endif else begin
	flux = perinoise( tsd, pixel, noisemodel=noisemodel, detstep=detstep, $
	    tukscale=tukscale, unit=glunit, flags=flags, $
	    plot=do_plot, status=status, _extra=_extra )
    endelse

    if ( status EQ 0 ) then begin
	status = tsd->set_value_tsd( 'fis_obs', 'flux', flux )
	if ( NOT keyword_set( sinefit ) ) then $
	    status = tsd->set_pix_flag_tsd( 'fis_obs', 'no_peri_corr', flags )
    endif

    if ( n_elements( periodiclist ) GT 0 ) then $
	free_lun, glunit
    if ( pltype EQ 7 ) then stop_hc

;;; Write HISTORY in header ;;;

    lmessage, 7, 'Writing HISTORY',/CONTINUE
    hist_message = ' gb_periodic_noise.pro,v 0.0 done. CCF= '

    IF KEYWORD_SET(ccf) NE 1 THEN BEGIN
	hist_message = hist_message + 'NONE'
    ENDIF ELSE BEGIN
	hist_message = hist_message + ccf
    ENDELSE

    dummy = tsd_fxaddhist(tsd, hist_message)
    IF dummy NE 0 THEN $
	MESSAGE,'Unable to write HISTORY',/CONTINUE

;    print, systime(1) - syst

    return, status

END


