;#> perinoise.dc1
; Identifier   perinoise
;
; Purpose      remove periodic noise in FIS data stream
;
; Synopsis     flux = perinoise( tsd, [detlist, noisemodel=noisemodel,
;			tukscale=tukscale, detstep=detstep, status=status,
;			flags=flags, range=range, unit=unit, yrange=yrange, $
;			plot=plot, _extra=_extra] )
;
; Arguments    Name	 I/O   Type:	 Description:
;	       -------------------------------------------------------------
;	       tsd	  I    struct	 FIS data
;	       detlist	  I    intarr	 list of detectors to be processed
;	       noisemodel O    fltarr	 resulting noise model
;					 one line per detector row
;	       tukscale   I    float	 Tukey scale to de-weight outliers
;					 about 6 * noisescale
;					 default: 1e-4
;	       detstep	  I    int	 # detectors to be taken together
;					 default: one row (20 or 15)
;	       range	  I    fltarr[2] limit operations to samples
;					 within range.
;					 default: all
;	       plot	  I    int	 various plotting options
;					 0: no plotting
;					 1: Only noise reduction plot
;					 2: individual detector plots
;					 3: stop after each plot
;	       unit	  I    int	 unit number to write info to
;					 default : none
;	       status	  O    byte	 OK
;	       flags	  O    bytarr	 where the correction could no be done
;	       yrange	  I    fltarr	 for plotting
;	       _extra	  I		 for plotting
;
; Returns      flux	  O    fltarr	 corrected fro periodic noise
;
; Description
;
; Example      flux = perinoise( tsd, indgen(20)+1, plot=3 )
;
; Filename     perinoise.pro
;
; Author       Do Kester
;
; Version      1.0
;
; History      1.0  2006-10-05 DK, first implementation
;#<

function perinoise, tsd, detlist, noisemodel=noisemodel, tukscale=tukscale, $
	detstep=detstep, range=range, unit=unit, yrange=yrange, $
	flags=flags, plot=plot, status=status, _extra=_extra

compile_opt idl2

status = 0B

type = tsd->get_datatype()
if ( n_elements( detlist ) EQ 0 ) then $
    detlist = ( ( type EQ 'FIS_SW' ) ? indgen( 100 ) : indgen( 75 ) ) + 1
ndet = n_elements( detlist )
bnd = ( type EQ 'FIS_SW' ) ? 0 : 1
sbd = ( type EQ 'FIS_SW' ) ? 'FIS_SW' : 'FIS_LW'

if ( n_elements( plot ) EQ 0 ) then plot = 0
;if ( n_elements( unit ) EQ 0 ) then unit = -1		; stdout

if ( n_elements( tukscale ) EQ 0 ) then tukscale = 1e-4
if ( n_elements( detstep ) EQ 0 ) then detstep = 20

flux = tsd->get_value_tsd( 'fis_obs', 'flux' )
rstf = tsd->get_pix_flag_tsd( 'fis_obs', 'reset' )
nodif = tsd->get_pix_flag_tsd( 'fis_obs', 'no_diff' )
infl = tsd->get_pix_flag_tsd( 'fis_obs', 'saturate' ) OR $
       tsd->get_pix_flag_tsd( 'fis_obs', 'no_rp_corr' ) OR $
       rstf OR nodif

; cannot use the calibration flashes
if ( type EQ 'FIS_LW' ) then $
    calstat = tsd->get_status_tsd( 'fis_obs', 'calalon' )
if ( type EQ 'FIS_SW' ) then $
    calstat = tsd->get_status_tsd( 'fis_obs', 'calason' )
calstat = calstat OR shift( calstat, 4 ) OR shift( calstat, -2 )
infl = infl OR ( make_array( n_elements( infl[*,0] ), /byte, value=1 ) # $
		 calstat )
;print, size( infl )

nsamples = n_elements( flux[0,*] )

if ( n_elements( range ) EQ 2 ) then begin
    b = lindgen( range[1] - range[0] ) + range[0]
    flux = flux[*,b]
    rstf = rstf[*,b]
    infl = infl[*,b]
    nodif = nodif[*,b]
    time = b
    nsamples = range[1] - range[0]
endif else time = lindgen( nsamples )

noisemodel = fltarr( ndet / detstep, nsamples )
wgt  = float( infl EQ 0 )
before = fltarr( ndet )
after  = fltarr( ndet )
factor = fltarr( ndet )

kdet = 0
kmodel = 0
for kk = 0, ndet-detstep, detstep do begin

    dlist = detlist[kk:kk+detstep-1]
    jd = dlist - 1
    smofl = fltarr( detstep, nsamples ) 	    ; store for later use

; Construct weights scaled to the distance from a smoothed version of itself
    for k = 0, n_elements( jd ) - 1 do begin
	ff = flux[jd[k],*]
	smf = ff
	b = where( infl[jd[k],*] )
	smf[b] = !values.f_nan
	for i = 1, 3 do smf = smooth( smf, 3, /nan, /edge_truncate )
	dist = abs( ff - smf ) / tukscale	   ; scale to Tukey's scale
	ww = ( ( dist < 1 )^2 - 1 )^2		   ; make a Tukey weight
	b = where( infl[jd[k],*] EQ 0, nb )
	if ( nb GT 0 ) then wgt[jd[k],b] = wgt[jd[k],b] * ww[b]
	smofl[k,*] = smf
    endfor

; Take a weighted average of all fluxes in this detstep
    wf = total( wgt[jd,*], 1 )
    xx = total( flux[jd,*] * wgt[jd,*], 1 )
    bf = where( wf GT 0, nbf )
    if ( nbf GT 0 ) then xx[bf] /= wf[bf]

; Make a smoothed version agian, now from the averaged flux, and subtract
    bno = where( nodif[jd[0],*] )
    sm = xx
    sm[bno] = !values.f_nan
    for k = 1, 3 do sm = smooth( sm, 3, /nan, /edge_truncate )
    xx -= sm
    xx[bno] = 0
    noisemodel[kmodel,*] = xx
    kmodel++
    xr = cliprange( xx, 0.01 )			    ; for plotting only

;plot, time, xx+sm, yr=[-1,10]*1e-4, ps=-1
;oplot, time, ss, color=5
;oplot, time, sm, color=2
;oplot, time, wf, color=3
;oplot, bno, fltarr( n_elements(bno) )-0.00005, ps=1, color=4
;stop

; Loop over the dets in this detstep
    for k = 0, n_elements( dlist ) - 1 do begin
	kd = jd[k]
	fl = flux[kd,*]
	fls = fl - smofl[k,*]
	b = where( infl[kd,*] EQ 0, nb )
	if ( nb EQ 0 ) then continue		; all data are flagged out

	flm = median( fl[b] )
	at = rotate_min( xx[b], fls[b], fmin=fmin )

	flux[kd,*] -= at * xx

	before[kdet] = median( abs( fls[b] ) )
	after[kdet]  = median( abs( fls[b] - at * xx[b] ) )
	factor[kdet] = at
	kdet++

; Plotting the results
	if ( plot GT 1 ) then begin
	  if ( n_elements( yrange ) NE 2 ) then begin
	    yr = cliprange( fl[b], 0.01 )
	    yp = yr
	    yr[0] = yr[0] < xr[0]
	    yr[1] = ( 3 * yr[1] ) > 0.001
	  endif else begin
	    yr = yrange
	    yp = yrange
	  endelse
	  plot, time[b], fl[b], position=[0.1,0.1,0.9,0.9], yr=yr, /nodata, $
		title=string( sbd, " Detector ", dlist[k], format='(2a,i3)')
;	  oplot, time, wgt[kd,*]*0.001, color=4
	  oplot, time[b], fl[b], ps=3
	  oplot, time[b], xx[b], color=2
	  oplot, time[b], fl[b] - xx[b], color=3
;	  oplot, time[b], sm[b], color=5
	  bf = where( infl[kd,*] NE 0, nbf )
	  oplot, bf, fltarr( nbf ) + !y.range[0], ps=1
	  bf = where( nodif[kd,*] NE 0, nbf )
	  oplot, bf, fltarr( nbf ) + !y.range[0], ps=1, color=4

	  rr = ( yp[1] - yp[0] ) > ( xr[1] - xr[0] )
	  xr[0] = ( xr[0] + xr[1] - rr ) / 2
	  xr[1] = xr[0] + rr
	  yp[1] = yp[0] + rr
	  xr = [-1,1]*1e-4
	  yp = xr + flm
	  plot, xx[b], fl[b], ps=3, position=[0.2,0.6,0.4,0.9], /noerase, $
		xticks=2, yticks=2, xminor=0, yminor=0, yr=yp, xr=xr
	  oplot, xr, at * xr + flm, color=3
	  xyouts, -0.00009, 0.00008 + flm, string( at, format='(f4.1)' )
	  oplot, xx[b], fl[b]- at*xx[b], ps=3, color=2
	  if ( plot GT 2 ) then waitforcursor, cont
	endif
    endfor

endfor						; loop on detectors

if ( keyword_set( plot ) ) then begin
    yr = cliprange( [before,after], [0,.99] )
    plot, detlist, before, ps=10, yr=yr, title="Periodic Noise Reduction", $
	    ytitle="Noise level", $
	    xtickname=replicate( " ", 10), position=[0.1,0.3,0.95,0.9]
    oplot, detlist, after, ps=10, color=3
    plot, detlist, factor, ytitle="amount", ps=10, yminor=0, /noerase, $
	    xtitle="detectors", position=[0.1,0.1,0.95,0.3]
endif

;; Print to a logfile (at unit)
if ( n_elements( unit ) GT 0 ) then begin
    for k = 0, ndet-1 do printf, unit, detlist[k], factor[k], before[k], $
		    after[k], format='(i3,f5.1,2f10.6)'
endif

flags = infl			; where the correction could not be done

return, flux

end

