; $Id: clipdt.pro,v 1.8 2008/03/18 11:14:57 makiuti Exp $
;
; Copyright (c) 2006-2008, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;   CLIPDT
;
; PURPOSE:
;   Clip out outlier from an array of data
;
; CATEGORY:
;   Function. Library for ASTRO-F pipeline.
;
; CALLING SEQUENCE:
;    outdata = clipdt(indata[, CLIPNUM=value, AVE=variable, STDDEV=variable, &'
;                     /TWICE, /MED, INDEX=variable, /PLOT])'
;
; INPUTS:
;   Data array
;
; OUTPUTS:
;   Output data array outlier of which has been clipped out
;
; MODIFICATION HISTORY:
;   Written by Makiuti, S
;   2006/05/10 modified by SHOH: for GB DC_resp correction
;   2006/10/04 modified by SHOH: add a new keyword Err=err (measure_err)
;   2008/03/17 Last modified by Makiuti
;
;-
FUNCTION CLIPDT, $
    data, $                   ; Data array
    t, $                      ; added by SHOH, used with function fitting
    CLIPNUM=clipnum, $        ; data are clipped at clipnum-sigma
    AVE=ave, ST_DEV=st_dev, $ ; average and std-dev values
    TWICE=twice, $            ; keyword to make clipping twice
    MED=med, $                ; keyword to replace average value by median value
    INDEX=idx, $              ; index number of the returned data
    PLOT=plot, $              ; keyword to display a result plot window
    FUNCTION_NAME=fnct, $     ; added by SHOH, function name for fitting
    TSTART=t0, $              ; added by SHOH, a param for function fitting
    ERR=err                   ; added by SHOH, array to be passed to measure_err of LMFIT

  IF N_params() EQ 0 THEN BEGIN
    print, '## USAGE ##'
    print, ' > outdata = clipdt(indata[, CLIPNUM=value, AVE=variable, STDDEV=variable, &'
    print, ' >             /TWICE, /MED, INDEX=variable, /PLOT])'
    return, -1
  ENDIF

  IF (N_Elements(clipnum) EQ 0) THEN clipnum=3.0
  LMESSAGE, 7, 'Clipping at '+strcompress(string(clipnum))+' sigma.', /PRINT

  num_data = N_elements(data)
  IF (num_data LE 1) THEN BEGIN
    LMESSAGE, 5, 'Input data must be array!'
    return, -1
  ENDIF

  IF Keyword_Set(fnct) THEN BEGIN
  ;; IF clip with a specified function fitting
  ;;    functions are assumed gb_cal_fit2 or gb_cal_fit3 for GB DC_resp correction

      dummy = Call_Function(fnct, data, t, t, YFIT=yFit, PARAMS=pa, TSTART=t0, ERR=err)
      res = data - yFit
      rms = sqrt(Total(res*res)/num_data)
      idx = Where(ABS(res) LT rms*clipnum, cnt)
;      stop
  ENDIF ELSE BEGIN
  ;; Default procedures
  ;;
      IF Keyword_set(med) THEN ave = median(data, /even) ELSE ave = mean(data)
      st_dev = stdev(data)

      IF (st_dev EQ 0) THEN BEGIN
        LMESSAGE, 5, 'SD = 0'
        return, -1
      ENDIF

      LMESSAGE, 7, 'Size of input data :'+strcompress(string(num_data)), /PRINT
      LMESSAGE, 7, 'Average :'+strcompress(string(ave)), /PRINT
      LMESSAGE, 7, 'StdDev  :'+strcompress(string(st_dev)), /PRINT

      idx = where( (data GT ave-clipnum*st_dev) AND (data LT ave+clipnum*st_dev) ,cnt)
      
  ENDELSE

  IF (cnt GT 0) THEN BEGIN
    outdata = data[idx]
  ENDIF ELSE BEGIN
    idx = -1
    return, -1
  ENDELSE

  msg = strcompress(string(cnt))+' out of'+ $
         strcompress(string(N_Elements(data)))+' samples remain.'
  LMESSAGE, 7, msg, /PRINT

  IF Keyword_set(twice) THEN BEGIN
    outdata = clipdt(outdata, CLIPNUM=clipnum, MED=med, AVE=ave, $
                                              ST_DEV=st_dev, INDEX=idx_2nd, ERR=err)
    IF (n_elements(outdata) GT 1) THEN BEGIN
      idx = idx[idx_2nd]
    ENDIF
  ENDIF

  IF Keyword_set(plot) THEN BEGIN
    plot, outdata
    oplot, [0,num_data], [ave, ave], color='00FF00'xl
    oplot, [0,num_data], [ave+st_dev, ave+st_dev], color='0000FF'xl
    oplot, [0,num_data], [ave-st_dev, ave-st_dev], color='0000FF'xl
  ENDIF

  return, outdata

END
