; $Id: gb_dc_responsivity_corr.pro,v 1.39 2008/03/21 07:13:07 makiuti Exp $
;
; Copyright (c) 2002-2008, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
;
;FUNCTION NAME:
;       GB_DC_Responsivity_Corr.pro
;
;PURPOSE:
;
;0. Basic Scheme
;
;  This is to correct the time-varying responsivity of (mostly) FIS SW/LW
;detectors.  Priciple of the correction is to use the output counts of the
;detectors when calibration lamps is lit that are presumably assumed to have
;contant brightness.  In other words, the output count (hereafter cal. data),
;where dark is subtracted, can be considered the responsivity or gain of the
;detectors.  By dividing the output counts obtained for targets (hereafter
;obs. data) with cal. data which are close to obs. data in time, the obs.
;data can be corrected for the time variation of responsivity.
;  Mathematically, the output, Vout can be modeled as follows:
;
;  Vout  = G Iin + Voff,
;
;where, Iin is the brightness of input and G and Voff may be time varying.
;
;  For 'dark' and 'cal. lamp on' mode,
;
;  Vdark =         Voff  (since no input)
;
;  Vcal  = G Io  + Voff
;
;  We can get two unknowns, G and Voff, and Iin can be obtained as follows.
;
;  Iin = (Vout - Vdark)/(Vcal - Vdark) * Io      ---------- (*)
;
;  If something like radiation from telescope itself is included, reduction
;scheme will be more complicated, and since it involves one more unknown, one
;more measurement is needed.  This will not be discussed here any more.  We assume
;that the case (*) mentioned above holds.
;
;1. Two kinds of options:
;
; - whether cal. data obtained with the 'open-shutter' status are used or not
; - use least square fit of cal. data over rather long time span or just use
;   adjacent cal. data (i.e., linear interpolation)
;
;  We need to take these two options into account due to the following
;reasons.
;
; (1) Basically neayby cal. data in both sides of observed data set should
;be used to correct the responsivity variation of detectors.  Usually
;calibration lamps are periodically lit for this purpose.  However, time
;interval of calibration with the closed shutter (~1000 seconds?) may be
;too long to reflect the true gain variation, particularly, after bias boost.
;  So it is considered to turn on calibration lamp more frequently (every
;~60 seconds?) but without closing the shutter.
;  This will help monitor the gain variation of detectors, but sources may be
;on detectors during this calibration period.  Therefore, some may prefer
;to use cal. data when the shutter is opened, while others not.
;
; (2) As escaping from SAA, bias boost will be applied in order to recover
;detectors after showers of cosmic rays.  However, this will result in rather
;rapid responsivity changes as shown for IRTS/FILM (Oh et al. 2004).  One
;may prefer to fit the responsivity between consecutive bias boosts as a whole
;with a combination of exponential functions as Oh et al. did, while others
;still may want to use conventional method using cal. data obtained before and
;after the obs. data under consideration.
;
;
;2. Typical time interval and duration
;
; - for 'open-shutter' : ~60 seconds and ~ a few hundred milli-seconds(?)
; - for 'closed-shutter' : ~1000 seconds and ~ a few seconds(?)
;
;
;3. Default options
;
;  We consider the following options be used as standard options but allow to
;use the different options.
;
; - 'open-shut' : calibration data obtained with shutter open
; - 'global-fit' : calibration is made using fitted responsivity curve over a certain
;           time interval
;
; * opposite of these are 'closed-shutter' and 'adj-fit'.
;
;  They will be controlled by two keyword parameters, SHUT_CLO and ADJ_FIT in
;  the program, where if they are set (1), they mean 'closed-shutter' and
;'adj-fit', respectively.
;
;4. Responsivity variation due to glitches out of SAA
;
;  Since detectors are about the same as those of IRTS/FILM, we expect that
;the responsivity change follows that of IRTS/FILM.  In IRTS/FILM, for orbits
;which do not pass SAA, responsivity does not change much, exhibiting
;monotonic decrease or increase.  But after glitches and responsivity
;changes suddenly with a time scale of >1000 seconds.  Of course, we do not
;apply bias boost in these cases.
;  If this occurs, we apply global fitting for time spans separated by
;glitch.  Though we added an algorithm to find the instant of glitch, it may
;not be so satisfactory.  It may be better to use status flag on glitch which
;will be set by other glitch finding modules. (as of July 2004)
;  We use the same fitting function as used for SAA region.  The functional
;form used for SAA also allows to fit responsivity variation after glitch
;outside of SAA region.
;
;
;CATEGORY:
;       Function. Green Box procedure in ASTRO-F pipeline.
;
;CALLING SEQUENCE:
;       Result = GB_Dc_Responsivity_Corr(tsd)
;
;INPUTS:
;       tsd             TSD object reference
;       cal_table       Cal table made in the previous procedure,
;                       It is automatically searched from CALTABLE/ directory
;
;OUTPUTS:
;       tsd             TSD object reference. Responsivity-corrected TSD.
;
;RETURN:
;       !SUCCEED if function ended successfully.
;       !FAIL in failure.
;
;RESTRICTIONS:
;       NONE
;
;PROCEDURE:
;
;1. Read data, status, and flags.
;2. check options. If none, use default options.
;3. Correct cal. data in the beginning of the file, since they may be in error.
;(Here we do not read cal. data from cal. table for better correction of obs.
;data in the both ends of a file, as was done in gb_dark_subtraction, since
;particularly for cal. data with 'open-shutter', we need obs. data as well in
;order to derive pure contribution by cal. lamp.  However, they are not recorded
;in the cal. file.  We may record obs. data taken before and after open-shutter
;cal. data as well in the cal. file, which may violate the policy of making and
;maintaining cal. table files.)
;4. If 'adj-fit'
;     select cal. data according to shutter status
;     for each channel
;       if 'open-shutter' then
;         derive pure contribution by cal. lamp subtracting obs. data taken
;                                               before and after cal. data
;         get cal. data by averaging each set of cal. data
;       else
;         get cal. data after simple transient correction
;       endelse
;       quadratic interpolation
;       correct (divide) obs. data with these interpolated values
;     endfor
; endif
;5. If 'closed-shutter' and 'global-fit'
;     print, waring, since data points are not sufficient
;6. If 'open-shutter' and 'global-fit'    --- THIS IS RECOMMENDED.
;    select cal. data according to shutter status
;    identify bias boost
;    for each channel
;      for each bias boost
;        get cal. data
;        derive pure contribution by cal. lamp subtracting obs. data taken
;                                              before and after cal. data
;        get cal. data by averaging each set of cal. data
;        search for sudden responsivity jump (glitch) and then gradual decay
;        for each sudden responsivity jump
;          global fitting using Oh's functional form
;        endfor
;      endfor --- bias boost loop
;    endfor --- channel loop
; endif
;
;7. write data
;
;8. write flag
;
;9. write history
;
;AUXILIARY PROCEDURE:
;    FUNCTION gb_cal_fit1.pro - This will be used for simple transient
;                           correction of cal. data.  We are going to
;                           adopt the method Oh et al. (2004) used.
;
;    FUNCTION gb_cal_fit2.pro - An algorithm to make a fitting curve for
;                           time varying responsivity between bias boosts or
;                           after glitch outside of SAA, which we will also
;                           borrow from Oh et al. (2004).
;
;EXAMPLE:
;
;REFERENCE:
;       Baba et al., "ASTRO-F/FIS Survey Data Structure"
;       Oh et al. 2004, "Analysis of IRTS/FILM Responsivity in Orbit",
;                        submitted to PASJ
;
;MODIFICATION HISTORY:
;       Oct. 2002    Written by A. Kawamura.
;       Jun. 2004    Modified by YS (adoption of options, global
;                    fitting algorithm, and many error handlings)
;       Oct. 2006    Modified by SangHoon OH
;       Mar. 2007    Modified by Makiuti
;       Jun. 2007    Modified by SangHoon OH
;       Mar. 2008    Modified by Makiuti
;
;-
;
FUNCTION GB_DC_Responsivity_Corr, tsd, $
           SHUT_CLO = shut_clo, ADJ_FIT = adj_fit, $
           TABLEDIR=tabledir, $    ; path of the directory from which table files are read
           RESULT_OUT = result_out ; to make output files for results (.ps)

;ON_ERROR 2

  ;;; Starting GB_DC_Responsivity_Corr ;;;
  LMESSAGE, 5, 'Executing GB_DC_Responsivity_Corr...', /CONTINUE

  ;; Check if the TSD file contains SAA info.
  tmsaa = tsd->Get_Value_TSD('SE','TM_SAA')
  IF total(tmsaa) EQ 0 THEN BEGIN
    LMESSAGE, 1, 'No TM_SAA (SAA info.) in the file.'
  ENDIF

  ; Set the path of the directory from which table files are read.
  IF (n_elements(tabledir) EQ 0) THEN tabledir='.' ; current directory by default
  table_dir = tabledir + '/CALTABLE'

  ; Table file division unit
  fldiv=1.D5  ;  seconds, table file is generated at this rate

  ;;; Set default options
  IF NOT keyword_set(shut_clo) THEN shut_clo=0
  IF NOT keyword_set(adj_fit)  THEN adj_fit=0

  ;;-----------------------
  ; Result Plot Option
  ;;-----------------------
  kplot = keyword_set(result_out)
  IF kplot THEN BEGIN
    LoadCT, 5
    thisDev = !D.Name
    Set_Plot, 'PS'
    !P.THICK=3
    !X.THICK=3
    !Y.THICK=3
    tsdName = tsd->Get_FileName()
    tsdBodyName = strmid(tsdName, 0, 21)
    !P.FONT=0  
    Device, FILE="rcfit_"+tsdBodyName+'.ps', /COLOR, XS=7, YS=5, XOFF=0.75, YOFF=5, $
            /INCHES, /TIMES
    clr = FSC_COLOR(['black', 'red', 'green', 'blue', 'orange', 'pink', 'purple', 'white'], $
          [0,5,10,15,20,25,30,255])
    clrs = [clr[2],clr[3],clr[1],clr[6]]
  ENDIF

  ;;-----------------------
  ;;; Extract values from TSD object reference ;;;
  ;;-----------------------
  in_flux = tsd->Get_value_tsd('FIS_OBS','Flux')
  ret     = af_finite(in_flux, flux)
  ferr    = tsd->Get_value_tsd('FIS_OBS','FErr')
  ti      = tsd->Get_value_tsd('FIS_OBS','AFtime')
  tmsaa = tsd->Get_Value_TSD('SE','TM_SAA')
  ; number of channels ;
    si = SIZE(flux)
    num_ch   = si[1]
    num_data = si[2]
      corr_flux = flux
      corr_ferr = ferr

;  ;;; Define duration of lighting by CalA for pulse(short) Cal.
;  ;   ## This values need to be optimized. ##
;      ## => The definition here was replaced by empirical values (2008/02/15)
;
;  IF (num_ch EQ 75) THEN BEGIN ;; LW
;    calAtime = 0.5     ; (05/11/26 since IT#4)
;    idura = fix(calAtime*16.86)+1 ; Sampling rate is 16.86Hz in Survey mode
;                                  ; => idura = 9
;  ENDIF ELSE IF (num_ch EQ 100) THEN BEGIN ;; SW
;    ;calAtime = 0.2     ; Old default
;    ;calAtime = 0.1399  ; Crossing time of point source in survey obs.
;    calAtime = 0.16    ; (05/11/26 since IT#4)
;    idura = fix(calAtime*25.28)+1 ; Sampling rate is 25.28Hz in Survey mode
;                                  ; => idura = 5
;  ENDIF


  ;;; Tolerance for comparing rms residuals of 4 fitting methods.
  ;;  ## Temporary values are set. Probably it depends on pixels.
  IF (num_ch EQ 75) then begin
    tol_rms = dblarr(75)
    tol_rms[0:44]  = 2d-7
    tol_rms[45:74] = 2d-7 
  ENDIF ELSE BEGIN
    tol_rms = dblarr(100)
    tol_rms[0:59]  = 2d-6
    tol_rms[60:99] = 2d-7
  ENDELSE

;;-----------------------
;; Read status and flags
;;-----------------------
op_shutter = tsd->Get_status_tsd('FIS_OBS','ShtOp')  ; '1' means shutter open
cl_shutter = 1 - op_shutter
pid = tsd->Get_Value_TSD('FIS_OBS', 'PacketID')

st_bad   = tsd->Get_Pix_Flag_TSD('FIS_OBS', 'Bad')
no_bad   = 1 - st_bad

IF (num_ch EQ 75) THEN BEGIN ; LW
  on_calA = tsd->Get_status_tsd('FIS_OBS','CalALon')
  calA_pls_cnt = tsd->Get_value_tsd('FIS_OBS','CalPlsCntAL')
  on_boost = tsd->Get_status_tsd('FIS_OBS','LWBoostOn')
  of_boost = 1 - on_boost
  caldac = tsd->get_value_tsd('FIS_OBS', 'DACCalAL')
  cal_dac_ok = make_array(n_elements(caldac), /byte)
  nonzero_cal_dac_idx = where(caldac NE 0, cnt)
  IF (cnt GT 0) THEN cal_dac_ok[nonzero_cal_dac_idx] = 1 ; DAC setting values not equal zero
ENDIF ELSE BEGIN ; SW
  on_calA = tsd->Get_status_tsd('FIS_OBS','CalASon')
  calA_pls_cnt = tsd->Get_value_tsd('FIS_OBS','CalPlsCntAS')
  on_boost = tsd->Get_status_tsd('FIS_OBS','SWBoostOn')
  of_boost = 1 - on_boost
  caldac = tsd->get_value_tsd('FIS_OBS', 'DACCalAS')
  cal_dac_ok = make_array(n_elements(caldac), /byte)
  nonzero_cal_dac_idx = where(caldac NE 0, cnt)
  IF (cnt GT 0) THEN cal_dac_ok[nonzero_cal_dac_idx] = 1 ; DAC setting values not equal zero
ENDELSE

on_calB = tsd->Get_status_tsd('FIS_OBS','CalBon')
of_calB = 1 - on_calB
calB_pls_cnt = tsd->Get_value_tsd('FIS_OBS','CalPlsCntB')

; Make status flag array to know if co-add mode (pid = $42) or not
on_coadd42 = BytArr(num_data)
idx_coadd42 = where(pid EQ 66, cnt_coadd)
IF (cnt_coadd GT 0) THEN BEGIN
  on_coadd42[idx_coadd42] = 1B
ENDIF
of_coadd42 = 1B - on_coadd42

; Blank & SAA statuses
st_blank = tsd->Get_flag_tsd('FIS_OBS','Blank')
no_blank = 1 - st_blank
in_in_saa= tsd->Get_flag_tsd('FIS_OBS','In_SAA')
no_in_saa= 1 - in_in_saa

; Flag to be set as a result
no_dccorr_flg = bytarr(num_ch, num_data)

;;-----------------------
;; Get quality indicators
;;-----------------------
qual_rp_data  = tsd->Get_quality_tsd('FIS_OBS','Qual_rp_data')
qual_rp_param = tsd->Get_quality_tsd('FIS_OBS','Qual_rp_param')
qual_rp_table = tsd->Get_quality_tsd('FIS_OBS','Qual_rp_table')

;;-----------------------
;; Correct status bit of 'CalAS(L)on' using 'CalAPlsCnt' and duration
;;   - Status of pulse (short) Cal is to be set again because the timing
;;     is supposed to be different from the real data.
;;   - Status of continuous Cal is not changed.
;;   - Pulse Cal is regarded to be done during shutter open, while continuous Cal
;;     is done during shutter close.
;;-----------------------

; Find out positions of frames at which CalA turned on
idum = where(calA_pls_cnt - shift(calA_pls_cnt, 1) NE 0, cnt)

IF (cnt GE 2) THEN BEGIN ; If CalA turned on more than once

  IF (on_calA[0]*op_shutter[0] EQ 1) THEN BEGIN ; If data begins during A Cal pulse On.
    dum = where(on_calA[0:15] EQ 1, cntp) ; '15' may be less than idura
    on_calA[0:cntp-1] = 1B
    on_calA[cntp:num_data-1L] = 0B
  ENDIF ELSE BEGIN
    tmp_idx = where(on_calA*op_shutter EQ 1, cnt_tmp) ; CalA ON but not continuous Cal
    IF (cnt_tmp GT 0) THEN on_calA[tmp_idx] = 0B
  ENDELSE

  ist=0 & IF (idum[0] EQ 0) THEN ist=1
  FOR i = ist, cnt-1 DO BEGIN

    ;ib = idum[i] ; until IT#4
    IF (num_ch EQ 100) THEN BEGIN
      ib = idum[i]+3 ; new for SW
      IF (ib GE num_data) THEN continue
    ENDIF ELSE BEGIN
      ib = idum[i]+1 ; for LW, since 2008/02/15
    ENDELSE

      IF (cl_shutter[ib] EQ 1) THEN continue ; If not pulse but continuous, then continue

      IF (num_ch EQ 100) THEN BEGIN ; for SW
        ; 5 by default, 8 for maneuver sampling mode
        IF (pid[ib] EQ 82) THEN idura =  8 ELSE idura = 5
      ENDIF ELSE BEGIN ; for LW
        ; 12 for every modes
        IF (pid[ib] EQ 82) THEN idura = 12 ELSE idura = 12
      ENDELSE
      ;ie = min([num_data, ib+idura-1L]) ; until IT#4
      ie = min([num_data-1, ib+idura]) ; new

      ;; Redefinition of CAL_ON
      on_calA[ib:ie] = 1B
    ENDFOR

  of_calA = 1 - on_calA

ENDIF ELSE BEGIN
  LMESSAGE, 3, 'Too small number of pulse CAL.', /CONTINUE
  return, !SUCCEED
ENDELSE

;;; to discard continuous (B) cal. data taken at transition period, special
;;; care is needed.
;
 idx_c=where(on_calB EQ 1, cnt)
 idx_b=where(idx_c - shift(idx_c, 1) NE  1)
 idx_e=where(idx_c - shift(idx_c,-1) NE -1)
 on_calB[idx_c[idx_b]] = 0   &   on_calB[idx_c[idx_b]+1] = 0
 on_calB[idx_c[idx_e]] = 0   &   on_calB[idx_c[idx_e]-1] = 0

 of_calB = 1 - on_calB

 ; Make status flag arrays for CDS and On_Gal_Plane
 on_cds = BytArr(num_data)
 on_gal = BytArr(num_data)
 idx_cds = where(pid EQ 65, ncds)
 IF (ncds GT 0) THEN BEGIN
     on_cds[idx_cds] = 1B
;; comment out by SHO (Jun '07)
;     idx_gal1 = where( ((ti - ti[idx_cds[0]]) GE -60.* 3) AND ((ti - ti[idx_cds[0]]) LT 0), ngal1)
;     if ngal1 GT 0 then on_gal[idx_gal1] = 1B ; 180sec before CDS
;     idx_gal2 = where( ((ti - ti[idx_cds[ncds-1]]) LE 60.* 3) AND ((ti - ti[idx_cds[ncds-1]]) GT 0), ngal2)
;     if ngal2 GT 0 then on_gal[idx_gal2] = 1B ; after CDS
 ENDIF
 no_cds = 1B - on_cds
; no_gal = 1B - on_gal

;;-----------------------
;; Check if sufficient number of data exist or not
;;-----------------------
ok1_dat = op_shutter * of_calA * of_calB * of_boost * no_blank * no_in_saa
ok2_dat = bytarr(num_ch,num_data)
FOR i = 0, num_ch-1 DO BEGIN
  ok2_dat[i,*] = ok1_dat[*] * no_bad[i,*]
ENDFOR

; though not sufficient condition....
idx_dat = where(ok1_dat GT 0.D, cnt)
IF (cnt EQ 0) THEN BEGIN
  LMESSAGE, 3, 'Not sufficient obs. data...Go back...'
  RETURN, !SUCCEED
ENDIF

;;----------------------------------------------------------
;; Define a Fitting Block between Bias Boosts
;;----------------------------------------------------------
idx_pos_tmsaa = where(tmsaa GT 0)
flg_pos_tmsaa = bytarr(n_elements(tmsaa))
flg_pos_tmsaa[idx_pos_tmsaa] = 1
dif_flg = flg_pos_tmsaa - shift(flg_pos_tmsaa, 1)
dif_flg[0] = 0 ; The first element in a differentiated array is invalid
fnum_bb_blk_str = where(dif_flg EQ 1, cnt_bb_blk_str) ; Frame# of a block between bias boosts starts
                                                      ;  at where TM_SAA become positive from negative.
                                                      ;  (also, near a bias boost end)

IF (cnt_bb_blk_str GT 0) THEN BEGIN
  fnum_bb_blk_end = fnum_bb_blk_str - 1  ; Frame# of a block between bias boosts end
  fnum_bb_blk_str = [0, fnum_bb_blk_str] ; add file begining frame
  fnum_bb_blk_end = [fnum_bb_blk_end, n_elements(ti)-1] ; add file end frame
  cnt_bb_blk = cnt_bb_blk_str + 1
ENDIF ELSE BEGIN ; No bias boost end point in the current file
  fnum_bb_blk_str = 0
  fnum_bb_blk_end = n_elements(ti) - 1
  cnt_bb_blk = 1
ENDELSE
bb_blk_str_ti = ti[fnum_bb_blk_str] ; AFTIME of block start
bb_blk_str_ti[0] = ti[0] - tmsaa[0] ; t@C擪܂ރubN̊JnTI͒Oboost-endTI
bb_blk_end_ti = ti[fnum_bb_blk_end] ; AFTIME of block end
bb_blk_end_ti[n_elements(bb_blk_end_ti)-1] = ti[num_data-1] + 2000
                               ; t@CŌ܂ރubN̏ITI̓t@CŌTI+2000sec

;;----------------------------------------------------------
;; Redefine fitting blocks
;; LONG ARRAY fnum_bg_blk_str[cnt_bg_blk], fnum_bg_blk_end[cnt_bg_blk]
;;----------------------------------------------------------
ra  = tsd->get_value_tsd('AOCU', 'RA')
dec = tsd->get_value_tsd('AOCU', 'DEC')
Euler, ra, dec, gal_lon, gal_lat, 1
; dummy = min(abs(gal_lat), idx_midplane)
; ti_midplane = ti[idx_midplane]

idx_on_gal = where(gal_lat LE 30. AND gal_lat GE -30., cnt_on_gal)
flg_on_gal = bytarr(num_data)
if (cnt_on_gal GT 0) then flg_on_gal[idx_on_gal] = 1B

dif_flg = fix(flg_on_gal) - shift(fix(flg_on_gal),1)
dif_flg[0] = 0B

fnum_gp_blk_str = where(dif_flg EQ  1, cnt_gp_blk_str)
fnum_gp_blk_end = where(dif_flg EQ -1, cnt_gp_blk_end)

IF (cnt_bb_blk GT 1) then begin

  if (cnt_gp_blk_str GT 0) and (cnt_gp_blk_end GT 0) then $
    fnum_bg_blk_sep = [fnum_bb_blk_str[1:cnt_bb_blk-1], fnum_gp_blk_str, fnum_gp_blk_end] $
  else if (cnt_gp_blk_str GT 0) and (cnt_gp_blk_end EQ 0) then $
    fnum_bg_blk_sep = [fnum_bb_blk_str[1:cnt_bb_blk-1], fnum_gp_blk_str] $
  else if (cnt_gp_blk_str EQ 0) and (cnt_gp_blk_end GT 0) then $
    fnum_bg_blk_sep = [fnum_bb_blk_str[1:cnt_bb_blk-1], fnum_gp_blk_end] $
  else fnum_bg_blk_sep = [fnum_bb_blk_str[1:cnt_bb_blk-1]]

  fnum_bg_blk_sep = fnum_bg_blk_sep[sort(fnum_bg_blk_sep)]

ENDIF ELSE BEGIN

  if (cnt_gp_blk_str GT 0) and (cnt_gp_blk_end GT 0) then $
    fnum_bg_blk_sep = [fnum_gp_blk_str, fnum_gp_blk_end] $
  else if (cnt_gp_blk_str GT 0) and (cnt_gp_blk_end EQ 0) then $
    fnum_bg_blk_sep = [fnum_gp_blk_str] $
  else if (cnt_gp_blk_str EQ 0) and (cnt_gp_blk_end GT 0) then $
    fnum_bg_blk_sep = [fnum_gp_blk_end] $
  else fnum_bg_blk_sep = -1 ;(cnt_gp_blk = 0) -1

ENDELSE

fnum_bg_blk_sep = fnum_bg_blk_sep[sort(fnum_bg_blk_sep)]

IF (size(fnum_bg_blk_sep, /N_Elements) EQ 1) THEN BEGIN
  IF (fnum_bg_blk_sep EQ -1) THEN BEGIN
    fnum_bg_blk_str = 0
    fnum_bg_blk_end = num_data-1
  ENDIF ELSE BEGIN
    fnum_bg_blk_str = [0, fnum_bg_blk_sep] 
    fnum_bg_blk_end = [fnum_bg_blk_sep-1, num_data-1]
  ENDELSE
ENDIF ELSE BEGIN
    fnum_bg_blk_str = [0, fnum_bg_blk_sep] 
    fnum_bg_blk_end = [fnum_bg_blk_sep-1, num_data-1]
ENDELSE

fnum_bg_blk_str = fnum_bg_blk_str[uniq(fnum_bg_blk_str)]
fnum_bg_blk_end = fnum_bg_blk_end[uniq(fnum_bg_blk_end)]

cnt_bg_blk_str = n_elements(fnum_bg_blk_str)
;cnt_bg_blk_end = n_elements(fnum_bg_blk_end)
cnt_bg_blk = cnt_bg_blk_str

IF ((cnt_gp_blk_str GT 0) AND (cnt_gp_blk_end GT 0)) THEN BEGIN

  if (cnt_gp_blk_str GT cnt_gp_blk_end) then begin ; off-planen܂on-planeŏI
    fnum_gp_blk_end = [fnum_gp_blk_end, num_data-1] 
  endif else begin
    if (cnt_gp_blk_str EQ cnt_gp_blk_end) then begin ; off-off 邢 on-on
      ;; if data begins and ends in galactic plane
      if ((fnum_gp_blk_end[0] LT fnum_gp_blk_str[0]) and $ ; on n܂Kr off ɂȂ
        (fnum_gp_blk_end[cnt_gp_blk_str-1] LT fnum_gp_blk_str[cnt_gp_blk_end-1])) then begin
        ;  t@C̍Ō on
        fnum_gp_blk_end = [fnum_gp_blk_end, num_data-1] 
        fnum_gp_blk_str = [0, fnum_gp_blk_str]
      endif
    endif else fnum_gp_blk_str = [0, fnum_gp_blk_str]
  endelse 

ENDIF ELSE BEGIN
  fnum_gp_blk_str = [0]
  fnum_gp_blk_end = [num_data-1] 
ENDELSE



;; Replace contents of BB blocks with BG blocks for test purpose.
;; In future, this should be removed and the name BB blocks will be
;; replaced with BG block in the following contents.

fnum_tmp_blk_str = fnum_bb_blk_str
fnum_tmp_blk_end = fnum_bb_blk_end
cnt_tmp_blk = cnt_bb_blk

fnum_bb_blk_str = fnum_bg_blk_str
fnum_bb_blk_end = fnum_bg_blk_end
cnt_bb_blk = cnt_bg_blk

bb_blk_str_ti = ti[fnum_bb_blk_str] ; AFTIME of block start
bb_blk_str_ti[0] = ti[0] - tmsaa[0] ; t@C擪܂ރubN̊JnTI͒Oboost-endTI
bb_blk_end_ti = ti[fnum_bb_blk_end] ; AFTIME of block end
bb_blk_end_ti[n_elements(bb_blk_end_ti)-1] = ti[num_data-1] + 2000
                               ; t@CŌ܂ރubN̏ITI̓t@CŌTI+2000sec

;;----------------------------------------------------------
;  Read CCF for Flat field correction
;   flat_ccf contains 'Ratio of A-Cal/(flat source)' and 'error'
;;----------------------------------------------------------
LMESSAGE, 7, "Read CCF for Flat field correction"
IF (num_ch EQ 100) THEN BEGIN ;; SW
  flat_ccf = CCF_Read_CCF('FLATS', tsd, HEADER=hdr, ERROR=err)
ENDIF ELSE IF (num_ch EQ 75) THEN BEGIN ;; LW
  flat_ccf = CCF_Read_CCF('FLATL', tsd, HEADER=hdr, ERROR=err)
ENDIF

;;-----------------------
;  Set values of TI at Gal. plane
;;-----------------------
bmin = min(abs(gal_lat), ibmin)
IF (abs(bmin) GT 0.1) THEN LMESSAGE, 5, "%% The time of galactic plane passage cannot be determined. This may lead to inappropriate block definition when external CALTAB is in use. "
t_b0_next = (t_b0_prev = (t_b0 = ti[ibmin]))
WHILE ((t_b0_prev -= 3000d) GT ti[0]) DO continue
WHILE ((t_b0_next += 3000d) LT ti[num_data-1]) DO continue
; 3000 sec is roughly corresponding to a half revolution.

;;----------------------------------------------------------
;; Read CAL table files
;;----------------------------------------------------------

  ;;-----------------------
  ;; Identify table files to be read
  ;;-----------------------

  ;; If there are any file to supplement data before and after the time
  ;    range in the current TSD file, those are read.
  flnm = intarr(4)
  timin = min(ti, max=timax) ; ti is AF-time

  ; Find files ; Up to 4000sec before the current TSD file
  flnm[0] = fix(timin/fldiv)
  IF fix(timin/fldiv) NE fix((timin-4000.D)/fldiv) THEN BEGIN
    flnm[1] = fix((timin-4000.D)/fldiv)
  ENDIF

  ; Find files : Up to 4000sec after the current TSD file
  flnm[2]=fix(timax/fldiv)
  IF fix(timax/fldiv) NE fix((timax+4000.D)/fldiv) THEN BEGIN
    flnm[3] = fix((timax+4000.D)/fldiv)
  ENDIF

  ;; Find files : For the current file
  ; Assortment (Find ID numbers for Cal files)
  tidn = fix(ti[21, *]/fldiv) ; ti is in AF-time; fldiv is an interval of the file.
                              ; tidn is the ID number of Cal files to be read.
  idx_idc = where(tidn - shift(tidn, 1) NE 0, cntds) ; index# of changing points of tidn
                                            ; cntds is the number of changing points of tidn

  IF (cntds EQ 0) THEN BEGIN ; If all data in the current file are to be in one file,
    cntds = 1                ;   only one file will be read.
    idx_idc = [0, N_elements(tidn)]
  ENDIF

  FOR i = 0, cntds-1 DO BEGIN
    fileID = tidn[idx_idc[i]]
  ENDFOR

  ;; Sort and uniq fileID numbers
  fileID = [flnm, fileID]
  fileID = fileID[uniq(fileID, sort(fileID))] ; File ID numbers (including 0)

  ;;-----------------------
  ;; Read table files
  ;;-----------------------
  dmy_flg=0
  ;; Select a band name
  IF (num_ch EQ 75)  THEN bnd='_LW_'
  IF (num_ch EQ 100) THEN bnd='_SW_'

  FOR i = 0, n_elements(fileID)-1 DO BEGIN

    ;; In case that the data of +-4000sec are in the same table file as others.
    ;; i.e. if flnm[1](flnm[3])=0
    IF (fileID[i] EQ 0) THEN Continue

    cal_file = table_dir+'/CAL'+bnd+string(fileID[i],format='(i4.4)')+'.fits'
    ret = file_test(cal_file)

    IF (ret EQ 0) THEN BEGIN ; If the file does not exist
      LMessage, 5, 'Info: Cal table ('+cal_file+') does not exist. ', /Continue
      continue
    ENDIF

    IF (ret NE 0) THEN BEGIN ; If the file does exist
      LMessage, 5, cal_file+' is read.', /Continue
      cal = MRDFITS(cal_file, 1, STATUS=status)

      ; Omit the data which are apart from the time of the current file more than 4000sec.
      IF (status GE 0) THEN   $
        ti_idx = where((cal.aftime GT t_b0_prev+500d) AND (cal.aftime LT t_b0_next-500d), cnt)

      IF (cnt GE 1) THEN BEGIN
        IF (dmy_flg EQ 0) THEN BEGIN ; for the first file
          ti_tab     = cal[ti_idx].aftime
          flux_tab   = cal[ti_idx].flux
          err_tab    = cal[ti_idx].err
          no_bad_tab = 1-cal[ti_idx].bad
          pid_tab    = cal[ti_idx].pid
          tmsaa_tab  = cal[ti_idx].tm_saa
          dmy_flg = 1
        ENDIF ELSE BEGIN ; for 2nd or the later files
          ti_tab     = [cal[ti_idx].aftime, ti_tab]
          flux_tab   = transpose([transpose(cal[ti_idx].flux),  transpose(flux_tab)])
          err_tab    = transpose([transpose(cal[ti_idx].err),   transpose(err_tab)])
          no_bad_tab = transpose([transpose(1-cal[ti_idx].bad), transpose(no_bad_tab)])
          pid_tab    = [cal[ti_idx].pid, pid_tab]
          tmsaa_tab  = [cal[ti_idx].tm_saa, tmsaa_tab]
        ENDELSE
      ENDIF

    ENDIF
  ENDFOR

  ;;; Sort and uniq dark data
  IF (n_elements(ti_tab) GT 0) THEN BEGIN
    idx_tmp = where(ti_tab GT 0.D, cnt)
    IF (cnt GT 0) THEN BEGIN
      idx_tmp    = uniq(ti_tab[sort(ti_tab)])
      ti_tab     = ti_tab[idx_tmp]
      flux_tab   = flux_tab[*,idx_tmp]
      err_tab    = err_tab[*,idx_tmp]
      no_bad_tab = no_bad_tab[*,idx_tmp]
      pid_tab    = pid_tab[idx_tmp]
      tmsaa_tab  = tmsaa_tab[idx_tmp]
    ENDIF
  ENDIF

;;----------------------------------------------------------
;; Begin DC responsivity correction.
;;----------------------------------------------------------
;;----------------------------------------------------------
;; Case 1
;;----------------------------------------------------------
;;; Use nearby cal. data, i.e., no global fit.
IF ADJ_FIT EQ 1 THEN BEGIN
;
  IF (shut_clo EQ 1) THEN BEGIN
    ok1_cal = cl_shutter * on_calB * of_boost * no_in_saa
  ENDIF ELSE BEGIN
    ok1_cal = op_shutter * on_calA * of_boost * no_in_saa
  ENDELSE

  ok2_cal = bytarr(num_ch,num_data)
  FOR i = 0, num_ch-1 DO BEGIN
    ok2_cal[i,*] = ok1_cal[*] * no_blank[*] * no_bad[i,*]
  ENDFOR

  ; see if cal. data exist, though not sufficient condition...
  idx_cal = where(ok1_cal GT 0, cnt_ok1_cal)

  IF (cnt_ok1_cal GT 5) THEN BEGIN
    inum_str = where(idx_cal - shift(idx_cal, 1) NE  1, cnt_calb) ; index# of Cal start
    inum_end = where(idx_cal - shift(idx_cal,-1) NE -1         ) ; index# of Cal end
                                                      ; cnt_calb is a number of Cal. blocks
    IF (cnt_calb LE 2) THEN BEGIN
       LMESSAGE, 3, 'Not sufficient cal. data...Go back...', /CONTINUE
       RETURN,!SUCCEED
    ENDIF
  ENDIF ELSE BEGIN
    LMESSAGE, 3, 'Not sufficient cal. data...Go back...', /CONTINUE
    RETURN,!SUCCEED
  ENDELSE

  cal_flux = dblarr(cnt_calb)
  cal_time = dblarr(cnt_calb)
  cal_tsaa = dblarr(cnt_calb)   ; SHOH : elapsed time after saa or bias boost
  ; derive cal. values

  FOR ch = 0, num_ch-1 DO BEGIN ; loop for channel or pixel
    FOR j = 0, cnt_calb-1 DO BEGIN
      ii_str = idx_cal(inum_str[j]) ; start frame# of 'j'th cal data
      ii_end = idx_cal(inum_end[j])

      ; check if whole data in a block are bad
      ibcal = where(ok2_cal[ch,ii_str:ii_end] EQ 1, cntp)
      IF (cntp LE 2) THEN continue

      IF (shut_clo EQ 0) THEN BEGIN
        ib1 = ii_str-10  &  IF (ib1 LT 0) THEN ib1 = 0
        ie1 = ii_str-1   &  IF (ie1 LT 0) THEN ie1 = 0
        idx_ok_precal = where(ok2_dat[ch,ib1:ie1] EQ 1, cnt_ok_precal)

        IF (cnt_ok_precal GE 3) THEN BEGIN
          titemp  = ti[idx_ok_precal[cnt_ok_precal-3:cnt_ok_precal-1]+ib1]
          valtemp = flux[ch,idx_ok_precal[cnt_ok_precal-3:cnt_ok_precal-1]+ib1]

          ib2 = ii_end+5  & IF (ib2 GE num_data-1L) THEN ib2 = num_data-1L
          ie2 = ii_end+15 & IF (ie2 GT num_data-1L) THEN ie2 = num_data-1L
          idx_ok_postcal = where(ok2_dat[ch,ib2:ie2] EQ 1, cnt_ok_postcal)

          IF (cnt_ok_postcal GE 3) THEN BEGIN
            titemp  = [titemp, ti[idx_ok_postcal[0:2]+ib2]]
            temp    = flux[ch, idx_ok_postcal[0:2]+ib2]
            valtemp = transpose([transpose(valtemp),transpose(temp)])
          ENDIF
        ENDIF ELSE BEGIN
          ib2 = ii_end+5  & IF (ib2 GT num_data-1L) THEN ib2 = num_data-1L
          ie2 = ii_end+15 & IF (ie2 GT num_data-1L) THEN ie2 = num_data-1L
          idx_ok_postcal = where(ok2_dat[ch,ib2:ie2] EQ 1, cnt_ok_postcal)
          IF (cnt_ok_postcal GE 3) THEN BEGIN
            titemp  = ti[idx_ok_postcal[0:2]+ib2]
            valtemp = flux[ch, idx_ok_postcal[0:2]+ib2]
          ENDIF
        ENDELSE

        IF (n_elements(valtemp) GT 4) THEN BEGIN
          ;coef = poly_fit(titemp, valtemp, 1, /double)
          ;flux[ch,ibcal+ii_str]=flux[ch,ibcal+ii_str]-(coef[0]+coef[1]*ti[ibcal+ii_str])
          ;cal_flux[j]=mean(flux[ch,ibcal+ii_str])
          ;; something wrong above; we take simple approach
          cal_flux[j] = mean(flux[ch,ibcal+ii_str]) - mean(valtemp)
          cal_time[j] = mean(ti[ibcal+ii_str])
          cal_tsaa[j] = mean(tmsaa[ibcal+ii_str]) ; SHOH : elapsed time since saa
        ENDIF
      ENDIF ELSE BEGIN
        cal_flux[j] = gb_cal_fit1(flux[ch,ibcal+ii_str], ti[ibcal])
                                                  ;!! not ti[ibcal+ii_str]
        cal_time[j] = mean(ti[ibcal+ii_str])
        cal_tsaa[j] = mean(tmsaa[ibcal+ii_str]) ; SHOH : elapsed time since saa
      ENDELSE

    ENDFOR

    ; to eliminate blocks where data in the blocks are all bad
    ok_flux_idx = where(cal_flux GT 0.D, cnt)
    IF (cnt GT 1) THEN BEGIN
      cal_fi = cal_flux[ok_flux_idx]
      tim_fi = cal_time[ok_flux_idx]
      tsa_fi = cal_tsaa[ok_flux_idx] ; SHOH
      ; interpolation
      ;     cal_tb[idx_dat]=interpol(cal_fi,tim_fi,ti[idx_dat],/quadratic)
      ;     corr_flux[ch,idx_dat]=flux[ch,idx_dat]/cal_tb[idx_dat]
      cal_tb = interpol(cal_fi,tim_fi,ti,/quadratic)
      cal_zero = where(cal_tb LE 0., cnt_zero)
      IF (cnt_zero GT 0) THEN BEGIN
         cal_tb[cal_zero] = +1.
         no_dccorr_flg[ch, cal_zero] = 1 ; set flag
         qual_rp_table[ch, cal_zero] = 3 ; set quality
      ENDIF
      corr_flux[ch,*] = flux[ch,*]/cal_tb[*]
     ;corr_flux[ch,*] = cal_tb[*] ; for debug
    ENDIF ELSE BEGIN
      no_dccorr_flg[ch, *] = 1 ; set flag
      qual_rp_table[ch, *] = 3 ; set quality
    ENDELSE

  ENDFOR ; End of loop for channel or pixel

ENDIF


;;----------------------------------------------------------
;; Case 2
;;----------------------------------------------------------
;;; Fit cal data with a curve fitted onto cal data obtained with CLOSED
;;; shutter over a longer time span

IF (ADJ_FIT EQ 0) AND (SHUT_CLO EQ 1) THEN BEGIN
  print, ''
  print, '  WARNING: THIS OPTION IS NOT RECOMMENDED.  '
  print, '  RETURN TO THE DEFAULT OPTION, ADJ_FIT=0 AND SHUT_CLO=0.  '
  print, ''
  SHUT_CLO = 0
ENDIF


;;----------------------------------------------------------
;; Case 3
;;----------------------------------------------------------
;;; Fit cal data with a curve fitted onto cal data obtained with OPEN
;;; shutter over a longer time span
IF (ADJ_FIT EQ 0) AND (SHUT_CLO EQ 0) THEN BEGIN

  ok1_cal = op_shutter * on_calA * of_boost * no_in_saa * cal_dac_ok

  ok2_cal = bytarr(num_ch, num_data)
  FOR ch = 0, num_ch-1 DO BEGIN
    ok2_cal[ch,*] = ok1_cal[*] * no_blank[*] * no_bad[ch,*] * on_coadd42[*]; * no_gal[*]
  ENDFOR

  ; see if cal data exist, though not sufficient condition...
  idx_cal = where(ok1_cal GT 0, cnt_ok1_cal)

  IF (cnt_ok1_cal GT 5) THEN BEGIN
    inum_str = where(idx_cal - shift(idx_cal, 1) NE  1, cnt_calb) ; index# of Cal start
    inum_end = where(idx_cal - shift(idx_cal,-1) NE -1          ) ; index# of Cal end
                                                      ; cnt_calb is a number of Cal blocks

    IF (cnt_calb LE 2) THEN BEGIN
       LMESSAGE, 3, 'Not sufficient cal. data...Go back...', /CONTINUE
       RETURN,!SUCCEED
    ENDIF
  ENDIF ELSE BEGIN
    LMESSAGE, 3, 'Not sufficient cal. data...Go back...', /CONTINUE
    RETURN,!SUCCEED
  ENDELSE

  ;; Values at where each Cal lamp turned on
  ; AFTIME
  cal_time = ti[idx_cal[inum_str]]
  ; Packet ID
  cal_pid = pid[idx_cal[inum_str]]
  ; TM_SAA
  cal_tsaa = tmsaa[idx_cal[inum_str]]

  ;;-----------------------
  ;; Loop for channel (pixel)
  ;;-----------------------
  MAX_NUM_BLOCK = 5
  rms_ch = dblarr(num_ch,MAX_NUM_BLOCK) ; root-mean-square of fitting residual will be stored
  rms_ch[*,*] = !VALUES.F_NAN

  FOR ch = 0, num_ch-1 DO BEGIN
    i_real_block = 0
    first_plot = 1B

    ; Result Plot Option
    IF kplot THEN BEGIN
      ;;!P.MULTI=[0,2,2,0,0]
      !P.MULTI = 0
      PlotSym, 0, 0.5, THICK=4
    ENDIF

    IF (cnt_calb LE 1) THEN BEGIN
      no_dccorr_flg[ch, *] = 1 ; set flag
      qual_rp_table[ch, *] = 3 ; set quality
      continue
    ENDIF

    cal_flux = dblarr(cnt_calb)
    cal_sigm = dblarr(cnt_calb)

    ;;-----------------------
    ;; Loop for Cal data
    ;;  - Get 'cal_flux' (flux value for Cal On) -
    ;;-----------------------
    FOR j_cal = 0, cnt_calb-1 DO BEGIN

      ii_str = idx_cal[inum_str[j_cal]]
        ; frame# of 'j'th cal start in the 'i'th data.
        ; The index count starts from the beginning of the TSD
      ii_end = idx_cal[inum_end[j_cal]]

      ; Check if whole data of a calibation signal are bad
      ibcal = where(flux[ch, ii_str:ii_end]*ok2_cal[ch, ii_str:ii_end] NE 0.D, cnt_ok_cal_frm)
      IF (cnt_ok_cal_frm LE 2) THEN continue ;; Doesn't need any quality remark for this block?

      ;;
      ;; Get 'valtemp' (a flux value before and after Cal ON).
      ;;
      ib1 = ii_str-10  &  IF (ib1 LT 0) THEN ib1 = 0
      ie1 = ii_str-1   &  IF (ie1 LT 0) THEN ie1 = 0
      idx_ok_precal = where(flux[ch, ib1:ie1]*ok2_dat[ch, ib1:ie1] NE 0.D, cnt_ok_precal)

      IF (cnt_ok_precal GE 4) THEN BEGIN
        titemp  = ti[idx_ok_precal[cnt_ok_precal-4:cnt_ok_precal-1]+ib1]
        valtemp = flux[ch, idx_ok_precal[cnt_ok_precal-4:cnt_ok_precal-1]+ib1]

        ib2 = ii_end+5  & IF (ib2 GT num_data-1L) THEN ib2 = num_data-1L
        ie2 = ii_end+15 & IF (ie2 GT num_data-1L) THEN ie2 = num_data-1L
        idx_ok_postcal = where(flux[ch, ib2:ie2]*ok2_dat[ch, ib2:ie2] NE 0.D, cnt_ok_postcal)

        IF (cnt_ok_postcal GE 4) THEN BEGIN
          titemp  = [titemp, ti[idx_ok_postcal[0:3]+ib2]]
          temp    = flux[ch, idx_ok_postcal[0:3]+ib2]
          valtemp = transpose([transpose(valtemp),transpose(temp)])
        ENDIF
      ENDIF ELSE BEGIN
        ib2 = ii_end+5  & IF (ib2 GT num_data-1L) THEN ib2 = num_data-1L
        ie2 = ii_end+15 & IF (ie2 GT num_data-1L) THEN ie2 = num_data-1L
        idx_ok_postcal = where(flux[ch, ib2:ie2]*ok2_dat[ch, ib2:ie2] NE 0.D, cnt_ok_postcal)

        IF (cnt_ok_postcal GE 4) THEN BEGIN
          titemp  = ti[idx_ok_postcal[0:3]+ib2]
          valtemp = flux[ch, idx_ok_postcal[0:3]+ib2]
        ENDIF
      ENDELSE

      IF (n_elements(valtemp) GT 4) THEN BEGIN

        cal_flux[j_cal] = mean(flux[ch, ibcal+ii_str]) - mean(valtemp)
        cal_sigm[j_cal] = stddev(valtemp)
          ;<- Now, error is the deviation of background (before and after the Cal)

        ;; set quality
        ;IF (cal_sigm GT 5e-5) THEN qual_rp_data[ch, ???]=1 ; set quality
      ENDIF

    ENDFOR ;; End of loop for Cal

    ;; eliminate blocks where data in the blocks are all bad
    ok_flux_idx = where(cal_flux GT 0.D, cnt_ok_cal)
    IF (cnt_ok_cal LE 4) THEN BEGIN
      no_dccorr_flg[ch, *] = 1 ; set flag
      qual_rp_table[ch, *] = 3 ; set quality
      continue
    ENDIF

    cal_fi = cal_flux[ok_flux_idx]
    sig_fi = cal_sigm[ok_flux_idx]
    tim_fi = cal_time[ok_flux_idx]
    tsa_fi = cal_tsaa[ok_flux_idx]
    pid_fi =  cal_pid[ok_flux_idx]


    ;;----------------------------------------------------------
    ;; Get final Cal values: merge and sort Cal data
    ;;----------------------------------------------------------
    ; Merge new Cal data (obtained from the current file) with Cal table values
    IF (n_elements(ti_tab) GT 0) AND (n_elements(tim_fi) GT 0) THEN BEGIN
      tim_fi = [tim_fi, reform(ti_tab[*])]
      cal_fi = [cal_fi, reform(flux_tab[ch,*])]
      sig_fi = [sig_fi, reform(err_tab[ch,*])]
      pid_fi = [pid_fi, reform(pid_tab[*])]
      tsa_fi = [tsa_fi, reform(tmsaa_tab[*])]
    ENDIF ELSE IF (n_elements(ti_tab) GT 0) THEN BEGIN
      tim_fi = [reform(ti_tab[*])]
      cal_fi = [reform(flux_tab[ch,*])]
      sig_fi = [reform(err_tab[ch,*])]
      pid_fi = [reform(pid_tab[*])]
      tsa_fi = [reform(tmsaa_tab[*])]
    ENDIF

    IF (n_elements(tim_fi) GT 0) THEN BEGIN
      ; Sort and uniq
      idx_temp = uniq(tim_fi, sort(tim_fi))
      ;; Get Cal data set for the current file and channel
      tim_fi = tim_fi[idx_temp]
      cal_fi = cal_fi[idx_temp]
      sig_fi = sig_fi[idx_temp]
      pid_fi = pid_fi[idx_temp]
      tsa_fi = tsa_fi[idx_temp]

      ;;---------------REMOVE ME ---------------
      ;; Exclude anomaly which occurs at the next two resp. data after
      ;; PID='52'xl ends, probably because they are often badly
      ;; extracted. 
      ;; This SHOULD be removed after this kind of anomaly is removed in
      ;; estimation of calibration flux.
      idx_pid52 = where(pid_fi EQ '52'xl, cnt_pid52)
      if (cnt_pid52 GT 0) then begin
        diff = idx_pid52 - shift(idx_pid52,-1)
        i_end_pid52 = where(diff LT -1, cnt_i_end_pid52)
        if cnt_i_end_pid52 GT 0 then begin
          pid_fi[idx_pid52[i_end_pid52]+1] = '52'xl
          pid_fi[idx_pid52[i_end_pid52]+2] = '52'xl
        endif 
      endif
      ;;------------End of REMOVE ME------------

      ; Rule out the data of other than co-add mode
      idx_coadd42 = where(pid_fi EQ 66, cnt_coadd)
      IF (cnt_coadd GT 0) THEN BEGIN
        tim_fi = tim_fi[idx_coadd42]
        cal_fi = cal_fi[idx_coadd42]
        sig_fi = sig_fi[idx_coadd42]
        pid_fi = pid_fi[idx_coadd42]
        tsa_fi = tsa_fi[idx_coadd42]
      ENDIF

      ; Rule out invalid (zero) data
      idx_notzero = where(cal_fi GT 0, cnt_notzero)
      IF (cnt_notzero GT 0) THEN BEGIN
        tim_fi = tim_fi[idx_notzero]
        cal_fi = cal_fi[idx_notzero]
        sig_fi = sig_fi[idx_notzero]
        pid_fi = pid_fi[idx_notzero]
        tsa_fi = tsa_fi[idx_notzero]
      ENDIF

      ; Rule out extremely high value
      idx_notextreme = where(cal_fi LT 100e-4, cnt_notextreme)
      IF (cnt_notextreme GT 0) THEN BEGIN
        tim_fi = tim_fi[idx_notextreme]
        cal_fi = cal_fi[idx_notextreme]
        sig_fi = sig_fi[idx_notextreme]
        pid_fi = pid_fi[idx_notextreme]
        tsa_fi = tsa_fi[idx_notextreme]
      ENDIF

    ENDIF ELSE BEGIN
      IF (ch EQ 0) THEN LMESSAGE, 3, 'No Cal data available in both table and the current files.'
         ; ^ this massage is shown only once.
      no_dccorr_flg[ch, *] = 1
      qual_rp_table[ch, *] = 3
      Continue
    ENDELSE

    ;;----------------------------------------------------------
    ;; Searching for Sudden Responsivity Jump
    ;;----------------------------------------------------------
    ; search for positions of sudden responsivity jump and then gradual decay
    
    ;
    cnt_cal_data = n_elements(tim_fi)
    ;
    jump_flg = IntArr(cnt_cal_data)
    sigPre   = DblArr(cnt_cal_data)

    ; Make an array of rms for previous NDEV frames
    NDEV = 5 ; number of frames to estimate rms levels
    IF (cnt_cal_data LT NDEV) THEN continue

    FOR cur=NDEV, cnt_cal_data-1 DO BEGIN
      curMean = median(cal_fi[cur-NDEV:cur-1])
      sigPre[cur] = StdDev(cal_fi[cur-NDEV:cur-1])
  
; Revision of JUMP DETECTION - NOT COMPLETED! -
;      idx_out = where(abs(cal_fi[cur-NDEV:cur-1]-curMean) GT 3.0*sigPre[cur], nOut)
;      if nOut NE 0 then begin
;        curMean = mean(cal_fi[cur-NDEV:cur-1])
    ENDFOR  

    sigPre[0:NDEV-1] = sig_fi[0:NDEV-1]
    sigPre[cnt_cal_data-1] = sig_fi[cnt_cal_data-1]

    ; Detect sudden jump more than ?-sigma
    ;  - sensitivity (?-sigma) should be optimized
    idx_jmp = where(abs(cal_fi - shift(cal_fi,1)) GT 5.5*sigPre, cnt_jmp)
    idx_sep = where((tim_fi - shift(tim_fi,1)) GT 1500, cnt_sep) ; seperated

    ; Confirm jumps are real or not
    IF (cnt_jmp GT 0) THEN BEGIN
      FOR j=0, cnt_jmp-1 DO BEGIN
        ist = idx_jmp[j]
        IF (ist EQ 0) OR (ist+5 GT cnt_cal_data-1) THEN continue
                              ; number "5" depends on the time scale
                              ; of the decay
        mval  = cal_fi[ist-1]
        refval= cal_fi[ist] - mval
        sigma = sigPre[ist-1]

        ; conditions of deviation
        cond_a1 = abs(cal_fi[ist+1]-mval) GT 3.*sigma
        cond_a2 = abs(cal_fi[ist+2]-mval) GT 3.*sigma
        cond_a3 = abs(cal_fi[ist+3]-mval) GT 2.*sigma
        cond_a4 = abs(cal_fi[ist+4]-mval) GT 2.*sigma
        cond_a5 = abs(cal_fi[ist+5]-mval) GT 2.*sigma

        ; conditions of positive and negative
        cond_b1 = refval*(cal_fi[ist+1]-mval) GT 0.
        cond_b2 = refval*(cal_fi[ist+2]-mval) GT 0.
        cond_b3 = refval*(cal_fi[ist+3]-mval) GT 0.
        cond_b4 = refval*(cal_fi[ist+4]-mval) GT 0.
        cond_b5 = refval*(cal_fi[ist+5]-mval) GT 0.

          ;;; # for check
          ;print, ch+1, j, $
          ;       cond_a1+cond_a2+cond_a3+cond_a4+cond_a5, $
          ;       cond_b1+cond_b2+cond_b3+cond_b4+cond_b5, $
          ;       (cond_a1+cond_a2+cond_a3+cond_a4+cond_a5 GE 4) AND $
          ;           (cond_b1+cond_b2+cond_b3+cond_b4+cond_b5 GE 4)

        ; Judge on conditions
        IF (cond_a1+cond_a2+cond_a3+cond_a4+cond_a5 GE 3) AND $
           (cond_b1+cond_b2+cond_b3+cond_b4+cond_b5 GE 3) THEN BEGIN
          jump_flg[ist] = 1
        ENDIF
        if cnt_sep GT 0 then jump_flg[idx_sep] = 1
      ENDFOR
    ENDIF

    ;; Regard jump positions as breakpoints of fitting block
    ;
    ; Index of "responsivity-jump Cal"
    ser_jump_cal = where(jump_flg EQ 1, cnt_jmp_cal)

    IF (cnt_jmp_cal GT 0) THEN BEGIN ; If any jump detected Wvoꂽꍇ
      jump_cal_ti = tim_fi[ser_jump_cal]

      ser_in_tsd = where(jump_cal_ti GT ti[0] and jump_cal_ti LT ti[num_data-1], cnt_in_tsd)
      IF (cnt_in_tsd GT 0) THEN BEGIN ; If there is a "jump" in the current TSD

        ; TI at where a jump was detected inside the current TSD file
        jump_cal_ti_in_tsd = jump_cal_ti[ser_in_tsd]
        ; Frame# corresponding above TIs
        fnum_jump_cal = LonArr(cnt_in_tsd)
        FOR i = 0, cnt_in_tsd-1 DO BEGIN
          fnum_jump_cal[i] = where(jump_cal_ti_in_tsd[i] EQ ti)
          ; if reprocessing is performed after re-creation of TSD, it might cause an error 
          ; because TI is slightly changed (return value of where() might be -1).
          ; This part should be reconsidered.
        ENDFOR
        ; Wvt[ubNX^[gt[ɉ
        ; Frame# of a jump is included in the block start frame#
        fnum_blk_str = [fnum_bb_blk_str, fnum_jump_cal]
        fnum_blk_str = fnum_blk_str[uniq(fnum_blk_str, sort(fnum_blk_str))]  
        blk_str_ti = [bb_blk_str_ti, jump_cal_ti_in_tsd]
        blk_str_ti = blk_str_ti[uniq(blk_str_ti, sort(blk_str_ti))]  

        fnum_blk_end = fnum_blk_str - 1     ; Re-create frame# of block ends
        fnum_blk_end = fnum_blk_end[where(fnum_blk_end GE 0)] ; Remove "-1"
        fnum_blk_end = [fnum_blk_end, num_data-1] ; Add file end frame
        blk_end_ti = ti[fnum_blk_end]             ; AFTIME of block end

        cnt_blk = n_elements(fnum_blk_str) ; Num. of fitting blocks

      ENDIF ELSE BEGIN ; If there is NO "jump" in the current TSD
        fnum_blk_str = fnum_bb_blk_str
        blk_str_ti = bb_blk_str_ti
        fnum_blk_end = fnum_bb_blk_end
        blk_end_ti = bb_blk_end_ti
        cnt_blk = cnt_bb_blk
      ENDELSE

      ; ݂TSDt@C擪܂ރubÑt@C擪OŌoꂽWvCal̃VAԍ
      ser_pre_tsd = where(jump_cal_ti GT bb_blk_str_ti[0] and jump_cal_ti LT ti[0], cnt_pre_tsd)
      ; In the case that there is a jump in the block which includes file beginning
      IF (cnt_pre_tsd GT 0) THEN BEGIN
        ; ubNJnTIWvTIɃZbg
        ; (Wv̏ꍇ͍Ō)
        blk_str_ti[0] = jump_cal_ti[max(ser_pre_tsd)]
      ENDIF

      ; ݂TSDt@CŌ܂ރubÑt@CŌŌoꂽWvCal̃VAԍ
      file_end_ti = ti[n_elements(ti)-1]
      ser_post_tsd = where(jump_cal_ti GT file_end_ti and jump_cal_ti LT file_end_ti+1000, cnt_post_tsd)
      ; In the case that there is a jump in the block which includes file end
      IF (cnt_post_tsd GT 0) THEN BEGIN
        ; ubNITIWvTIɃZbg
        ; (Wv̏ꍇ͍ŏ̃Wv)
        blk_end_ti[n_elements(blk_str_ti)-1] = jump_cal_ti[min(ser_post_tsd)]
      ENDIF

    ENDIF ELSE BEGIN ; When no jump detected
      fnum_blk_str = fnum_bb_blk_str
      blk_str_ti = bb_blk_str_ti
      fnum_blk_end = fnum_bb_blk_end
      blk_end_ti = bb_blk_end_ti
      cnt_blk = cnt_bb_blk
    ENDELSE

    on_gp_blk = bytarr(cnt_blk)
    for iblk=0, cnt_blk-1 do begin
      idum = where(((fnum_gp_blk_str - fnum_blk_str[iblk]) LE 0), ndum)
      if (ndum GT 0) && (fnum_gp_blk_end[idum[ndum-1]] GE fnum_blk_end[iblk]) then $
        on_gp_blk[iblk] = 1B
    endfor

    ; Set jump_flg=1 at where new block after BiasBoost started
    tmp = jump_flg
    tmp[*] = 0
    tmp_bb = tmp
    
    diff_tsa_fi = tsa_fi - shift(tsa_fi, 1)
    just_aft_bb = where(diff_tsa_fi LT 0, cnt1)
    IF (cnt1 GT 0) THEN BEGIN
      jump_flg[just_aft_bb] = 1
      tmp_bb[just_aft_bb] = 1
    ENDIF

    id_jump = where(jump_flg EQ 1, cnt_jmp)
    id_bb_jump = where(tmp_bb EQ 1, cnt_bb_jmp)

    ;;-----------------------
    ;; Loop for fitting block
    ;;  - a block begins from the end of a bias-boost or a sudden jump of responsivity
    ;;-----------------------
    FOR i_block = 0, cnt_blk-1 DO BEGIN
      ;;; # for check
      ;print, '***** ch.', strcompress(string(ch+1)), ' loop for BB', strcompress(string(i_block)), $
      ; ' of', strcompress(string(cnt_blk))

      i_blk_str = fnum_blk_str[i_block]
        ; Start frame# of 'i_block'th data.
        ; The frame# starts from the beginning of the TSD.
      i_blk_end = fnum_blk_end[i_block]
        ; End frame# of 'i_block'th data.

      ; check data
      dummy = where(ok1_cal[i_blk_str:i_blk_end] GT 0, cnt_ok1_cal_blk)
      IF (cnt_ok1_cal_blk LE 0) THEN BEGIN
        no_dccorr_flg[ch, i_blk_str:i_blk_end] = 1 ; set flag
        qual_rp_table[ch, i_blk_str:i_blk_end] = 3 ; set quality
        continue
      ENDIF

      ; Index array of Cal data in the current block to be fitted
      id_this_blk_cal = where((tim_fi GE blk_str_ti[i_block]) AND $
                              (tim_fi LE blk_end_ti[i_block]), cnt_cal_this_blk)
      ; If no Cal data in the current block
      IF (cnt_cal_this_blk EQ 0) THEN continue
      ; Extract Cal data in the current block
      tim_fi_blk = tim_fi[id_this_blk_cal]
      cal_fi_blk = cal_fi[id_this_blk_cal]
      tsa_fi_blk = tsa_fi[id_this_blk_cal]
      sig_fi_blk = sig_fi[id_this_blk_cal]

      ;;-----------------------
      ;; Do responsivity correction
      ;;-----------------------
      IF (num_ch EQ 100) THEN BEGIN ; for SW
        fitFunction = ['gb_cal_fit_lin', 'gb_cal_fit2_sw', 'gb_cal_fit3_sw', 'gb_cal_fit3f_sw']
        IF (tsa_fi_blk[0] LT 100) THEN BEGIN
          t0 = blk_str_ti[i_block] ; TI from which a fitting start (= BB end)
        ENDIF ELSE BEGIN
           t0 = ti[i_blk_str]
        ENDELSE
      ENDIF ELSE BEGIN
        IF (num_ch EQ 75) THEN BEGIN ; for LW   
          fitFunction = ['gb_cal_fit_lin', 'gb_cal_fit2_lw', 'gb_cal_fit3_lw', 'gb_cal_fit3f_lw']
          IF (tsa_fi_blk[0] LT 100) THEN BEGIN
            t0 = blk_str_ti[i_block] ; TI from which a fitting start (= BB end)
          ENDIF ELSE BEGIN
            t0 = ti[i_blk_str]
          ENDELSE  
        ENDIF
      ENDELSE
      nMethod = n_elements(fitFunction)
      nparams = [2,5,7,7]
      cal_tb_arr = dblarr(nMethod, i_blk_end - i_blk_str + 1)
      
      ; Result Plot Option
      IF kplot THEN BEGIN
        calFit_arr = dblarr(nMethod, i_blk_end - i_blk_str + 1)
        tim_fi_clip_arr = dblarr(nMethod, i_blk_end - i_blk_str + 1)
        cnt_cal_clip_arr = lonarr(nMethod)
      ENDIF
      
      rms_arr = dblarr(nMethod) & pa_arr = dblarr(nMethod,7)

;       IF (num_ch EQ 100) THEN BEGIN ; for SW
;         IF (tsa_fi_blk[0] LT 100) THEN BEGIN
;           t0 = blk_str_ti[i_block] ; TI from which a fitting start (= BB end)
;           fitFunction = 'gb_cal_fit3_sw'
;         ENDIF ELSE BEGIN
;           t0 = ti[i_blk_str]
;           fitFunction = 'gb_cal_fit2_sw'
;         ENDELSE
;       ENDIF ELSE IF (num_ch EQ 75) THEN BEGIN ; for LW
;         IF (tsa_fi_blk[0] LT 100) THEN BEGIN
;           t0 = blk_str_ti[i_block] ; TI from which a fitting start (= BB end)
;           fitFunction = 'gb_cal_fit3_lw'
;         ENDIF ELSE BEGIN
;           t0 = ti[i_blk_str]
;           fitFunction = 'gb_cal_fit2_lw'
;         ENDELSE
;       ENDIF

;      if ch EQ 59 then stop
      
      ; Result Plot Option
      if (kplot and first_plot) then begin
        plotsym, 0,0.5, THICK=4
        Plot, tim_fi-ti[0], cal_fi*1d4, $
          XTITLE = 'aftime!X [+' + $
          String(F='(e13.6)', t0)+'sec]', $
          YTITLE = 'R!X [x10!E-4!NV]', PSYM = 8, CHARS = 1.0, $
          YR=[0, 15], YSTYLE=1, XTHICK=3, YTHICK=3
        first_plot=0B
      endif


      IF ((cnt_cal_this_blk GE 8) AND (NOT on_gp_blk[i_block])) THEN BEGIN
        for iMethod=0,nMethod-1 do begin
          paTmp = 0
;          if (ch + 1) EQ 4 then stop
          cal_fi_clip = $
            clipdt(cal_fi_blk, tim_fi_blk, clipnum=3.0, index=idx, $
                   FUNCTION_NAME=fitFunction[iMethod], TSTART=t0, ERR=sig_fi_blk, TWICE=0)

          IF (size(cal_fi_clip, /N_Elements) EQ 1) THEN BEGIN
            IF (cal_fi_clip EQ -1) then begin
              continue
            ENDIF
          ENDIF

          tim_fi_clip = tim_fi_blk[idx]
          sig_fi_clip = sig_fi_blk[idx]
          cnt_cal_this_blk_clip = n_elements(idx)
;      cal_tb_arr[nMethod,*] contains calibration information for
;      the current TSD file which is obtained by using ith method
;      calFit contains the fitted result at the position of
;      calibration lamp.
          if (iMethod EQ 2) AND (tsa_fi_blk[0] GT 200) AND (num_ch EQ 100) then begin
            FITA=[1,1,1,1,1,0,1] 
            cal_tb_arr[iMethod,*] = $
              Call_Function(fitFunction[iMethod], cal_fi_clip, tim_fi_clip, $
                            ti[i_blk_str:i_blk_end], YFIT=calFit, $
                            PARAMS=paTmp, TSTART=t0, CHISQ=chisq, $
                            ITER=iter, ITMIN=10, ERR=sig_fi_clip, FITA=FITA)
          endif else begin
            cal_tb_arr[iMethod,*] = $
              Call_Function(fitFunction[iMethod], cal_fi_clip, tim_fi_clip, $
                            ti[i_blk_str:i_blk_end], YFIT=calFit, $
                            PARAMS=paTmp, TSTART=t0, CHISQ=chisq, $
                            ITER=iter, ITMIN=10, ERR=sig_fi_clip)
          endelse
          res = calFit-cal_fi_clip
          rms_arr[iMethod] = sqrt(total(res^2)/cnt_cal_this_blk_clip)
          pa_arr[iMethod,0:nparams[iMethod]-1] = paTmp
          date = cnv_aftime(blk_str_ti[i_block])

          ; Result Plot Option
          IF kplot THEN BEGIN
            tim_fi_clip_arr[iMethod,0:cnt_cal_this_blk_clip-1] = tim_fi_clip
            calFit_arr[iMethod,0:cnt_cal_this_blk_clip-1] = calFit
            cnt_cal_clip_arr[iMethod] = cnt_cal_this_blk_clip

;             PlotSym,0,0.5, THICK=4
            if iMethod EQ 0 then begin
;               if i_real_block EQ 0 then begin 
;                 Plot, tim_fi-ti[0], cal_fi*1d4, $
;                   XTITLE = 'aftime!X [+' + $
;                   String(F='(e13.6)', t0)+'sec]', $
;                   YTITLE = 'R!X [x10!E-4!NV]', PSYM = 8, CHARS = 1.0, $
;                   YR=[0, 15], YSTYLE=1, XTHICK=3, YTHICK=3

;               endif
              plotsym, 0,0.5, FILL=1
              oplot, tim_fi_clip-ti[0], cal_fi_clip*1d4, PSYM=8
            endif

;            OPlot, ti[i_blk_str:i_blk_end]-ti[0], $
;            cal_tb_arr[iMethod,*]*1d4, LINE=0, COLOR=clrs[iMethod]
;            oplot, tim_fi_clip-ti[0], calFit*1d4, LINE=1, COLOR=clrs[iMethod]
;            oplot, ti[i_blk_str:i_blk_end:10]-ti[0], cal_tb_arr[iMethod,0:*:10]*1d4, LINE=0, COLOR=clrs[iMethod]

          ENDIF                 ; END of if KPLOT

        endfor                  ; END of loop for 4 methods.
      ENDIF ELSE BEGIN          ; If not (cnt_cal_this_blk GE 8)
        IF (cnt_cal_this_blk GE 3) THEN BEGIN
          IF (NOT on_gp_blk[i_block]) THEN BEGIN
            cal_fi_clip = clipdt(cal_fi_blk, tim_fi_blk, clipnum=3.0, index=idx, /MED)
            
            IF (n_elements(cal_fi_clip) EQ 1) THEN BEGIN
              IF (cal_fi_clip EQ -1) THEN BEGIN
                no_dccorr_flg[ch, i_blk_str:i_blk_end] = 1 ; set flag
;               qual_rp_table[ch, i_blk_str:i_blk_end] = 3 ; set quality
                continue
              ENDIF
            ENDIF ELSE BEGIN
              tim_fi_clip = tim_fi_blk[idx]
              cnt_cal_this_blk_clip = n_elements(idx)
            ENDELSE
            
          ENDIF ELSE BEGIN
            cal_fi_clip = cal_fi_blk
            tim_fi_clip = tim_fi_blk
          ENDELSE
        ENDIF ELSE BEGIN
          no_dccorr_flg[ch, i_blk_str:i_blk_end] = 1 ; set flag
;          qual_rp_table[ch, i_blk_str:i_blk_end] = 3 ; set quality
          continue
        ENDELSE
      ENDELSE
        
      IF ((cnt_cal_this_blk GE 8) AND (NOT on_gp_blk[i_block])) THEN begin

        ;; Fitted functions are not allowed to have negative values or
        ;; extremely large values such as larger than 100e-4 V.
        i_not_diverge = where(cal_tb_arr[*,0] GT 0d and cal_tb_arr[*,0] LT 100d-4,n_not_diverge)
        ;; If all 4 functions have such negative or extremely large
        ;; values, then just choose the simplest method, linear fit.
        if n_not_diverge EQ 0 then begin
          LMESSAGE, 3, "%% NO Proper Fits among the 4 methods %%", /CONTINUE
          i_not_diverge = [0,1,2,3] & iMin=0 ;; Choose linear fit.
        endif else begin
          rms_min = min(rms_arr[i_not_diverge], iMin)
          ;; If the fitting quality in terms of RMS is NOT so different,
          ;; choose the simpler one. The critical value of the
          ;; difference should be stored in 'tol_rms[100]'
          for iM=0, n_not_diverge-1 do begin
            if ((rms_arr[i_not_diverge[iM]] - rms_min) LE tol_rms[ch]) then begin
              iMin = i_not_diverge[iM]
              break
            endif 
          endfor

        endelse
        cal_tb = reform(cal_tb_arr[iMin, *])
        pa = pa_arr[iMin]

        ; Result Plot Option
        IF kplot THEN BEGIN
          items = ['CH. '+string(F='(i3)',ch+1), $
                   't!I!N0='+date, $
                   'rms (linear fit ) ='+string(f='(e11.3)',rms_arr[0]), $
                   'rms (2-comp exp.) ='+string(f='(e11.3)',rms_arr[1]), $
                   'rms (3-comp exp.) ='+string(f='(e11.3)',rms_arr[2]), $
                   'rms (3f function) ='+string(f='(e11.3)',rms_arr[3]) $
                  ]
          prefix = ['  ','  ','  ','  ']
          prefix[iMin] = ' V'
          items[2:5] = prefix + items[2:5]
          tcolors = [clr[0],clr[0],clrs]
          legend, items, LEFT=1, POS=[0.11+i_real_block*0.25, 0.9], box=0,textcolor=tcolors, color=colors, chars=0.7, /NORM, CHARTHICK=2
          rms_ch[ch,i_real_block] = rms_arr[iMin]
          print, ' CH='+ string(F='(i3)',ch)+ $
            '  RMS_MIN='+ string(F='(e13.4)',rms_ch[ch,i_real_block])+$
            ' by '+ fitFunction[iMin]
          plotsym, 0, 0.5, /FILL
          oplot, tim_fi_clip_arr[iMin,0:cnt_cal_clip_arr[iMin]-1]-ti[0], $
            calFit_arr[iMin,0:cnt_cal_clip_arr[iMin]-1]*1d4, LINE=2, COLOR=clrs[iMin], thick=3
          idx_corr = where(ti[i_blk_str:i_blk_end] LE tim_fi_blk[cnt_cal_this_blk-1] AND ti[i_blk_str:i_blk_end] GE tim_fi_blk[0] , cnt_corr)
          if cnt_corr GT 0 then $ 
            oplot, ((ti[i_blk_str:i_blk_end])[idx_corr])[0:*:10]-ti[0], (cal_tb[idx_corr])[0:*:10]*1d4, LINE=0, COLOR=clrs[iMin], thick=3
          i_real_block++
;          stop
        ENDIF


        ; Flat field correction and Normalize (0.001V for standard pulse cal.)
        cal_tb = cal_tb/flat_ccf[ch, 0]/0.001
      ENDIF ELSE BEGIN
        cal_tb = interpol(cal_fi_clip, tim_fi_clip, ti[i_blk_str:i_blk_end], /quadratic) 
        
        ; Result Plot Option
        IF kplot THEN BEGIN
          idx_corr = where(ti[i_blk_str:i_blk_end] LE tim_fi_blk[cnt_cal_this_blk-1] AND ti[i_blk_str:i_blk_end] GE tim_fi_blk[0] , cnt_corr)
          IF cnt_corr GT 0 THEN $ 
            oplot, ((ti[i_blk_str:i_blk_end])[idx_corr])[0:*:10]-ti[0], (cal_tb[idx_corr])[0:*:10]*1d4, LINE=0, COLOR=clr[4]
        ENDIF

        ; Flat field correction and Normalize (0.001V for standard pulse cal.)
        cal_tb = cal_tb/flat_ccf[ch, 0]/0.001

      ENDELSE
        ;; # for check & debug
        ;plot, tim_fi_clip, cal_fi_clip/flat_ccf[ch, 0]/0.001
        ;oplot, ti[j_j_frm:j1j_frm], cal_tb, color='0000FF'xl

      ; error check
      cal_zero = where(cal_tb LE 0., cnt_zero)
      
      IF (cnt_zero GT 0) THEN BEGIN
        cal_tb[cal_zero] = +1.
        no_dccorr_flg[ch, [cal_zero+i_blk_str]] = 1 ; set flag
        qual_rp_table[ch, [cal_zero+i_blk_str]] = 3 ; set quality
      ENDIF

      ; Doesn't make corrections on the data outside the fitting block.
      ; But do it if the length is shorter than 1500 frames.
      ; (This avoids data gap around the Galactic plane.)
      ;
      ; - check the data before the fitting block
      idx_no_corr = where((ti[i_blk_str:i_blk_end] LT tim_fi_blk[0]) , cnt_no_corr)
      IF (cnt_no_corr GT 1500) THEN BEGIN
        cal_tb[idx_no_corr] = +1. ; set as unity
        no_dccorr_flg[ch, [idx_no_corr+i_blk_str]] = 1 ; set flag
        qual_rp_table[ch, [idx_no_corr+i_blk_str]] = 3 ; set quality
      ENDIF
      ; - check the data behind the fitting block
      idx_no_corr = where((ti[i_blk_str:i_blk_end] GT tim_fi_blk[cnt_cal_this_blk-1]), cnt_no_corr)
      IF (cnt_no_corr GT 1500) THEN BEGIN
        cal_tb[idx_no_corr] = +1. ; set as unity
        no_dccorr_flg[ch, [idx_no_corr+i_blk_str]] = 1 ; set flag
        qual_rp_table[ch, [idx_no_corr+i_blk_str]] = 3 ; set quality
      ENDIF

      ;;;;;
      ; Do responsivity correction
      ;;;;;
      corr_flux[ch, i_blk_str:i_blk_end] = flux[ch, i_blk_str:i_blk_end]/cal_tb[*]
      corr_ferr[ch, i_blk_str:i_blk_end] = ferr[ch, i_blk_str:i_blk_end]/cal_tb[*]
;      corr_flux[ch, i_blk_str:i_blk_end] = cal_tb[*] ; # for debug

      IF (cnt_cal_this_blk GE 8) THEN BEGIN
        qual_rp_table[ch, i_blk_str:i_blk_end] = 0 ; set quality
      ENDIF ELSE BEGIN
        qual_rp_table[ch, i_blk_str:i_blk_end] = 1 ; set quality
      ENDELSE

    ENDFOR  ;; End of fitting block loop

    ; Result Plot Option
    IF kplot THEN BEGIN
        XYOutS, 0.4, 0.96, tsdName, /NORM, charthick=2
;        !P.MULTI=0
    ENDIF

  ENDFOR     ;; End of channel loop

ENDIF ; End of the case 3

; Result Plot Option
IF kplot THEN BEGIN
  Device, /CLOSE
  !P.THICK=0
  !X.THICK=0
  !Y.THICK=0
  !P.MULTI=0
  if !D.NAME EQ 'PS' then Device, XS=7, YS=5, XOFF=0.75, YOFF=5, /INCHES, SCALE=1.0, LANDSCAPE=0
  Set_Plot, thisDev
  writefits, 'rms_'+tsdBodyName+'.fits', rms_ch
ENDIF

;;----------------------------------------------------------
;; Set quality indicators
;;----------------------------------------------------------
;
  ret = tsd->Set_quality_tsd('FIS_OBS','Qual_rp_data', qual_rp_data)
  ret = tsd->Set_quality_tsd('FIS_OBS','Qual_rp_param', qual_rp_param)
  ret = tsd->Set_quality_tsd('FIS_OBS','Qual_rp_table', qual_rp_table)

;;----------------------------------------------------------
;; Write TSD
;;----------------------------------------------------------
  ret = tsd->Set_value_tsd('FIS_OBS','Flux', corr_flux)
  ret = tsd->Set_value_tsd('FIS_OBS','FErr', corr_ferr)
  LMESSAGE, 7, 'Writing responsivity corrected TSD...', /CONTINUE

  ret = tsd->Set_pix_flag_tsd('FIS_OBS','No_DCcal_corr', no_dccorr_flg)
  fi_bad = where(st_bad + no_dccorr_flg GT 0, cnt)
  IF (cnt GT 0) THEN st_bad[fi_bad]=1
  ret = tsd->Set_pix_flag_tsd('FIS_OBS','Bad', st_bad)
  LMESSAGE, 7, 'Setting DCCAL_CORR, BAD flag,...', /CONTINUE

;;----------------------------------------------------------
;; Write HISTORY in TSD header
;;----------------------------------------------------------
  LMESSAGE, 7, 'Writing HISTORY', /CONTINUE
  name_ver = gb_get_name_ver(infile='gb_dc_responsivity_corr')
  IF (name_ver NE '') THEN BEGIN
    hist_message = STRCOMPRESS(' '+name_ver+' done. CCF= ')
  ENDIF ELSE BEGIN
    hist_message = 'gb_dc_responsivity_corr.pro,v ???'
  ENDELSE

  IF KEYWORD_SET(ccf) NE 1 THEN BEGIN
    hist_message = hist_message + 'NONE'
  ENDIF ELSE BEGIN
    hist_message = hist_message+ccf
  ENDELSE

  dummy = tsd_fxaddhist(tsd, hist_message, /prres)
  IF dummy NE 0 THEN $
    LMESSAGE, 3, 'Unable to write HISTORY', /CONTINUE

  Return, !SUCCEED
END
