; $Id: gb_detect_flutter.pro,v 1.14 2008/05/21 23:34:36 nikeda Exp $
;
; Copyright (c) 2007, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;   gb_detect_flutter.pro
;
; PURPOSE:
;   Detect flutter after a heavy cosmic ray hit and set flags.
;   Flags, flutter and bad are set.
;
;
; CATEGORY:
;   Function. Green Box procedure in ASTRO-F pipeline.
;
; CALLING SEQUENCE:
;   Result = gb_detect_flutter(tsd
;            [, /mask_till_cal, t_chunk = t_chunk, cut_std = cut_std, 
;               n_ref = n_ref, n_buf = n_buf,
;								neighbor_number = neighbor_number, margin = margin, 
;								t_source = t_source, 
;								n_baseline = n_baseline, flt_std = flt_std, resid_std = resid_std, 
;								/log, data_struct = data_struct])
;
;    ;; Default value for t_chunk is 5.0 [sec]
;			   cut_std is 6.0
;			   n_ref is 30 (SHOULD BE LARGER THAN 2*60/t_chunk)
;			   n_buf is 1
;			   margin = 1
;			   t_source is 150 [sec]
;			   neighbor_number is 2
;				 flt_std is 4
;				 resid_std is 7 (Changed from 8; 2008/04/23)
;        smtwidth is 3 (Changed from 2; 2008/04/23)
;				 n_baseline is 12
;
; INPUTS:
;   tsd         TSD object reference.
;		mask_till_cal		mask the flutter untill the turn-on time of cal lamp next to the detection.
;   t_chunk	binning time.
;   cut_std	threshold.
;   n_ref	to calculate threshold.
;   n_buf	to calculate threshold.
;   neighbor_number	number of minimum neighbor pixels to have also big jump
;   margin	bin margin
;   t_source	examine time range of the differentiation of the mean flux
;		flt_std
;		resid_std
;		smtwidth
; 	n_baseline
;
; OUTPUTS:
;   tsd          TSD object reference with new flags (flutter).
;
; RETURN:
;   !SUCCEED if function ended successfully.
;   !FAIL in failure.
;
; RESTRICTIONS:
;
; PROCEDURE:
;
; EXAMPLE:
;
; REFERENCE:
;
; MODIFICATION HISTORY:
;   Sep. 2007     Written by N. Ikeda.
;   Feb. 2008			Improved the avoidance of the false detection.
;									Implemented masking procedure using CCF (FLTR[LS]).
;-


function fetchtyp, tsd, str

value_str = tsd->get_value_members_tsd('FIS_OBS')
pixflag_str = tsd->get_pix_flag_members_tsd('FIS_OBS')
flag_str = tsd->get_flag_members_tsd('FIS_OBS')
status_str = tsd->get_status_members_tsd('FIS_OBS')
qual_str = tsd->get_quality_members_tsd('FIS_OBS')

dummy = where(str eq value_str, count)
if (count ne 0) then return, 'value'

dummy = where(str eq pixflag_str, count)
if (count ne 0) then return, 'pix_flag'

dummy = where(str eq flag_str, count)
if (count ne 0) then return, 'flag'

dummy = where(str eq status_str, count)
if (count ne 0) then return, 'status'

dummy = where(str eq quality_str, count)
if (count ne 0) then return, 'quality'

return, ''

end


FUNCTION stddev_fit, arr, nan = nan, cut = cut

if (not keyword_set(cut)) then cut = 0

if (keyword_set(cut)) then begin
	for i = 1, cut do begin
		index_max = where(arr eq $
			max(arr, subscript_min = index_min, /nan), count)
		if (count ne 0) then begin
			arr[index_max] = !values.d_nan
			arr[index_min] = !values.d_nan
		endif
	endfor
endif

dummy = finite(arr) & dummy2 = where(dummy, count)
if (count gt 1) then begin
	coeff = goodpoly(indgen(n_elements(arr)), arr, 1, 1e5)
	subtract = arr - (coeff[0] + coeff[1] * indgen(n_elements(arr)))
	return, stddev(subtract, /nan)
endif else return, !values.d_nan

end


FUNCTION gb_detect_flutter, tsd, $
	mask_till_cal = mask_till_cal, $
	cut_std = cut_std, $
	t_chunk = t_chunk, $
	n_ref = n_ref, $
	n_buf = n_buf, $
	n_baseline = n_baseline, $
	margin = margin, $
	t_source = t_source, $
	neighbor_number = neighbor_number, $
	flt_std = flt_std, $
	resid_std = resid_std, $
	smtwidth = smtwidth, $
	log = log, $
; +-- debug ---
	data_struct = data_struct
; --- debug ---

version = gb_get_name_ver(infile = 'gb_detect_flutter')

timescale_check = 1B

errcnt = 0

; Preparation for log.
; ------------------------------
if (keyword_set(log)) then begin

	log_dir = 'GB_DETECT_FLUTTER'
	if (not file_test(log_dir, /directory)) then begin
		file_mkdir, log_dir
	endif

	openw, lun, log_dir + '/gb_detect_flutter.log.' + $
		tsd->get_filename(), /get_lun

	printf, lun, strcompress(' '+version+' started :')+systime()
endif


lmessage, 5, strcompress(' '+version+' started :'), /continue

; Flutter detection parameters
; set default values if not specified
; -----------------------------------------
if (not keyword_set(t_chunk)) then t_chunk = 5.0
if (not keyword_set(cut_std)) then cut_std = 6.0
cut_std_p = cut_std
cut_std_m = cut_std
if (not keyword_set(n_ref)) then n_ref = 30	;t_ref = n_ref * t_chunk
if (n_ref le 2.0*60.0/t_chunk) then begin
	n_ref = 2.0*60.0/t_chunk + 2
	lmessage, 7, 'N_REF should be larger than 2*60.0/t_chunk. Using N_REF = '+$
		string(n_ref)
	if (keyword_set(log)) then printf, lun, $
		'N_REF should be larger than 2*60.0/t_chunk. Using N_REF = '+string(n_ref)
endif
if (not keyword_set(n_buf)) then n_buf = 1	;t_buf = n_buf * t_chunk
if (not keyword_set(n_baseline)) then n_baseline = 12L
if (not keyword_set(margin)) then margin = 1
if (not keyword_set(t_source)) then t_source = 150.0
if (not keyword_set(neighbor_number)) then neighbor_number = 2
if (not keyword_set(flt_std)) then flt_std = 4.0
if (not keyword_set(resid_std)) then resid_std = 7.0
if (not keyword_set(smtwidth)) then smtwidth = 3.0

; get aftime and the type of detector
; -----------------------------------------
aftime = tsd->get_value_tsd('FIS_OBS', 'aftime')

detector = strtrim(tsd->get_header_keyword_tsd('primary', 'detector'), 2)
if (detector eq 'SW') then begin
	chsize = 100
	caltyp = 'CALASON'
	daccaltyp = 'DACCALAS'
	ccf_plso = 'PLSOS'
	ccf_fltr = 'FLTRS'
endif else begin
	chsize = 75
	caltyp = 'CALALON'
	daccaltyp = 'DACCALAL'
	ccf_plso = 'PLSOL'
	ccf_fltr = 'FLTRL'
endelse

; get status and flags
; -----------------------------------------
lmessage, 7, 'Getting Status and Flags..', /continue
if (keyword_set(log)) then printf, lun, 'Getting Status and Flags..'

typname = ['FLUX', $
;	'SHTOP', $
	'DEAD', 'SATURATE', $
	'RESET', 'RSTANOM', 'NO_DIFF', 'NO_PERI_CORR', caltyp, $
	'NO_RP_CORR', $
	'GPGL_TYPE1', 'GPGL_TYPE2', 'GPGL_TYPE3', $
	'GPGL_TYPE4', 'GPGL_TAIL', $
	'MTGL_TYPE1', 'MTGL_TYPE2', 'MTGL_TYPE3', $
	'MTGL_TYPE4', 'MTGL_TAIL', 'PACKETID']

;msg_level_bak = !MSG_LEVEL
;!MSG_LEVEL = 1
plso = ccf_read_ccf(ccf_plso, tsd, error = error)
if (error eq !FAIL) then begin
	lmessage, 1, 'CCF ' + ccf_plso + ' cannot be found.', /cont
;	!MSG_LEVEL = msg_level_bak
	return, !FAIL
endif
;!MSG_LEVEL = msg_level_bak

for typ = 0, n_elements(typname)-1 do begin
	cmd = 'value = tsd->get_'+fetchtyp(tsd, typname[typ])+$
		'_tsd("FIS_OBS", "'+typname[typ]+'")'

	result = execute(cmd)
	if (not result) then begin
		lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
		errcnt ++
		continue
	endif

	if fetchtyp(tsd, typname[typ]) eq 'status' then begin

		cmd = typname[typ] + ' = replicate(1.0, chsize) # value'

		result = execute(cmd)
		if (not result) then begin
			lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
			errcnt ++
			continue
		endif

	endif else begin

		cmd = typname[typ]+' = value'

		result = execute(cmd)
		if (not result) then begin
			lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
			errcnt ++
			continue
		endif

	endelse
endfor

;; for SW, cal lamp turn on at 3 frame after calason flag set
;; -----------------------------------------
;if keyword_set(calason) then begin
;	time_delay_calason = 3
;	if (detector eq 'SW') then $
;		calason = shift(calason, 0, time_delay_calason)
;		for i = 0, time_delay_calason-1 do $
;			replicate_inplace, calason, 0, 1, [0, i]
;endif


; flag out the data
; -----------------------------------------
lmessage, 7, 'Flag out the data..', /continue
if (keyword_set(log)) then printf, lun, 'Flag out the data..'

flux_org = flux
flux_cal = flux
no_bd = bytarr(chsize, n_elements(aftime))

for i = 0, n_elements(typname)-1 do begin

	case typname[i] of 
		'SHTOP' : begin
			cmd = 'index = where(not '+typname[i]+', count)'
			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
				errcnt ++
			endif

			; flag out the data
			if (count ne 0) then begin
				flux[index] = !values.d_nan
				flux_org[index] = !values.d_nan
				no_bd[index] = no_bd[index] or 1
			endif

			; reset the variable
			cmd = typname[i]+' = 0'
			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd, /continue
				errcnt ++
			endif
		end

		'FLUX' : begin
;			flux_unflagged = flux
		end

		'PACKETID' : begin
; ignoring FTS mode
; ------------------------------
			cmd = 'index = where(' + typname[i] + ' eq 124, count)'
			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
				errcnt ++
			endif

			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
				errcnt ++
			endif

			; flag out the data
			if (count ne 0) then begin
				flux[index] = !values.d_nan
				flux_org[index] = !values.d_nan
				flux_cal[index] = !values.d_nan
				no_bd[index] = no_bd[index] or 1
			endif
		end

		'RSTANOM' : goto, RESET_label
		'NO_DIFF' : goto, RESET_label
		'RESET' : begin
RESET_label:
			cmd = 'index = where(' + typname[i] + ', count)'
			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
				errcnt ++
			endif

			; flag out the data
			if (count ne 0) then begin
				flux[index] = !values.d_nan
				flux_org[index] = !values.d_nan
				flux_cal[index] = !values.d_nan
				no_bd[index] = no_bd[index] or 1
			endif

			; reset the variable
			cmd = typname[i]+' = 0'
			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd, /continue
				errcnt ++
			endif
		end

		'CALALON' : goto, CALASON_label
		'CALASON' : begin
CALASON_label:
			result = execute('cal = '+typname[i])
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
				errcnt ++
			endif

			; reset the variable
			cmd = typname[i]+' = 0'
			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd, /continue
				errcnt ++
			endif

  		; Shift the turn-on and -off timings of cal. lamp
  		; follwoing the calibration table PLSO[LS].
  		; -----------------------------------------
  		lmessage, 9, 'Offsetting turn-on and -off timings of cal lamp..', /cont
  		index = where(cal[0, *] eq 1, count)
  		if (count ne 0) then begin
    		intval, index, s_pos, e_pos

    		; calculate overshoot using pulse cal flag.
    		pulse_cal_pos = min(where( (e_pos - s_pos +1) le 10, count))
    		if (count eq 0) then begin
      		lmessage, 9, 'No pulse cal lamp. Continue..', /cont
					goto, shiftlabel
    		endif
    		overshoot = plso.off - (e_pos - s_pos + 1)[pulse_cal_pos]

    		; shift flags
    		for k = 0, n_elements(s_pos)-1 do begin
      		cal[*, e_pos[k]:( (e_pos[k] + overshoot)<(n_elements(cal[0, *])-1) )] = 1B
      		if (plso.on ne 0) then $
        		cal[*, s_pos[k]:( (s_pos[k] + plso.on-1)<(n_elements(cal[0, *])-1) )] = 0B
    		endfor

  		endif
shiftlabel:

;			cmd = 'index = where('+typname[i]+', count)'
;			result = execute(cmd)
;			if (not result) then begin
;				if (keyword_set(log)) then $
;					printf, lun, 'Failed executing: "'+cmd+'"'
;				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
;				errcnt ++
;			endif
			index = where(cal, count)

			; flag out the data
			if (count ne 0) then begin
				flux[index] = !values.d_nan
; cal flags are ignored for flux_cal
				flux_org[index] = !values.d_nan
				no_bd[index] = no_bd[index] or 1
			endif

		end

		else : begin
			cmd = 'index = where('+typname[i]+', count)'
			result = execute(cmd)
			if (not result) then begin
				if (keyword_set(log)) then $
					printf, lun, 'Failed executing: "'+cmd+'"'
				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
				errcnt ++
			endif

			; flag out the data
			if (count ne 0) then begin
				flux[index] = !values.d_nan
; others are ignored for flux_org and flux_cal
;				flux_org[index] = !values.d_nan
				no_bd[index] = no_bd[index] or 1
			endif

			; reset the variable
;			cmd = typname[i]+' = 0'
;			result = execute(cmd)
;			if (not result) then begin
;				if (keyword_set(log)) then $
;					printf, lun, 'Failed executing: "'+cmd+'"'
;				lmessage, 3, 'Failed executing: "'+cmd+'"', /continue
;				errcnt ++
;			endif

		end

	endcase

endfor

;----------------------------------------
; Flutter detection
; 1) calculate time bin size (t_chunk)
; 2) calculate the mean and standard deviation of the flux within each time bin
; 3) derive the differentiation fo the mean and standard derivation
; 4) searching big jumps as flutter candidates
; 5) examine whether the neigbor pixels also have big jumps or not
; 6) examine the timescales of the excitation and relaxation of the big jumps
; 7) removing redundant flags and ignoring the transition to CDS mode
; 8) restore the flag data
;----------------------------------------
lmessage, 7, 'Creating data for detection..', /continue
if (keyword_set(log)) then printf, lun, 'Creating data for detection..'

; 1) calculate time bin size (t_chunk)
;----------------------------------------
dt = aftime[1] - aftime[0]
ind_int = round(t_chunk/dt)

size = size(flux, /dimension)
ind_length = size[1]
n_bin = ind_length/ind_int*2


; If the data time length is less than the reference time length,
; clear flags and exit.
; --------------------
if (n_bin lt n_ref + n_buf -1) then begin
	tmpstr = 'No enough time length for flutter detection.'
	lmessage, 7, tmpstr, /cont
	if (keyword_set(log)) then $
		printf, lun, tmpstr
	errcnt ++

;	flutter = tsd->get_pix_flag_tsd('FIS_OBS', 'flutter')
;	flutter[*] = 0B
;	stat = tsd->set_pix_flag_tsd('FIS_OBS', 'flutter', flutter)
;	if (stat eq !FAIL) then begin
;		tmpstr = 'Failed to save pix flag (flutter) to TSD'
;		lmessage, 1, tmpstr, /continue
;		errcnt ++
;		if (keyword_set(log)) then $
;			printf, lun, tmpstr
;	endif

	qual_flutter = tsd->get_quality_tsd('FIS_OBS', 'qual_flutter')
	qual_flutter[*] = 1B
	stat = tsd->set_quality_tsd('FIS_OBS', 'qual_flutter', qual_flutter)
	if (stat eq !FAIL) then begin
		tmpstr = 'Failed to save quality flag (qual_flutter) to TSD'
		lmessage, 1, tmpstr, /continue
		errcnt ++
		if (keyword_set(log)) then $
			printf, lun, tmpstr
	endif

	tmpstr = strcompress(' '+version+' done.')
	lmessage, 5, tmpstr, /continue
	dummy = tsd_fxaddhist(tsd, tmpstr, /prres)
	if (dummy ne 0) then begin
		tmpstr = 'Unable to write history to TSD'
		lmessage, 3, tmpstr, /continue
		errcnt ++
		if (keyword_set(log)) then $
			printf, lun, tmpstr
	endif

	if (keyword_set(log)) then begin
		if (errcnt eq 0) then begin
			printf, lun, $
				'Seems to be finished successfully.. :' + systime()
		endif else begin
			printf, lun, $
				'Seems to be finished with some ERRORS :' + systime()
		endelse
		close, lun
		free_lun, lun
	endif

	return, !SUCCEED

endif


ind_start = floor(dindgen(n_bin)*ind_int/2.)
ind_end = floor(ind_start+ind_int-1)

index = where(ind_end gt ind_length-1, count)
if (count) then ind_end[index] = ind_length-1

; 2) calculate the mean and standard deviation of the flux within each time bin
; and derive the differentiation of the mean and standard derivation
;----------------------------------------
mn = fltarr(chsize, n_elements(ind_start))
mn_org = fltarr(chsize, n_elements(ind_start))
sd = fltarr(chsize, n_elements(ind_start))
d_mn = fltarr(chsize, n_elements(ind_start))
d_mn_org = fltarr(chsize, n_elements(ind_start))
d_sd = fltarr(chsize, n_elements(ind_start))

for ch = 0, chsize-1 do begin
	for i = 0, n_elements(ind_start)-1 do begin
		fragment = reform(flux[ch, ind_start[i]:ind_end[i]])
		dummy = finite(fragment) & dummy2 = where(dummy, count)
		if (count ge 2) then begin
			mn[ch, i] = mean(fragment, /nan);, cut = 2)
;			mn[ch, i] = mean_fit(fragment, /nan)
;			sd[ch, i] = stddev(fragment, /nan)
			sd[ch, i] = stddev_fit(fragment, /nan);, cut = 2)
		endif else begin
			sd[ch, i] = !values.f_nan
			mn[ch, i] = !values.f_nan
		endelse

		fragment_org = reform(flux_org[ch, ind_start[i]:ind_end[i]])
		dummy = finite(fragment_org) & dummy2 = where(dummy, count)
		if (count ge 2) then begin
			mn_org[ch, i] = mean(fragment, /nan)
		endif else begin
			mn_org[ch, i] = !values.f_nan
		endelse

	endfor
	tmp = deriv(mn[ch, *]) & d_mn[ch, *] = tmp
	tmp = deriv(mn_org[ch, *]) & d_mn_org[ch, *] = tmp
	tmp = deriv(sd[ch, *]) & d_sd[ch, *] = tmp
endfor

; 3) derive the mean and stddev of the differentiation 
;----------------------------------------
mn_d_mn = fltarr(chsize, n_elements(ind_start))
sd_d_mn = fltarr(chsize, n_elements(ind_start))
mn_d_sd = fltarr(chsize, n_elements(ind_start))
sd_d_sd = fltarr(chsize, n_elements(ind_start))
for ch = 0, chsize - 1 do begin
	for i = 0, n_elements(ind_start)-n_ref-1 do begin
		fragment_mn = reform(d_mn[ch, i:i+n_ref-1])
		fragment_sd = reform(d_sd[ch, i:i+n_ref-1])

		dummy = finite(fragment_mn) & dummy2 = where(dummy, count)
		if (count ge 2) then begin
			mn_d_mn[ch, i] = mean(fragment_mn, /nan)
;			mn_d_mn[ch, i] = mean_fit(fragment_mn, /nan)
;			sd_d_mn[ch, i] = stddev(fragment_mn, /nan)
			sd_d_mn[ch, i] = stddev_fit(fragment_mn, /nan)
		endif else begin
			mn_d_mn[ch, i] = !values.f_nan
			sd_d_mn[ch, i] = !values.f_nan
		endelse

		dummy = finite(fragment_sd) & dummy2 = where(dummy, count)
		if (count ge 2) then begin
			mn_d_sd[ch, i] = mean(fragment_sd, /nan)
;			sd_d_sd[ch, i] = stddev(fragment_sd, /nan)
			sd_d_sd[ch, i] = stddev_fit(fragment_sd, /nan)
		endif else begin
			mn_d_sd[ch, i] = !values.f_nan
			sd_d_sd[ch, i] = !values.f_nan
		endelse

	endfor
endfor

; 4)
; First, search a big jump in the differentiation of 
; the mean and standard deviation of flux as a candidate of flutter.
; If found, set bdflag.
; Set no_bdflag for sample points not to be able to detect flutter.
;----------------------------------------
lmessage, 7, '== Now into the jump-detection loop.. ==', /continue
if (keyword_set(log)) then $
	printf, lun, '== Now into the jump-detection loop.. =='

bdflag = bytarr(chsize, n_elements(ind_start))
no_bdflag = bytarr(chsize, n_elements(ind_start))

no_bdflag[*, 0:n_ref+n_buf-1] = 1

; +-- debug ---
thmn = fltarr(chsize, n_elements(ind_start))
thsd = fltarr(chsize, n_elements(ind_start))
thmn_m = fltarr(chsize, n_elements(ind_start))
thsd_m = fltarr(chsize, n_elements(ind_start))
; --- debug ---

; Channel Loop
; -------------------------
for ch = 0, chsize-1 do begin

	bjc = 0

	for j = n_ref+n_buf, n_elements(ind_start)-1 do begin

; Set threshold of the following jump detection
; -------------------------
		th_d_mn = mn_d_mn[ch, j-n_buf-n_ref] + $
			cut_std_p*sd_d_mn[ch, j-n_buf-n_ref] 
		th_d_sd = mn_d_sd[ch, j-n_buf-n_ref] + $
			cut_std_p*sd_d_sd[ch, j-n_buf-n_ref]

		th_d_mn_m = mn_d_mn[ch, j-n_buf-n_ref] - $
			cut_std_p*sd_d_mn[ch, j-n_buf-n_ref] 
		th_d_sd_m = mn_d_sd[ch, j-n_buf-n_ref] - $
			cut_std_p*sd_d_sd[ch, j-n_buf-n_ref]

; +-- debug ---
thmn[ch, j] = th_d_mn
thsd[ch, j] = th_d_sd
thmn_m[ch, j] = th_d_mn_m
thsd_m[ch, j] = th_d_sd_m
; --- debug ---

; Condition for searching big jump
; -------------------------
;		if (d_mn[ch, j] ge th_d_mn and d_sd[ch, j] ge th_d_sd) then begin
		if ( (d_mn[ch, j] ge th_d_mn and d_sd[ch, j] ge th_d_sd) or $
			(d_mn[ch, j] le th_d_mn_m and d_sd[ch, j] le th_d_sd_m) ) then begin

			bdflag[ch, j] = 1
			bjc ++

;			mn_d_mn[ch, j] = mn_d_mn[ch, j-n_buf-n_ref]
;			sd_d_mn[ch, j] = sd_d_mn[ch, j-n_buf-n_ref]
;			mn_d_sd[ch, j] = mn_d_sd[ch, j-n_buf-n_ref]
;			sd_d_sd[ch, j] = sd_d_sd[ch, j-n_buf-n_ref]

			tmpstr = string(bjc, ch+1, j, $
				format = '("Big jump # ", i0, " in ch ", i0, " is detected (", i0, ")")')
			lmessage, 9, tmpstr, /continue
			if (keyword_set(log)) then $
				printf, lun, tmpstr

		endif 

	endfor

;Just after the flagged data, flutter cannot be detected correctly.
; -------------------------
	sign = finite(d_mn[ch, *])
	index = where(sign eq 0, count)
	if (count ne 0) then begin
		intval, index, s_pos, e_pos
		for k = 0, n_elements(e_pos)-1 do begin
			if ( (e_pos[k] - s_pos[k]) ge n_ref ) then begin
				bdflag[ch, $
					(e_pos[k]+1)<(n_elements(bdflag[ch, *])-1):$
					(e_pos[k]+4)<(n_elements(bdflag[ch, *])-1)] = 0
				no_bdflag[ch, $
					(e_pos[k]+1)<(n_elements(no_bdflag[ch, *])-1):$
					(e_pos[k]+4)<(n_elements(no_bdflag[ch, *])-1)] = 1
			endif
		endfor
	endif

endfor

; 5)
; Next, examine whether neighbor pixels also shows big jumps or not.
; If yes, it may not be a flutter candidate. 
; ------------------------------
lmessage, 7, '== Now into the cross-pixel checking loop.. ==', /continue
if (keyword_set(log)) then $
	printf, lun, '== Now into the cross-pixel checking loop.. =='
confirm = bdflag 

for ch = 0, chsize - 1 do begin
	pos = where(bdflag[ch, *], count)

	if (count ne 0) then begin
		neighbor = neighbor_pixels(ch+1, detector, searching_width = 5, $
; Ignoring pixels next to the examined one
			ignore_range = 1)

		if (n_elements(neighbor) lt neighbor_number) then $
			neighbor_number_apply = n_elements(neighbor) $
		else neighbor_number_apply = neighbor_number

		for i = 0, n_elements(pos)-1 do begin
			n_count = 0

			tmpstr = string(i+1, ch+1, pos[i], neighbor, neighbor_number_apply, $
				format = '("Big jump # ", i0, " in ch ", i0, " (", i0, ") : ' + $
					'examining other pixels .. (", ' + $
					strtrim(string(n_elements(neighbor)),2) + $
					'(i0, x),"), threshold > ", i0 )')
			lmessage, 9, tmpstr, /continue
			lmessage, 9, string("-> found in ", format = '(a0, x, $)')

			if (keyword_set(log)) then begin
				printf, lun, tmpstr
				printf, lun, "-> found in ", format = '(a0, x, $)'
			endif

			for j = 0, n_elements(neighbor)-1 do begin
				if total( bdflag[neighbor[j] - 1, $
					( (pos[i]-margin) > 0 ): $
					( (pos[i]+margin) < (n_elements(ind_start)-1) ) ] ) ne 0 then begin
					n_count ++

					if (keyword_set(log)) then $
						printf, lun, neighbor[j], format = '(i0, ", ", $)' 

				endif
			endfor

			if (keyword_set(log)) then $
				printf, lun, ''

			if n_count ge neighbor_number_apply then begin
				confirm[ch, pos[i]] = 0

				tmpstr = string(i+1, ch+1, pos[i], $
					format = '(" -- Big jump # ", i0, " in ch ", i0, ": ' + $
						'jumps can be seen in the neighbor pixels. ' + $
						'Possibly an astronomical source (", i0, ")")')
				lmessage, 9, tmpstr, /continue
				if (keyword_set(log)) then $
						printf, lun, tmpstr

			endif else begin

				tmpstr = string(i+1, ch+1, pos[i], $
					format = '(" -- Big jump # ", i0, " in ch ", i0, ": ' + $
						'jumps cannot be found in the neighbor pixels. ' + $
						'Proceeding.. (", i0, ")")')
				lmessage, 9, tmpstr, /continue
				if (keyword_set(log)) then $
					printf, lun, tmpstr

			endelse

		endfor
	endif
endfor
bdflag = bdflag and confirm

; Examine whether heavy glitches exist in the time bin.
; If heavy glitches exist, skip the following timescale check.
; -------------------------
lmessage, 7, '== Now into the heavy glitch checking loop.. ==', /continue
if (keyword_set(log)) then $
	printf, lun, '== Now into the heavy glitch checking loop.. =='
glitch_conf = bdflag 
size = size(bdflag, /dimension)

glitch_flag = $
	gpgl_type1 or gpgl_type2 or gpgl_type3 or gpgl_type4 or gpgl_tail or $
	mtgl_type1 or mtgl_type2 or mtgl_type3 or mtgl_type4 or mtgl_tail

for ch = 0, chsize-1 do begin

	pos = where(bdflag[ch, *], count)
	if (count ne 0) then begin

		for i = 0, n_elements(pos)-1 do begin

			af_start = aftime[ind_start[pos[i]]]
			af_end = aftime[ind_end[pos[i]]]

			tmpstr = string(i+1, ch+1, pos[i], af_start, af_end, $
				format = '("Big jump # ", i0, " in ch ", i0, " (", i0, ") : ' + $
				'aftime = ", f0, " - ", f0, " : ")' )
			lmessage, 9, tmpstr, /continue
			if (keyword_set(log)) then $
				printf, lun, tmpstr

			glitch_index = where(glitch_flag[ch, ind_start[pos[i]]:ind_end[pos[i]]], count)
			if (count ne 0) then begin

				glitch_flux = flux_org[ch, ind_start[pos[i]]:ind_end[pos[i]]]
				glitch_flux = glitch_flux[glitch_index]

				mean_baseline = mean( $
					mn[ch, ((pos[i] - n_buf - n_ref - 1)>0):((pos[i] - n_buf - 1)>0)], /nan)
				stddev_baseline = stddev( $
					mn[ch, ((pos[i] - n_buf - n_ref - 1)>0):((pos[i] - n_buf - 1)>0)], /nan)
				th_hg = 200.0D * stddev_baseline + mean_baseline

				heavy_glitch_index = where(glitch_flux ge th_hg, count)
				if (count ne 0) then begin

					tmpstr = string(count,  th_hg, $
						format = '(" -- ", i0, " heavy glitches > ", f0, " are found. Skipping the following timescale check..")')
					lmessage, 9, tmpstr, /continue
					if (keyword_set(log)) then $
						printf, lun, tmpstr

					glitch_conf[ch, pos[i]] = 1B

				endif else begin

					tmpstr = string(th_hg, $
						format = '(" -- no heavy glitches > ", f0, " are found. Continue..")')
					lmessage, 9, tmpstr, /continue
					if (keyword_set(log)) then $
						printf, lun, tmpstr

					glitch_conf[ch, pos[i]] = 0B

				endelse

;plot, flux_org[ch, ind_start[pos[i]]:ind_end[pos[i]]], psym = 10
;oplot, flux[ch, ind_start[pos[i]]:ind_end[pos[i]]], psym = 10, color = '00ff00'x
;oplot, glitch_flag[ch, ind_start[pos[i]]:ind_end[pos[i]]], psym = 10, $
;	color = '0000ff'x
;plots, !x.crange, [th_hg, th_hg]
;read, gomi

			endif else begin

				tmpstr = " -- no glitches are found. Continue.."
				lmessage, 9, tmpstr, /continue
				if (keyword_set(log)) then $
					printf, lun, tmpstr

				glitch_conf[ch, pos[i]] = 0B

			endelse

		endfor
		
	endif

endfor


; 6)
; Examine whether the jump detection is flutter or 
; just a astronomical source:
; Typical flutter shows that
; the flux, as a function of time, increases very rapidly
; but decreases relatively slower.
; On the other hand, many astronomical sources show that
; the increase and decrease of the flux are comparable to each other.
; -------------------------
if (keyword_set(timescale_check)) then begin

lmessage, 7, '== Now into the timescale checking loop.. ==', /continue
if (keyword_set(log)) then $
	printf, lun, '== Now into the timescale checking loop.. =='
for ch = 0, chsize-1 do begin

	pos = where(bdflag[ch, *], count)

	if (count ne 0) then begin

		for k = 0, n_elements(pos)-1 do begin

			if (glitch_conf[ch, pos[k]] eq 1B) then begin

				tmpstr = string(k+1, ch+1, pos[k], $
					format = '(" -- Big jump # ", i0, " in ch ", i0, " ' + $
						'is skipped. Proceeding.. (", i0, ")")')
				lmessage, 9, tmpstr, /continue
				if (keyword_set(log)) then printf, lun, tmpstr

				goto, timescale_skip_label

			endif

			if (pos[k]+t_source/t_chunk ge (size(d_mn_org, /dimension))[1]) then begin
; Ignore big jumps located at the edge of the data
; -------------------------

				tmpstr = string(k+1, ch+1, pos[k], $
					format = '(" -- Big jump # ", i0, " in ch ", i0, " ' + $
						'is located at the edge of data. Do nothing. (", i0, ")")')
				lmessage, 9, tmpstr, /continue
				if (keyword_set(log)) then printf, lun, tmpstr

				bdflag[ch, pos[k]] = 0

			endif else begin

				index = min( where( $
					d_mn_org[ch, $
						pos[k]:$
						(pos[k]+t_source/t_chunk)<((size(d_mn_org, /dimension))[1]-1)] $
						le 0, count) )

				if (count eq 0) then begin

					tmpstr = string(k+1, ch+1, pos[k], $
						format = '(" -- Failed to derive first_minus : ' + $
							'# ", i0, " in ch ", i0, " (", i0, ")")')
					lmessage, 9, tmpstr, /continue
					if (keyword_set(log)) then printf, lun, tmpstr

					index = 0
				endif
				first_minus = pos[k] + index

				index = min( where( $
					d_mn_org[ch, $
						first_minus:$
						(first_minus+t_source/t_chunk)<$
						((size(d_mn_org, /dimension))[1]-1)] $
					ge 0, count) )

				if (count eq 0) then begin
					tmpstr = string(k+1, ch+1, pos[k], $
						format = '(" -- Failed to derive first_plus : ' + $
							'# ", i0, " in ch ", i0, " (", i0, ")")')
					lmessage, 9, tmpstr, /continue
					if (keyword_set(log)) then printf, lun, tmpstr

					index = 0
				endif
				first_plus = first_minus + index

				integ = total( d_mn_org[ch, pos[k]:first_plus], /nan)
				th_plus = 4d*stddev_fit($
					mn_org[ch, (pos[k]-n_buf-n_ref):(pos[k]-n_buf)], /nan)
				th_minus = -1d*30d*stddev_fit($
					mn_org[ch, (pos[k]-n_buf-n_ref):(pos[k]-n_buf)], /nan)

				if (integ lt th_plus) and (integ gt th_minus) then begin

					tmpstr = string(k+1, ch+1, pos[k], first_minus, first_plus, $
						integ, th_plus, th_minus, $
						format = '(" -- Big jump # ", i0, " in ch ", i0, ' + $
							'" may be an astronomical source: ' + $
							'timescales are comparable (", i0, ") ' + $
							'", i0, " - ", i0, " -> ", f0, ' + $
							'" within ", f0, " - ", f0, ")")')
					lmessage, 9, tmpstr, /continue
					if (keyword_set(log)) then $
						printf, lun, tmpstr

					bdflag[ch, pos[k]] = 0

				endif else begin

					tmpstr = string(k+1, ch+1, pos[k], first_minus, first_plus, $
						integ, th_plus, th_minus, $
						format = '(" -- Big jump # ", i0, " in ch ", i0, ' + $
							'" : may be a flutter. Proceeding.. (", i0, ") ' + $
							'", i0, " - ", i0, " -> ", f0, ' + $
							'" out of the range ", f0, " - ", f0, ")")')
					lmessage, 9, tmpstr, /continue
					if (keyword_set(log)) then $
						printf, lun, tmpstr

				endelse

			endelse

timescale_skip_label:

		endfor
	endif

endfor

endif


; 7)
; remove redundant flags and ignore the transition to CDS 
; ------------------------------

; ignoring the transition to CDS
; ------------------------------
;index = where(packetid eq 65, count)	;CDS mode
;if (count ne 0) then begin
;	tmpstr = '== CDS mode is found.. =='
;	lmessage, 7, tmpstr, /continue
;	if (keyword_set(log)) then $
;		printf, lun, tmpstr
;
;	intval, index, to_cds, dummy
;	for i = 0, n_elements(to_cds)-1 do begin
;		distance = to_cds[i] - ind_start
;		cds_pos = where(distance ge 0 and distance lt ind_int/2.)
;		cds_pos = cds_pos[0]
;		cds_mask = cds_pos + lindgen(7)-3
;
;		tmpstr = string(cds_pos, min(cds_mask), max(cds_mask), $
;			format = '(" -- Transition to CDS is found at ",' + $
;				'(i0, x), ". ignoring..", i0, " - ", i0)')
;		lmessage, 9, tmpstr, /continue
;		if (keyword_set(log)) then $
;			printf, lun, tmpstr
;
;		bdflag[*, cds_mask] = 0
;		no_bdflag[*, cds_mask] = 1
;	endfor
;endif

; ignoring the transtion to other sampling mode
; ------------------------------
transition = packetid - shift(packetid, 1)
transition[0] = 0B
trans_index = where(transition gt 0, count)
if (count ne 0) then begin
	tmpstr = '== Sampling mode transitions are found.. =='
	lmessage, 7, tmpstr, /continue
	if (keyword_set(log)) then $
		printf, lun, tmpstr

	for i = 0, n_elements(trans_index)-1 do begin
		distance = trans_index[i] - ind_start
		trans_pos = where(distance ge 0 and distance lt ind_int/2.)
		trans_pos = trans_pos[0]
		trans_mask = trans_pos + lindgen(7)-3

		tmpstr = string(trans_pos, min(trans_mask), max(trans_mask), $
			format = '(" -- Sampling mode transition is found at ",' + $
				'(i0, x), ". ignoring..", i0, " - ", i0)')
		lmessage, 9, tmpstr, /continue
		if (keyword_set(log)) then $
			printf, lun, tmpstr

		bdflag[*, trans_mask] = 0
		no_bdflag[*, trans_mask] = 1
	endfor
endif


lmessage, 7, ' -- end of the detection using binned data --', /continue
if (keyword_set(log)) then $
	printf, lun, ' -- end of the detection using binned data --'

; Remove redundant flags and store "FLUTTER" pixel flag (actually a variable "bd")
; ------------------------------
bd = bytarr(chsize, n_elements(aftime))
for ch = 0, chsize-1 do begin
	index = where(bdflag[ch, *], count)

	if (count ne 0) then begin
		intval, index, start_pos, end_pos

		for k = 0, n_elements(start_pos)-1 do begin
			if not ( start_pos[k] eq end_pos[k] ) then $
				bdflag[ch, start_pos[k]:end_pos[k]-1] = 0

; find the position of flutter within the time bin
; ------------------------------
			bin_start_prev = ind_start[end_pos[k]-1]
			bin_end_prev = ind_end[end_pos[k]-1]
			avrg = mean(flux[ch, bin_start_prev:bin_end_prev], /nan)
			stdev = stddev(flux[ch, bin_start_prev:bin_end_prev], /nan)

			bin_start = ind_start[end_pos[k]]
			bin_end = ind_end[end_pos[k]]
			pos_bin = where(flux[ch, bin_start:bin_end] ge $
				avrg + flt_std * stdev, count)

			if (count ne 0) then flutter_pos = bin_start+ min(pos_bin) $
			else flutter_pos = bin_end

;			wset, 0
;			plot, flux[ch, bin_start:bin_end], psym=10, title = string(ch+1)
;			oplot, flux[ch, bin_start_prev:bin_end_prev], psym=10, color = '00ff00'x
;			plots, !x.crange, [avrg, avrg], color = '0000ff'x
;			plots, !x.crange, [avrg + flt_std * stdev, avrg + flt_std * stdev], $
;				color = '0000ff'x
;			plots, [flutter_pos-bin_start, flutter_pos-bin_start], !y.crange
;read, gomi

			tmpstr = string(k+1, ch+1, $
				end_pos[k], aftime[flutter_pos], $
				format = '("Flutter candidate # ", i0, ' + $
					'" is found in ch ", i0, " (", i0, ") aftime = ", f0)')
			lmessage, 7, tmpstr, /continue

			if (keyword_set(log)) then $
				printf, lun, tmpstr

; Store "FLUTTER" pixflag 
; -------------------------
			bd[ch, flutter_pos] = 1B

		endfor
	endif
endfor


; Checking misidentification of astronomical sources as flutters
; with the ORIGINAL data, i.e., not binned along aftime.
; ------------------------------
lmessage, 7, ' -- Checking misidentification of sources.. --', /continue
if (keyword_set(log)) then $
	printf, lun, ' -- Checking misidentification of sources.. --'

sample_int = aftime[1] - aftime[0]
confirm = bd

daccal = tsd->get_value_tsd('FIS_OBS', daccaltyp)

for ch = 0, chsize-1 do begin

	fltr_index = where(bd[ch, *], count)
	if (count eq 0) then continue

	for i = 0, n_elements(fltr_index)-1 do begin

		flutter_pos = fltr_index[i]

		str = string(i+1, ch+1, aftime[flutter_pos], $
			format = '("Examining flutter candidate # ", i0, ' + $
				'" in ch ", i0, " (aftime = ", f0, ")..")')
		lmessage, 9, str, /cont
		if (keyword_set(log)) then $
			printf, lun, str

		;get the last lamp before the detection
		;--------------------
		cal_index = where(cal[ch, 0:flutter_pos-1], count)
		if (count eq 0) then begin
			str= ' - Cannot found lamp before the detection (' + $
				string(ch, flutter_pos, format = '(i0, ",", i0)') + '). Continue..'
			lmessage, 9, str, /cont
			if (keyword_set(log)) then $
				printf, lun, str
			continue
		endif
		intval, cal_index, s_cal, e_cal

		; determine the turn-on time of the pulse cal.
		; ------------------------------
		index = where(e_cal - s_cal lt 15, count)
		if (count ne 0) then begin

			plst = mean(e_cal[index] - s_cal[index] + 1, /nan)

		endif else begin

			str= ' - Cannot found lamp before the detection (' + $
				string(ch, flutter_pos, format = '(i0, ",", i0)') + $
				'). Using 10 samples as the turn-on time of the pulse cal..'
			lmessage, 9, str, /cont
			if (keyword_set(log)) then $
				printf, lun, str

			plst = 10L
		endelse

		; Get DAC value for cal lamps
		; ------------------------------
		dacval = daccal[ s_cal + plst/2 ]
		turn_on_idx = where(dacval ne 0, count)
		if (count eq 0) then begin
			str= ' - No turn-on cal lamps (' + $
				string(ch, flutter_pos, format = '(i0, ",", i0)') + '). Continue..'
			lmessage, 9, str, /cont
			if (keyword_set(log)) then $
				printf, lun, str
			continue
		endif
		turn_on_idx = max(turn_on_idx)

;print, turn_on_idx, dacval[turn_on_idx]

;		s_cal = s_cal[n_elements(s_cal)-1]
;		e_cal = e_cal[n_elements(e_cal)-1]
		s_cal = s_cal[turn_on_idx]
		e_cal = e_cal[turn_on_idx]
		dacfactor = dacval[turn_on_idx]/96d		;Normalize by the pulse cal DAC value.

		;get the scan rate along the in-scan direction at the time of the detection
		;--------------------
		body_rate = tsd->get_value_tsd('AOCU', 'body_rate')
		scan_rate = body_rate[1, flutter_pos]

		s_pos1 = (flutter_pos - round(50.0/sample_int))>0
		e_pos1 = (flutter_pos + round(50.0/sample_int))<(n_elements(aftime)-1)

		s_pos2 = (flutter_pos - 50L)>0
		e_pos2 = (flutter_pos + 50L)<(n_elements(aftime)-1)

		neighbor = neighbor_pixels(ch+1, detector, search = 2)
		if (neighbor[0] eq -1) then begin
			str= ' - No neighbor pixels. Continue..'
			lmessage, 9, str, /cont
			if (keyword_set(log)) then $
				printf, lun, str
			continue
		endif else begin
			if (keyword_set(log)) then begin
				printf, lun, format = '(" - Neighbor pixels : ", $)'
				for j = 0, n_elements(neighbor)-1 do begin
					printf, lun, neighbor[j], format = '(i0, " ", $)'
				endfor
				printf, lun, ''
			endif
		endelse

		check = 0B
		for j = 0, n_elements(neighbor)-1 do begin

			; ignore the neighbors which have other flutter flags
			; within 50 secs around the detection
			; --------------------
			dummy = where(bd[neighbor[j]-1, s_pos1:e_pos1] eq 1B, count)
			if (count ne 0) then begin
				str= ' - Neighbor pixel (ch ' + strtrim(string(neighbor[j]), 2) + $
					') has flutter flags. Continue..'
				lmessage, 9, str, /cont
				if (keyword_set(log)) then $
					printf, lun, str
				continue
			endif

;			cal_target = total(flux_cal[ch, s_cal:e_cal])
			cal_target = total(flux_cal[ch, s_cal:s_cal + plst-1], /nan)/dacfactor
;			cal_comp = total(flux_cal[neighbor[j]-1, s_cal:e_cal])
			cal_comp = total(flux_cal[neighbor[j]-1, s_cal:s_cal + plst-1], /nan)/dacfactor

			if (ch+1 gt neighbor[j]) then sign = -1.0D else sign = 1.0D
			sample_sep = pixel_separation_inscan(ch+1, neighbor[j], detector) / $
				scan_rate / sample_int * sign

;			smtwidth = 2

			ofst = shift(smooth(flux[neighbor[j]-1, *], smtwidth, /nan), round(sample_sep))
			resid = smooth(flux[ch, *], smtwidth, /nan)/cal_target  - ofst/cal_comp

			resid_sigma = stddev_fit( $
				resid[ $
					s_pos2:(s_pos2 + 45L)], /nan)
;					((flutter_pos - ind_int - n_buf)>0):$
;					((flutter_pos - n_buf)<(n_elements(aftime)-1))], /nan)
			resid_mean = mean( $
				resid[ $
					s_pos2:(s_pos2 +45L)], /nan)
;					((flutter_pos - ind_int - n_buf)>0):$
;					((flutter_pos - n_buf)<(n_elements(aftime)-1))], /nan)
			threshold = resid_mean + resid_std * resid_sigma
			if ((finite(threshold))[0] eq 0) then begin
				str= ' - Cannot calculate threshold (with ch ' + $
					strtrim(string(neighbor[j]), 2) + '). Continue..'
				lmessage, 9, str, /cont
				if (keyword_set(log)) then $
					printf, lun, str
;				check = [check, 0B]
				continue
			endif

			dummy = where(resid[s_pos2:e_pos2] ge threshold, count)
			if (count ge 1) then begin
				str = string(neighbor[j], threshold, $
					format = '(" - residual (with ch ", i0, ") > ", f0, ". Continue..")')
				lmessage, 9, str, /cont
				if (keyword_set(log)) then $
					printf, lun, str
				check = [check, 1B]
			endif else begin
				str = string(neighbor[j], threshold, $
					format = '(" - residual (with ch ", i0, ") < ", f0, ". Continue..")')
				lmessage, 9, str, /cont
				if (keyword_set(log)) then $
					printf, lun, str
				check = [check, 0B]
			endelse

;plot_min = min( $
;	[min(resid[s_pos2:e_pos2]), $
;	min(smooth(flux[ch, s_pos2:e_pos2], smtwidth, /nan)/cal_target, /nan), $
;	min(smooth(flux[neighbor[j]-1, s_pos2:e_pos2], smtwidth, /nan)/cal_comp, /nan), $
;	min(ofst[s_pos2:e_pos2]/cal_comp)], /nan )
;plot_max = max( $
;	[max(resid[s_pos2:e_pos2]), $
;	max(smooth(flux[ch, s_pos2:e_pos2], smtwidth, /nan)/cal_target, /nan), $
;	max(smooth(flux[neighbor[j]-1, s_pos2:e_pos2], smtwidth, /nan)/cal_comp, /nan), $
;	max(ofst[s_pos2:e_pos2]/cal_comp, /nan), $
;	threshold+threshold/4.0], /nan )
;
;			plot, smooth(flux[ch, *], smtwidth, /nan)/cal_target, $
;				xrange = [s_pos2, e_pos2], psym=10, $
;				yrange = [plot_min, plot_max], /xstyle, /ystyle, $
;				title = 'ch '+string(ch+1)+' w/ neighbor '+string(neighbor[j])
;			oplot, smooth(flux[neighbor[j]-1, *], smtwidth, /nan)/cal_comp, $
;				psym=10, color = '00ff00'x
;			oplot, ofst/cal_comp, color = 'ff0000'x, psym=10
;
;print, ch+1, neighbor[j], threshold, $
;	min(smooth(flux[neighbor[j]-1, *], smtwidth, /nan), /nan), $
;	max(smooth(flux[neighbor[j]-1, *], smtwidth, /nan), /nan)
;
;			oplot, resid, color = '0000ff'x, psym=10
;			plots, !x.crange, [threshold, threshold]
;			wait, 2
;;			read, gomi & if gomi eq 2 then stop

		endfor

		if (n_elements(check) eq 1) then begin
			str = " --> No available data. Do nothing.. "
			lmessage, 9, str, /cont
			if (keyword_set(log)) then $
				printf, lun, str
			continue
		endif

		check = check[1:*]

		dummy = where(check eq 1B, count)
		if (count ge n_elements(check) * 0.5) then begin
			str = " --> : more than 50 % > threshold. Continue.. "
			lmessage, 9, str, /cont
			if (keyword_set(log)) then $
				printf, lun, str
		endif else begin
			str = " --> : less than 50 % > threshold. Resetting flags.. "
			lmessage, 9, str, /cont
			if (keyword_set(log)) then $
				printf, lun, str
			confirm[ch, flutter_pos] = 0B
		endelse

	endfor

endfor
bd = bd and confirm


lmessage, 7, '== Detection result ==', /continue
if (keyword_set(log)) then $
	printf, lun, '== Detection result =='

FLTR = ccf_read_ccf(ccf_fltr, tsd, error = error)
if (error eq !FAIL) then begin
	lmessage, 1, 'CCF ' + ccf_fltr + ' cannot be found.', /cont
	return, !FAIL
endif
;fltr = replicate({scale:1.912d, index:0.256d}, chsize)

; mask after the flutter 
; -------------------------
for ch = 0, chsize-1 do begin
	fltr_index = where(bd[ch, *], count)
	if (count eq 0) then continue

	intval, fltr_index, s_pos, e_pos

	for j = 0, n_elements(s_pos)-1 do begin

; ----------------------------------------
; masking START

;t_chunk = 5.0
;dt = aftime[1]-aftime[0]
;ind_int = round(t_chunk/dt)
;n_baseline = 12L

afpos = aftime[s_pos[j]]
f_s = s_pos[j]

smoothed = smooth(flux[ch, *], ind_int, /nan)

;if (keyword_set(plot_result)) then begin
;	xrange = [afpos - 80d, afpos + 40d]
;	wset, 1
;	!p.multi = [0, 1, 2]
;	plot, aftime, smoothed, xrange = xrange, /nodata
;	oplot, aftime, flux[ch, *], psym = 10, color = '0000ff'x
;	oplot, aftime, smoothed, psym = 10
;	;oplot, aftime, flutter[ch, *], psym = 10, color = '00ff00'x
;endif

peak = max( $
	smoothed[((f_s-ind_int)>0):(f_s+ind_int*1)<(n_elements(aftime)-1)], /nan)
baseline = mean( $
	smoothed[((f_s-(n_baseline+1)*ind_int)>0):((f_s-ind_int)>0)], /nan)

;if (peak le baseline) then continue

;if (keyword_set(plot_result)) then begin
;	plots, [1d, 1d]*(aftime[f_s]), !y.crange
;	plots, !x.crange, [1d, 1d]*peak
;	plots, [1d, 1d]*aftime[((f_s-(n_baseline+1)*ind_int)>0)], !y.crange, line=2
;	plots, [1d, 1d]*aftime[((f_s-ind_int)>0)], !y.crange, line=2
;	plots, !x.crange, [1d, 1d]*baseline, line = 2
;endif

cal_index = where(cal[ch, 0:(f_s-ind_int)>0], count)
if (count eq 0) then begin
	str= ' - Cannot found cal lamps before the detection (' + $
		string(ch+1, j+1, afpos, format = '(2(i0, ", "), f0)') + $
			'). Continue..'
	lmessage, 9, str, /cont
	if (keyword_set(log)) then $
		printf, lun, str
	continue
endif

intval, cal_index, s_cal, e_cal

; determine the turn-on time of the pulse cal.
; ------------------------------
dt_cal = e_cal - s_cal
index = where(dt_cal lt 15, count)
if (count ne 0) then begin
	plst = mean(e_cal[index] - s_cal[index] + 1, /nan)
endif else begin
	str= ' - Cannot found the pulse cal lamp before the detection (' + $
		string(ch, flutter_pos, format = '(i0, ",", i0)') + $
		'). Using 10 samples as the turn-on time of the pulse cal..'
	lmessage, 9, str, /cont
	if (keyword_set(log)) then $
		printf, lun, str
	plst = 10L
endelse

; Get DAC value for cal lamps
; ------------------------------
dacval = daccal[ s_cal + plst/2 ]
turn_on_idx = where(dacval ne 0, count)
if (count eq 0) then begin
	str= ' - No turn-on cal lamps (' + $
		string(ch, flutter_pos, format = '(i0, ",", i0)') + '). Continue..' 
	lmessage, 9, str, /cont
	if (keyword_set(log)) then $
		printf, lun, str
	continue
endif

; Get the last pulse cal before the detection
; ------------------------------
turn_on_idx = max(turn_on_idx)

;s_cal = s_cal[index]
;e_cal = e_cal[index]
;s_cal = s_cal[n_elements(s_cal)-1]
;e_cal = e_cal[n_elements(e_cal)-1]
s_cal = s_cal[turn_on_idx]
e_cal = e_cal[turn_on_idx]
dacfactor = dacval[turn_on_idx]/96d		;Normalize by the pulse cal DAC value.

;cal_lamp = replicate(!values.d_nan, n_elements(aftime))
;cal_lamp[s_cal:e_cal] = flux_unflagged[ch, s_cal:e_cal]
;if (keyword_set(plot_result)) then begin
;	plot, aftime, cal_lamp, color = '00ffff'x, psym = 10, $
;		xrange = [aftime[s_cal]-1, aftime[e_cal]+1]
;	oplot, aftime, flux_unflagged[ch, *]
;	;plots, aftime[s_cal]*[1d, 1d], !y.crange, line = 3
;	!p.multi = 0
;endif

;cal_integ = total(cal_lamp[s_cal:e_cal], /nan)
;cal_integ = total(flux_unflagged[ch, s_cal:e_cal], /nan)
;cal_integ = total(flux_cal[ch, s_cal:e_cal], /nan)
cal_integ = total(flux_cal[ch, s_cal:s_cal + plst-1], /nan)/dacfactor

;return, (peak-baseline)/(cal_integ-baseline)

;printf, lun, tsdfile[fileindex], afpos, $
;	ch+1, j+1, (peak-baseline)/(cal_integ-baseline), $
;	format = '(a0, x, f0, x, i0, x, i0, x, f0)'

masktime = 10^(fltr[ch].scale + $
	fltr[ch].index * alog10((peak-baseline)/(cal_integ-baseline)) )
if (not finite(masktime)) then begin

	tmpstr = string(j+1, ch+1, $
		afpos, $
		peak, baseline, cal_integ, $
		format = '("Flutter # ", i0, ' + $
			'" in ch ", i0, " (aftime = ", f0, ") : ' + $
			' peak = ", f0, " base = ", f0, " cal = ", f0, " : ' + $
			' failed to derive masktime. Continue..")')
	lmessage, 9, tmpstr, /continue

	if (keyword_set(log)) then begin
		printf, lun, tmpstr
	endif

	continue

endif
masklength = masktime/dt

;print, fltr[ch].scale, fltr[ch].index

if ( (s_pos[j] + masklength) ge n_elements(aftime)) then begin

	maskindex = n_elements(aftime) - 1

endif else begin

	if (keyword_set(mask_till_cal)) then begin
		maskindex = where(cal[ch, ((s_pos[j] + masklength)<(n_elements(aftime)-1)):*], count)
		if (count eq 0) then begin
			maskindex = n_elements(aftime) - 1
		endif else begin
			maskindex = min(maskindex) + s_pos[j] + masklength
		endelse
	endif else begin
		maskindex = s_pos[j] + masklength
	endelse

endelse

bd[ch, s_pos[j]:maskindex] = 1B


tmpstr = string(j+1, ch+1, $
	afpos, $
	(peak-baseline)/(cal_integ-baseline), $
	masktime, $
	(maskindex - s_pos[j])*dt, $
	format = '("Flutter # ", i0, ' + $
		'" in ch ", i0, " (aftime = ", f0, ") : ' + $
		' jump/lamp(max) = ", f0, ", masktime = ", f0, " : ' + $
		'masked ", f0, " seconds after the detection")')
lmessage, 9, tmpstr, /continue
if (keyword_set(log)) then begin
	printf, lun, tmpstr

endif


; masking END
; ----------------------------------------

	endfor

endfor



; restore quality flag 
; ------------------------------
for j = 0, n_elements(ind_start)-1 do begin
	no_bd[*, ind_start[j]:ind_end[j]] = $
		no_bdflag[*, j] # replicate(1.0, ind_end[j]-ind_start[j]+1)
endfor


; Save flags to TSD
; ------------------------------
flutter = tsd->get_pix_flag_tsd('FIS_OBS', 'flutter')
flutter = flutter or bd
stat = tsd->set_pix_flag_tsd('FIS_OBS', 'flutter', flutter)
if (stat eq !FAIL) then begin
	tmpstr = 'Failed to save pix flag (flutter) to TSD'
	lmessage, 1, tmpstr, /continue
	errcnt ++
	if (keyword_set(log)) then $
		printf, lun, tmpstr
endif

bad = tsd->get_pix_flag_tsd('FIS_OBS', 'bad')
bad = bad or bd
stat = tsd->set_pix_flag_tsd('FIS_OBS', 'bad', bad)
if (stat eq !FAIL) then begin
	tmpstr = 'Failed to save pix flag (bad) to TSD'
	lmessage, 1, tmpstr, /continue
	errcnt ++
	if (keyword_set(log)) then $
		printf, lun, tmpstr
endif

qual_flutter = tsd->get_quality_tsd('FIS_OBS', 'qual_flutter')
qual_flutter = qual_flutter or no_bd
stat = tsd->set_quality_tsd('FIS_OBS', 'qual_flutter', qual_flutter)
if (stat eq !FAIL) then begin
	tmpstr = 'Failed to save quality flag (qual_flutter) to TSD'
	lmessage, 1, tmpstr, /continue
	errcnt ++
	if (keyword_set(log)) then $
		printf, lun, tmpstr
endif


; +-- debug ---
data_struct = {$
	ind_start:ind_start, ind_end:ind_end, $
	flux:flux, flux_org:flux_org, mn:mn, mn_org:mn_org, sd:sd, d_mn:d_mn, d_sd:d_sd, $
	mn_d_mn:mn_d_mn, sd_d_mn:sd_d_mn, $
	mn_d_sd:mn_d_sd, sd_d_sd:sd_d_sd, $
	thmn:thmn, thsd:thsd, $
	thmn_m:thmn_m, thsd_m:thsd_m, $
	bdflag:bdflag, bd:bd, $
	no_bdflag:no_bdflag, no_bd:no_bd, $
	cal:cal $
}
; --- debug ---

tmpstr = strcompress(' '+version+' done.')
lmessage, 5, tmpstr, /continue
dummy = tsd_fxaddhist(tsd, tmpstr, /prres)
if (dummy ne 0) then begin
	tmpstr = 'Unable to write history to TSD'
	lmessage, 3, tmpstr, /continue
	errcnt ++
	if (keyword_set(log)) then $
		printf, lun, tmpstr
endif

if (keyword_set(log)) then begin
	if (errcnt eq 0) then begin
		printf, lun, $
			'Seems to be finished successfully.. :' + systime()
	endif else begin
		printf, lun, $
			'Seems to be finished with some ERRORS :' + systime()
	endelse
	close, lun
	free_lun, lun
endif

return, !SUCCEED

end
