; $Id: gb_set_flag_rstanom.pro,v 1.2 2006/01/16 00:01:27 baba Exp $
;
; Copyright (c) 2004-2005, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;   gb_set_flag_rstanom.pro
;
; PURPOSE:
;   Set flags at anomaly after reset. Flags, rstanom, and
;       bad are set.
;
; CATEGORY:
;   Function. Green Box procedure in ASTRO-F pipeline.
;
; CALLING SEQUENCE:
;   Result = gb_set_flag_rstanom(tsd
;            [, ccf_rstanom = ccf_rstanom])
;
;    ;; Default values for ccf_rstanom is 1
;
; INPUTS:
;   tsd          TSD object reference.
;   ccf_rstanom  Calibration constants file; array with
;                pixel numbers (75 or 100) times 2.
;
; OUTPUTS:
;   tsd          TSD with new flags.
;
; RETURN:
;   !SUCCEED if function ended successfully.
;   !FAIL in failure.
;
; RESTRICTIONS:
;
; PROCEDURE:
;
; EXAMPLE:
;
; REFERENCE:
;
; MODIFICATION HISTORY:
;   May. 2004     Written by A. Kawamura.
;   Oct. 2004     Changed from library to an independent module.
;                             TSD3 is implemented.
;                             IT#2
;   Nov. 2005     Last modified by Makiuti
;-

FUNCTION gb_set_flag_rstanom, tsd,$
    ccf_rstanom = ccf_rstanom

;ON_ERROR 2

;;; Starting gb_set_flag_rstanom ;;;
    LMESSAGE, 5, 'Executing GB_Set_Flag_Rstanom...', /CONTINUE

;;; Check detector type, LW or SW ;;;

   det_type = tsd->Get_datatype()

   CASE det_type OF

        'FIS_LW'  : BEGIN
                    num_ch=75
                    nw_delim=45
                    msg='Detector type is LW.'
                    LMESSAGE, 7, msg ,/CONTINUE
                    END

        'FIS_SW'  : BEGIN
                    num_ch=100
                    nw_delim=60
                    msg='Detector type is SW.'
                    LMESSAGE, 7, msg ,/CONTINUE
                    END

        ELSE      : BEGIN
                    msg='Check data. Unknown detector type.'
                    LMESSAGE, 0, msg ,/CONTINUE
                    return,!FAIL
                    END

   ENDCASE

;;----------------------------------------------------------
;; Main procedure
;;----------------------------------------------------------
IF (N_elements(ccf_rstanom) EQ 0) THEN ccf_rstanom = 1 ;#default value

IF (ccf_rstanom NE 0) THEN BEGIN ;; Read CCF and set flags

  ;;; read CCF ;;;
  LMESSAGE, 7, "Read CCF for Reset anomary", /CONTINUE
  IF (num_ch EQ 100) THEN BEGIN ;; SW
    gb_rst_anom_ccf = CCF_Read_CCF('RSTAS', tsd, HEADER=hdr, ERROR=err)
  ENDIF ELSE IF (num_ch EQ 75) THEN BEGIN ;; LW
    gb_rst_anom_ccf = CCF_Read_CCF('RSTAL', tsd, HEADER=hdr, ERROR=err)
  ENDIF


  ;;; setting flags (rstanom for TSD3) ;;;
  LMESSAGE, 7, 'Setting flag, rstanom, at reset anom. ...',/CONTINUE

  frstanom = tsd->Get_pix_flag_tsd('FIS_OBS','RstAnom')
  frstanom[*] = 0


  ;;; check reset frame position ;;;

  IF num_ch EQ 75 THEN BEGIN ; for LW

    rstctw = tsd->Get_value_tsd('FIS_OBS','RstCntWideL')
    rstctn = tsd->Get_value_tsd('FIS_OBS','RstCntN170')

  ENDIF ELSE BEGIN ; for SW

    rstctw = tsd->Get_value_tsd('FIS_OBS','RstCntWideS')
    ;inst=tsd->get_header_keyword_tsd('primary','instrume')
    ;IF (STRMATCH(inst, 'FISVI*')) THEN rstctw = rstctw ELSE rstctw = SHIFT(rstctw, 1); for past laboratory data
    rstctn = tsd->Get_value_tsd('FIS_OBS','RstCntN60')

  ENDELSE

  rstw = rstctw - SHIFT(rstctw, 1)
  rstn = rstctn - SHIFT(rstctn, 1)

  ;;; get an array of sampling mode
  packetid = tsd->Get_value_tsd('FIS_OBS','PacketID')
  aftime = tsd->Get_value_tsd('FIS_OBS','AFTIME')
  inst = tsd->Get_header_keyword_tsd('PRIMARY','INSTRUME')
  smode = smpl_mode(packetid, AFTIME=aftime[0], TYPE=inst)


  ;;-----------------------
  ;; Set temporal reset anomary flag
  ;;-----------------------

  ;;; WIDE band ;;;

  indrw=WHERE((rstw NE 0) AND (smode NE 1), cntrw)
  IF cntrw GT 1 THEN BEGIN

    dummyw = reform(frstanom[0:nw_delim-1,*])
    frm_num_of_data = n_elements(dummyw[0,*])

    FOR ch = 0, nw_delim-1 DO BEGIN
      FOR n = 0, cntrw-1 DO BEGIN
        IF (indrw[n]+gb_rst_anom_ccf[ch,1] LT frm_num_of_data) THEN BEGIN
          IF (gb_rst_anom_ccf[ch,0] GE 0) and (gb_rst_anom_ccf[ch,1] GE 1) THEN BEGIN
            dummyw[ch,indrw[n]+gb_rst_anom_ccf[ch,0]:indrw[n]+gb_rst_anom_ccf[ch,1]-1]=1
          ENDIF
        ENDIF
      ENDFOR
    ENDFOR

  ENDIF ELSE BEGIN
     LMESSAGE, 3, 'No reset is found in data (wide band)...', /CONTINUE
     RETURN,!SUCCEED
  ENDELSE

  ;;; NARROW band ;;;

  indrn=WHERE((rstn NE 0) AND (smode NE 1), cntrn)
  IF cntrn GT 1 THEN BEGIN
    dummyn = reform(frstanom[nw_delim:num_ch-1,*])
    frm_num_of_data = n_elements(dummyn[0,*])

    FOR ch = 0, num_ch-nw_delim-1 DO BEGIN
       FOR n = 0, cntrn-1 DO BEGIN
         IF (indrn[n]+gb_rst_anom_ccf[ch+nw_delim,1] LT frm_num_of_data) THEN BEGIN
           IF (gb_rst_anom_ccf[ch+nw_delim,0] GE 0) and (gb_rst_anom_ccf[ch+nw_delim,1] GE 1) THEN BEGIN
             dummyn[ch,indrn[n]+gb_rst_anom_ccf[ch+nw_delim,0]:indrn[n]+gb_rst_anom_ccf[ch+nw_delim,1]-1]=1
           ENDIF
         ENDIF
       ENDFOR
    ENDFOR

  ENDIF ELSE BEGIN
     LMESSAGE, 3, 'No reset is found in data (narrow band)...', /CONTINUE
     RETURN,!SUCCEED
  ENDELSE

  ;;;
  ;;-----------------------
  ;; Set flags ('rstanom' & 'bad')
  ;;-----------------------
  msg = "Setting flag rstanom..."
  LMESSAGE, 7, msg, /CONTINUE

  frstanom=[dummyw,dummyn]
  ret = tsd->Set_pix_flag_tsd('FIS_OBS','RstAnom', frstanom)

  fbad = tsd->Get_pix_flag_tsd('FIS_OBS','Bad')
  fbad_new = fbad + frstanom
  ind = WHERE (fbad_new NE 0, cnt)
  IF cnt NE 0 THEN fbad_new(ind)=1
  err = tsd->Set_pix_flag_tsd('FIS_OBS','Bad', fbad_new)

ENDIF ELSE BEGIN ;; Dose not read CCF, dose not set any flags

    LMESSAGE, 7, "CCF for reset anomary is not read.", /CONTINUE
    ;; No flag is not set

ENDELSE


;;----------------------------------------------------------
;; Write a history in TSD header
;;----------------------------------------------------------
   LMESSAGE, 7, 'Writing HISTORY', /CONTINUE
   name_ver = gb_get_name_ver(infile='gb_set_flag_rstanom')
   hist_message = STRCOMPRESS(' '+name_ver+' done. CCF= ')

    ;  hist_message = hist_message+ccf_rstanom

   dummy = tsd_fxaddhist(tsd, hist_message, /prres)
   IF dummy NE 0 THEN $
     LMESSAGE, 3, 'Unable to write HISTORY', /CONTINUE

   return, !SUCCEED

END
