; $Id: gb_tr_hk1.pro,v 1.4 2006/01/18 06:57:05 baba Exp $
;
; Copyright (c) 2002, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	gb_tr_hk1.pro
; PURPOSE:
;	Function for transient correction with a modified F-S model.
;       The transient effects of the detector and the cosmic-ray
;	after effects are corrected. 
; CATEGORY:
;	Function. Green Box procedure in ASTRO-F pipeline.
; CALLING SEQUENCE:
;	result = gb_tr_hk1(tsd
;	[, CAL_CONST_FILE = cal_const_file, SUMMARY = summary])
; INPUTS:
;	tsd		TSD object reference.
;	cal_const_file	Calibration constants file; 
;			format, usage are TBD (optional.)
; OUTPUTS:
;	tsd		TSD object reference. Transient corrected TSD.
;	(SUMMARY = summary)	Summary of the procedure.
;			How to manage the summaries is under 
;			considerations.
; RETURN:
;	!SUCCEED if function ended successfully.
;	!FAIL in failure.
; RESTRICTIONS:
;	
; PROCEDURE:
;	  In this pipeline process, the model reported by Kaneda 
;	is imported to correct the transient and cross-talk effects. 
;                           (1-betan)*Jninf*(Jnin - betan*Jninf)  
;    Jnf = betan*Jninf + ------------------------------------------
;                        Jnin-betan*Jninf+(Jninf-Jnin)exp(-dt/taun)
;    Jnin-Jn1f = betan1*(Jninf-Jn1inf)
;
;
;                               g1
;    gn =  -----------------------------------------------------
;           -- Jkinf - (Jk1inf-d1)   _                dt
;           \  -------------------  | | exp(-1* -------------- )     
;        1+a/   ave(Jk1inf-d1,d2)^ef1     l=k    b*Jlinf^(-1*ef2)
;           --
;           k=2
;           
;         calculate if Jkinf - (Jk1inf-c) > fsig * sigma ... fsig given
;
; REFERENCE:
;	Kaneda, internal document in Japanese on transient and 
;	cross-talk correction(2002 Apr. 23).
; MODIFICATION HISTORY:
; 	Sept. 2002		Written by A. Kawamura.
;	Dec. 2002		Header is added.
;-
FUNCTION gb_tr_hk1,vdetp,num_ch
COMPILE_OPT IDL2
MESSAGE,"Starting gb_tr_hk1 ...",/CONTINUE
;RESOLVE,'tr1.ccf'
;; g1=g1, ej_in=ej_in
;; a=a,b=b,d1=d1,d2=d2,ef1=ef1,ef2=ef2,fsig=fsig,
;; Jsf=Jsf, Joff=Joff, 

     ;;; Define parameters ;;;
;g1=4e-2 ;; tr_param[0]
g1=1.D
ej_in = 10000. ;; tr_param[1]
pi=!pi
c = 2.99792458*1e+8
epsiron = 16.6*8.8e-12*4*pi*c*c*1e-7  ;;; permiability in esu
ej=ej_in / c *1e+4                            ;;; cm^2 
area= 1.* 1e^2     

a=0.002
b=15.
d1=2.
d2=70.
ef1=1.
ef2=1.
fsig=5.
Jsf=1.
Joff=0.

sigma=stddev(vdetp[140:144])
;;  set baseline

     ;;; Set parameters ;;;
si = SIZE(vdetp)
arr=si[1]

Jin=vdetp
Joff=Joff*Jsf
;Jin = Jin*Jsf-(0.062*Jsf-0.01) ; * 1e-12 *c*10.                            ;;; pA->esu
;;plot,Jin[0:6] &wait,1

jf=findgen(arr)
jinf=findgen(arr)
g=findgen(arr)
beta=findgen(arr)
tau=findgen(arr)
IF num_ch EQ 75 THEN dt=1./24D ELSE dt=1./15.969D
     ;;; Get first parameters for the first sampling ;;;
x1=1
g[0:x1]= g1


beta1 = 1.-g1*(1.-exp(-1./g1))
beta[0:x1]=beta1

j1in = mean(Jin[0:x1])
j1inf = mean(Jin[0:x1])

tau1 = epsiron *area * ej / 4/pi/beta1 / j1inf
j1f=1D
;j1f = beta1*j1inf + (1.-beta1)*j1inf*(j1in-beta1*j1inf) /     (j1in - beta1*j1inf)+(J1inf-J1in)*exp(-1*dt/tau1)

Jinf[0:x1]=J1inf
Jf[0:x1]=J1f
tau[0:x1]=tau1
     ;;; Calculation ;;;

sfactor=fltarr(arr)
mfactor=fltarr(arr)

FOR n=x1+1, arr-1 DO BEGIN

IF n EQ x1+1 THEN BEGIN
   jn1inf = j1inf
   jn1f = j1f
   betan = beta1
   betan1 = beta1
ENDIF ELSE BEGIN   
   jn1inf = jninf
   jn1f = jnf
   betan1 = betan
ENDELSE

;;**;;
Jnin= Jin[n]

       ;;; getting Gn ;;;
;                               g1
;    gn =  -----------------------------------------------------
;           -- Jkinf - (Jk1inf-d1)   _                dt
;           \  -------------------  | | exp(-1* -------------- )     
;        1+a/   ave(Jk1inf-d1,d2)^ef1     l=k    b*Jlinf^(-1*ef2)
;           --
;           k=2

   jlin = jin[n]
   jkin = jin[n]

dd1=d1
dd2=d2

IF dd1 GT n-1 THEN dd1=n
IF dd2+dd1 GT n-1 THEN BEGIN
    dummy=mean(Jin[0:x1])
    Jk1inav = (dummy*(dd2+dd1-n)+((mean(Jin[0:(n-dd1)]))*(n-dd1)))/dd2
ENDIF ELSE BEGIN

   Jk1inav = mean(Jin[n-dd1-dd2:n-dd1])

ENDELSE

   Jk1in = Jin[n-1]
IF (Jkin-Jin[n-dd1]) GT 5*sigma THEN BEGIN
;sfactor[n-1]=(Jkin-Jin[n-dd1])/(Jk1inav)^ef1
sfactor=1
ENDIF
;IF n NE x1+1 THEN mfactor[0:n-2] = mfactor[0:n-2] +jkin^ef2
factor = sfactor*exp(-1.)
;factor = sfactor*exp(-1*dt/b*mfactor)
;factor=total(factor)
factor=1.
; gn=g1/(1.+a*factor)                     ;!!
; betan=1.-gn*(1.-exp(-1./gn))    ;!!
; g[n] = gn
; beta[n] = betan
g[n] = 1.D
beta[n]=1.D
jninf=(jnin-jn1f)
;jninf = (jnin - jn1f)/betan1 +jn1inf ;;**;;
jninf=jnin
jinf[n] = jninf 
    ;;; calculate
;taun =  epsiron *area * ej / 4/pi/betan1 / jninf ;;**;;
taun =  epsiron *area * ej / 4/pi/betan1
taun=1.
tau[n] = taun
 fsnanm = jninf*(jnin-betan1*jninf)-betan1*jninf*(jnin-betan1*jninf)		;!!
 fsdenom = jnin*(1.-exp(-1*dt/taun)) + jninf*(-1*betan1+exp(-1*dt/taun))	;!!
fsnanm=0.D
fsdenom=1.D
jnf = betan*jninf + fsnanm / fsdenom ;;**;;
jf[n] = jnf

ENDFOR

;;;; result
;print,min(g),max(g)
print,min(Jinf),max(Jinf)
;print,min(vdetp),max(vdetp)
   result=Jinf/g
    sf=1.D  
;   sf=mean(jin[100:200]) /mean(result[100:200])
   c_vdetp=result*sf
ind = WHERE(c_vdetp EQ 0, cnt)
  IF cnt NE 0 THEN c_vdetp[ind] = !values.d_nan
MESSAGE,"Ending gb_tr_hk1...",/CONTINUE
Return,c_vdetp

END
