; $Id: ldsrpc__define.pro,v 1.4 2007/09/07 07:43:19 cyamauch Exp $
;
; Copyright (C) 2004 Hajime BABA  All rights reserved.
;
;+
; CLASS_NAME:
;   LDSRPC_DESERIALIZER
;
; PURPOSE:
;   Parse XML-RPC methodResponse, de-serialize, and create result structure
;
; CATEGORY:
;   Data definition
;
; CALLING SEQUENCE:
;   ldsrpc = OBJ_NEW('ldsrpc')
;
; INPUTS:
;   'ldsrpc'	Name of object class
;
; OUTPUTS:
;   ldsrpc	Name of object instance
;
; RESTRICTIONS:
;   None
;
; COMMENTS:
;   See chapter 'Object Basics' in 'Building IDL Applications' in detail.
;
; DESCRIPTION:
;
; XML Protocols: XML-RPC and SOAP
; www.cafeconleche.org/books/ xmljava/chapters/ch02s05.html
;
;   <!ELEMENT methodCall (methodName, params)>
;   <!ELEMENT methodName (#PCDATA)>
;   <!ELEMENT params     (param*)>
;   <!ELEMENT param      (value)>
;   <!ELEMENT value
;      (i4|int|string|dateTime.iso8601|double|base64|struct|array)>
;  
;   <!ELEMENT i4               (#PCDATA)>
;   <!ELEMENT int              (#PCDATA)>
;   <!ELEMENT string           (#PCDATA)>
;   <!ELEMENT dateTime.iso8601 (#PCDATA)>
;   <!ELEMENT double           (#PCDATA)>
;   <!ELEMENT base64           (#PCDATA)>
;  
;   <!ELEMENT array            (data)>
;   <!ELEMENT data             (value*)>
;   <!ELEMENT struct           (member+)>
;   <!ELEMENT member           (name, value)>
;   <!ELEMENT name             (#PCDATA)>
;  
;   <!ELEMENT methodResponse   (params | fault)>
;   <!ELEMENT fault            (value)>
;
; EXAMPLE:
;
;   Set variables
;     IDL> proxy = 'http://localhost/ldsrpc'
;     IDL> method = 'query'
;     IDL> args = ['1', '2', 'abc'] ; Should be STRARR()
;
;   Initialize
;     IDL> ldsrpc = OBJ_NEW('ldsrpc')
;
;   Call step by step
;     IDL> ldsrpc = ldsrpc->proxy(proxy)
;     IDL> result = ldsrpc->call(method, args)
;
;   Call at once
;     IDL> result = (ldsrpc->proxy(proxy))->call(rpcmethod, args)
;   Below cause an error
;     IDL> result = ldsrpc->proxy(proxy)->call(rpcmethod, args)
;
;   Parse locally stored file
;     IDL> ldsrpc->ParseFile, xmlFile
;
;   Destroy object
;     IDL> OBJ_DESTROY, ldsrpc
;
;   Show results (anonymous strucuture)
;     IDL> help, result
;     IDL> print, result
;
; SIDE EFFECTS:
;   None
;
; MODIFICATION HISTORY:
;   Written by: Hajime Baba, June 2004
;
;-

;----------------------------------------------------------------------------
; Methods which are overrided
;----------------------------------------------------------------------------

PRO ldsrpc::characters, data
    COMPILE_OPT hidden
;    MESSAGE, 'Called: ' + data, /INFORMATIONAL
    self.buffer = TEMPORARY(self.buffer) + data
    data = 0 ; clear memory
END

PRO ldsrpc::startElement, URI, local, strName, attrName, attrValue
    COMPILE_OPT hidden
    MESSAGE, 'Called: ' + strName, /INFORMATIONAL
    CASE strName OF
        "methodResponse":	; Do nothing
        "params":
        "param": self.number.param = self.number.param + 1
        "value": self.type = ''
        ; scalar types
        "int":     self.buffer = ''
        "i4":      self.buffer = '' ; same as "int"
        "long":    self.buffer = ''
        "boolean": self.buffer = ''
        "string":  self.buffer = ''
        "double":  self.buffer = ''
        "dateTime.iso8601": self.buffer = ''
        "base64":  self.buffer = ''
        ; composite types
        "struct": self.nest.struct = self.nest.struct + 1
        "member": self.number.member = self.number.member + 1
        "name":   self.buffer = ''
        "array":  self.nest.array = self.nest.array + 1
        "data": ; TBD
        ; fault
        "fault":  self.isFault = 1
        ELSE: MESSAGE, 'Unknown start tag: "' + strName + '"', /CONTINUE
    ENDCASE
END

PRO ldsrpc::endElement, URI, Local, strName
    COMPILE_OPT hidden
    MESSAGE, 'Called: ' + strName, /INFORMATIONAL
    CASE strName of
        "methodResponse": BEGIN
            IF NOT !QUIET THEN HELP, /HEAP_VARIABLES
        END
        "params": BEGIN
            FOR i=0,self.number.param-1 DO BEGIN
                tag = 'PARAM_' + STRTRIM(STRING(i),2)
                tmp = CREATE_STRUCT(tag, *self.paramArray[i])
                PTR_FREE, self.paramArray[i]
                IF i EQ 0 THEN tmp2 = tmp $
                ELSE tmp2 = CREATE_STRUCT(tmp2, tmp)
            ENDFOR
; Below causes 'Conflicting data structures' error, or
; 'Named structures can't contain anonymous structure members' error.
; See 'Relaxed Structure Assignment' in IDL manual.
;            tmp2 = CREATE_STRUCT(NAME="LDSRPC_methodResponse", tmp2)
            self.params = self->PTR_NEW(tmp2)
            self.number.param = 0
        END
        "param": BEGIN
            self.paramArray[self.number.param - 1] = self.value
        END
        "value": BEGIN
            value = self->castValue2()
            self.value = self->PTR_NEW(value)
            ; mmm... TBD
            IF (self.nest.array GE 1 AND self.type EQ "struct") THEN BEGIN
                self.number.array = self.number.array + 1
                IF self.number.array GT N_ELEMENTS(self.dataArray) THEN BEGIN
                    MESSAGE, '[WARNING] size of data buffer exceeds!', /CONTINUE
                    self.number.array = self.number.array - 1
                ENDIF ELSE BEGIN
                    self.dataArray[self.number.array - 1].value = self.value
                    self.dataArray[self.number.array - 1].type = self.type
                ENDELSE
            ENDIF
        END
        ; scalar types
        "int":     self.type = strName
        "i4":      self.type = strName ; same as "int"
        "long":     self.type = strName
        "boolean": self.type = strName
        "string":  self.type = strName
        "double":  self.type = strName
        "dateTime.iso8601": self.type = strName
        "base64":  self.type = strName
        ; composite types
        "struct": BEGIN
            self.type = strName
            FOR i=0,self.number.member-1 DO BEGIN
                value = self->castValue(*self.memberArray[i].value, $
                                        self.memberArray[i].type)
                PTR_FREE, self.memberArray[i].value ; clear memory
                tmp = CREATE_STRUCT(self.memberArray[i].name, value)
                IF i EQ 0 THEN tmp2 = tmp $
                ELSE tmp2 = CREATE_STRUCT(tmp2, tmp)
            ENDFOR
; Below causes 'Conflicting data structures' error.
; See 'Relaxed Structure Assignment' in IDL manual.
;            name = "LDSRPC_STRUCT_" + STRTRIM(STRING(self.nest.struct-1),2)
;            tmp2 = CREATE_STRUCT(NAME=name, tmp2)
            self.struct = self->PTR_NEW(tmp2)
            self.number.member = 0
            self.nest.struct = self.nest.struct - 1
        END
        "member": BEGIN
            member = {MEMBER, self.name, self.value, self.type }
            self.memberArray[self.number.member - 1] = member
        END
        "name": self.name = self.buffer
        "array": BEGIN
            self.type = strName
            FOR i=0,self.number.array-1 DO BEGIN
                value = self->castValue(*self.dataArray[i].value, $
                                        self.dataArray[i].type)
                PTR_FREE, self.dataArray[i].value
                tag = "VALUE_" + STRTRIM(STRING(i),2)
                tmp = CREATE_STRUCT(tag, value)
                IF i EQ 0 THEN tmp2 = tmp $
                ELSE tmp2 = CREATE_STRUCT(tmp2, tmp)
            ENDFOR
; Below causes 'Conflicting data structures' error.
; See 'Relaxed Structure Assignment' in IDL manual.
;            name = "LDSRPC_ARRAY_" + STRTRIM(STRING(self.nest.array-1),2)
;            tmp2 = CREATE_STRUCT(NAME=name, tmp2)
            self.array = self->PTR_NEW(tmp2)
            self.number.array = 0
            self.nest.array = self.nest.array - 1
        END
        "data": 
        ; fault
        "fault": BEGIN
            self.fault = self.value
        END
        ELSE: MESSAGE, 'Unknown end tag: "' + strName + '"', /CONTINUE
    ENDCASE
END

;----------------------------------------------------------------------------
; Methods which are created
;----------------------------------------------------------------------------

FUNCTION ldsrpc::castValue, buffer, type
    COMPILE_OPT hidden
    MESSAGE, 'Called', /INFORMATIONAL
    value = ''
    CASE type OF
        'int':     value = LONG(buffer) ; A 32-bit signed integer
        'i4':      value = LONG(buffer) ; same as "int"
        'long':    value = LONG64(buffer) ; A 64-bit signed integer
        'boolean': value = BYTE(buffer) ; Should be 0 or 1
        'string':  value = STRING(buffer)
        'double':  value = DOUBLE(buffer)
        'dateTime.iso8601': value = STRING(buffer)
        'base64':  value = buffer
        'struct':  value = buffer
        'array':   value = buffer
        '':        value = STRING(buffer) ; in case of data type is omitted
        ELSE: MESSAGE, 'Unknown scalar type: ' + type
    ENDCASE
    buffer = 0 ; clear memory
    RETURN, value
END

FUNCTION ldsrpc::castValue2
    COMPILE_OPT hidden
    MESSAGE, 'Called', /INFORMATIONAL
    value = ''
    CASE self.type OF
        'int':     value = LONG(self.buffer) ; A 32-bit signed integer
        'i4':      value = LONG(self.buffer) ; same as "int"
        'long':    value = LONG64(self.buffer) ; A 64-bit signed integer
        'boolean': value = BYTE(self.buffer) ; Should be 0 or 1
        'string':  value = STRING(self.buffer)
        'double':  value = DOUBLE(self.buffer)
        'dateTime.iso8601': value = STRING(self.buffer)
;        'base64':  value = STRING(self.buffer)
        'base64': BEGIN
            b64 = OBJ_NEW('base64')
            value = b64->decode_base64(BYTE(self.buffer))
            OBJ_DESTROY, b64
        END
        'struct': BEGIN
            IF N_ELEMENTS(*self.struct) NE 0 THEN value = *self.struct
            PTR_FREE, self.struct
        END
        'array': BEGIN
            IF N_ELEMENTS(*self.array) NE 0 THEN value = *self.array
            PTR_FREE, self.array
        END
        '': value = STRING(self.buffer) ; in case of data type is omitted
        ELSE: MESSAGE, 'Unknown scalar type: ' + self.type
    ENDCASE
    self.buffer = 0 ; clear memory
    RETURN, value
END

; Writing a Recursive Function
; http://www.dfanning.com/tips/recursive_function.html
FUNCTION ldsrpc::createValueXML, var
    COMPILE_OPT hidden
    FORWARD_FUNCTION ldsrpc::createValueXML
    value = ''
    CASE SIZE(var, /TYPE) OF
        0: ; Undefined
        1: BEGIN
            IF N_ELEMENTS(var) EQ 1 THEN BEGIN
                value = '<int>' + STRTRIM(STRING(var),2) + '</int>' ; BYTE
            ENDIF ELSE BEGIN ; BYTARR
                b64 = OBJ_NEW('base64')
                value = '<base64>' + STRING(b64->encode_base64(var)) + '</base64>'
                OBJ_DESTROY, b64
            ENDELSE
        END
;        2: value = '<int>' + STRTRIM(STRING(var),2) + '</int>' ; INT
;        3: value = '<int>' + STRTRIM(STRING(var),2) + '</int>' ; LONG
;        4: value = '<double>' + STRTRIM(STRING(var),2) + '</double>' ; FLOAT
;        5: value = '<double>' + STRTRIM(STRING(var),2) + '</double>' ; DOUBLE
        2: BEGIN ; INT
            IF SIZE(var, /DIMENSIONS) EQ 0 THEN BEGIN ; scalar
                value = '<int>' + STRTRIM(STRING(var),2) + '</int>'
            ENDIF ELSE BEGIN ; array
                dims=SIZE(var,/DIMENSIONS)
                value = '<array><data>'
                FOR i=0,dims[0]-1 DO BEGIN
                    value = value + '<value><int>' + STRTRIM(STRING(var[i]),2) + '</int></value>'
                ENDFOR
                value = value + '</data></array>'
            ENDELSE
        END
        3: BEGIN ; LONG
            IF SIZE(var, /DIMENSIONS) EQ 0 THEN BEGIN ; scalar
                value = '<int>' + STRTRIM(STRING(var),2) + '</int>'
            ENDIF ELSE BEGIN ; array
                dims=SIZE(var,/DIMENSIONS)
                value = '<array><data>'
                FOR i=0,dims[0]-1 DO BEGIN
                    value = value + '<value><int>' + STRTRIM(STRING(var[i]),2) + '</int></value>'
                ENDFOR
                value = value + '</data></array>'
            ENDELSE
        END
        4: BEGIN ; FLOAT
            IF SIZE(var, /DIMENSIONS) EQ 0 THEN BEGIN ; scalar
                value = '<double>' + STRTRIM(STRING(var,FORMAT='($,%"%17.14g")'),2) + '</double>'
            ENDIF ELSE BEGIN ; array
                dims=SIZE(var,/DIMENSIONS)
                value = '<array><data>'
                FOR i=0,dims[0]-1 DO BEGIN
                    value = value + '<value><double>' + STRTRIM(STRING(var[i],FORMAT='($,%"%17.14g")'),2) + '</double></value>'
                ENDFOR
                value = value + '</data></array>'
            ENDELSE
        END
        5: BEGIN ; DOUBLE
            IF SIZE(var, /DIMENSIONS) EQ 0 THEN BEGIN ; scalar
                value = '<double>' + STRTRIM(STRING(var,FORMAT='($,%"%17.14g")'),2) + '</double>'
            ENDIF ELSE BEGIN ; array
                dims=SIZE(var,/DIMENSIONS)
                value = '<array><data>'
                FOR i=0,dims[0]-1 DO BEGIN
                    value = value + '<value><double>' + STRTRIM(STRING(var[i],FORMAT='($,%"%17.14g")'),2) + '</double></value>'
                ENDFOR
                value = value + '</data></array>'
            ENDELSE
        END
;         6: BEGIN ; COMPLEX
;             value = value + '<struct> + $
;                     '<member><name>real_part</name><value><double>' + $
;                     STRING(REAL_PART(var)) + $
;                     '</double></value></member>' + $
;                     '<member><name>imaginary</name><value><double>' + $
;                     STRING(IMAGINARY(var)) + $
;                     '</double></value></member></struct>'
;         END
        7: value = '<string>' + STRING(var) + '</string>' ; STRING
        8: BEGIN ; STRUCT
            value = value + '<struct>'
            tag_names = STRLOWCASE(TAG_NAMES(var))
            FOR j=0,N_TAGS(var)-1 DO BEGIN
                value = value + '<member>'
                value = value + '<name>' + tag_names[j] + '</name>'
                value = value + '<value>'
                value = value + self->createValueXML(var.(j))
                value = value + '</value>'
                value = value + '</member>'
            ENDFOR
            value = value + '</struct>'
        END
;         9: BEGIN ; DCOMPLEX
;             value = value + '<struct>' + $
;                     '<member><name>real_part</name><value><double>' + $
;                     STRING(REAL_PART(var)) +
;                     '</double></value></member>' + $
;                     '<member><name>imaginary</name><value><double>' + $
;                     STRING(IMAGINARY(var)) + $
;                     '</double></value></member></struct>'
;         END
        10: value = self->createValueXML(*var) ; POINTER
;         11: ; OBJECT Reference
;         12: value = '<int>' + STRING(LONG(var)) + '</int>' ; UINT
;         13: value = '<int>' + STRING(LONG(var)) + '</int>' ; ULONG
;         14: ; LONG64
;         15: ; ULONG64
        ELSE: MESSAGE, 'Unsupported data type: ' + SIZE(var,/TNAME), /CONTINUE
    ENDCASE
    RETURN, value
END


FUNCTION ldsrpc::createMethodCall
    COMPILE_OPT hidden
    ; HTTP body part (payload)
;    body = '<?xml version="1.0"?>'
    body = '<?xml version="1.0" encoding="UTF-8"?>'
    body = body + '<methodCall>'
    body = body + '<methodName>' + self.methodCall.methodName + '</methodName>'
    IF N_ELEMENTS(*self.methodCall.params) GT 0 THEN BEGIN
        body = body + '<params>'
        FOR i=0,N_ELEMENTS(*self.methodCall.params)-1 DO BEGIN
            body = body + '<param>'
            value = self->createValueXML((*self.methodCall.params)[i])
            body = body + '<value>' + value + '</value>'
            body = body + '</param>'
        ENDFOR
        body = body + '</params>'
    ENDIF
    body = body + '</methodCall>'

;    hostname = GETENV('HOST')
    ; UNIX and Windows only (/NOSHELL, /NULL_STDIN)
;    SPAWN, ['hostname'], result, errresult, $
;      COUNT=count, EXIT_STATUS=exit_status, PID=pid, /NOSHELL, /NULL_STDIN
;    hostname = STRTRIM(result,2)
    hostname = 'dummyhost.example.com'

    comma = STRPOS(self.methodCall.methodName, '.')
    class = STRMID(self.methodCall.methodName, 0, comma)
    method = STRMID(self.methodCall.methodName, comma+1, STRLEN(self.methodCall.methodName))
    ; for UNIX
    username = GETENV('USER')
    hostname = GETENV('HOSTNAME')
    ; for Windows
    IF username EQ '' THEN username = GETENV('USERNAME')
    IF hostname EQ '' THEN hostname = GETENV('COMPUTERNAME')

    ; HTTP header part
;    header = [ 'POST ' + self.methodCall.proxy + ' HTTP/1.0' ]
    header = [ 'Content-Type: text/xml' ]
    header = [ header, 'Content-Length: ' + STRTRIM(STRING(STRLEN(body)),2) ]
    header = [ header, 'User-Agent: ' + username + '@' + hostname + ' IDL/' + !VERSION.RELEASE + $
             ' on ' + !VERSION.OS + '/' + !VERSION.ARCH ]
;    header = [ header, 'Host: ' + hostname ]
;    header = [ header, 'SOAPAction: ' + 'LDS#queryViaLDSRPC' ]
    header = [ header, 'SOAPAction: ' + class + '#' + method ]

    RETURN, CREATE_STRUCT('header', header, 'body', body)
END


; psuedo_tmpnam - create a name for a temporary file
FUNCTION ldsrpc::psuedo_tmpnam
    COMPILE_OPT hidden
    chars = ['a','b','c','d','e','f','g','h','i','j','k','l','m',$
             'n','o','p','q','r','s','t','u','v','w','x','y','z',$
             'A','B','C','D','E','F','G','H','I','J','K','L','M',$
             'N','O','P','Q','R','S','T','U','V','W','X','Y','Z']
    nStr = 10
    seed = SYSTIME(/SECONDS)
    index = FIX(RANDOMU(seed, nStr) * N_ELEMENTS(chars))
    ; for UNIX
    username = GETENV('USER')
    ; for Windows
    IF username EQ '' THEN username = GETENV('USERNAME')
    RETURN, username + STRJOIN(chars[index], '')
END

;; imported from webget.pro
PRO ldsrpc::MimeType,  Header, Class, Type, Length, Serial
    COMPILE_OPT hidden
;;
;; MIME type recognition
;
  Class = 'text'
  Type = 'simple'               ; in case no information found...    
  Serial = -1
  def = strupcase(strmid(header,0,13))
  g = where(STRUPCASE(def) EQ 'CONTENT-TYPE:', Ng)
  if Ng GT 0 then begin
       ClassAndType = strmid(Header[g[0]], 14, strlen(Header[g[0]])-1)
       Class = (strsplit(ClassAndType, '/', /extract))[0]
       Type = (strsplit(ClassAndType, '/', /extract))[1]
  ENDIF 
  def = strupcase(strmid(header,0,15))
  g = where(def EQ 'CONTENT-LENGTH:', Ng)
  if Ng GT 0 then $
         Length = long(strmid(Header[g[0]], 15, strlen(Header[g[0]])-1))
  def = strupcase(strmid(header,0,7))
  g = where(def EQ 'SERIAL:', Ng)
  if Ng GT 0 then $
         Serial = long(strmid(Header[g[0]], 7, strlen(Header[g[0]])-1))
  return
END 
;; end of import

FUNCTION ldsrpc::proxy, proxy
    MESSAGE, 'Called', /INFORMATIONAL
    self.methodCall.proxy = proxy
    RETURN, self
END

FUNCTION ldsrpc::call, methodName, params, FILENAME=filename, DIRNAME=dirname, KEEP=keep
    MESSAGE, 'Called', /INFORMATIONAL
    self.methodCall.methodName = STRTRIM(STRING(methodName),2)
    self.methodCall.params     = PTR_NEW(params)
    PTR_FREE, self.params
    self.params = PTR_NEW()
    PTR_FREE, self.fault
    self.fault = PTR_NEW()

    url = self.methodCall.proxy
    IF self.methodCall.proxy EQ '' THEN $
      STOP, 'LDSRPC::CALL: Proxy does not set yet.'

    ; 
    methodCall = self->createMethodCall()

    ;; output methodCall to tmpfile (just for debug)
    ; make temporary file name with PSUEDO_TMPNAM() in util.pro
    ;tmpFile = GETENV('IDL_TMPDIR') + self->PSUEDO_TMPNAM() + '.methodCall'
    ;IF FILE_TEST(tmpFile) THEN FILE_DELETE, tmpFile
    ;OPENW, wunit, tmpFile, /GET_LUN, ERROR=error
    ;IF error NE 0 THEN MESSAGE, !ERROR_STATE.MSG
    ;FOR i=0,N_ELEMENTS(methodCall.header)-1 DO $
    ;  PRINTF, wunit, methodCall.header[i]
    ;PRINTF, wunit, methodCall.body
    ;FREE_LUN, wunit
    ;IF KEYWORD_SET(KEEP) NE 1 THEN FILE_DELETE, tmpfile

    ;; imported from webget.pro
    slash1 = STRPOS(STRMID(url, 7, STRLEN(url)), '/')
    server = STRMID(url, 7, slash1 )
    purl = STRMID(url, slash1+7, STRLEN(url))
    port = 80

    ; Open the HTTP socket and POST message (HTTP header and body)
    SOCKET, unit, server, port, /GET_LUN, /SWAP_IF_LITTLE_ENDIAN
    MESSAGE, 'sending request...', /CONTINUE
    PRINTF, unit, 'POST ' + purl +  ' HTTP/1.0'  
    FOR i=0,N_ELEMENTS(methodCall.header)-1 DO $
      PRINTF, unit, methodCall.header[i]
    PRINTF, unit, ''
    PRINTF, unit, methodCall.body

    ;; now read the header
    Header = strarr(256)
    LinesRead = 0
    text = 'xxx'
    MESSAGE, 'receiving results...', /CONTINUE
    On_IOERROR, done
    WHILE text NE '' do begin
        READF, unit, text
        Header[LinesRead] = text
        LinesRead = LinesRead+1
        IF LinesRead MOD 256 EQ 0 THEN $
          Header=[Header, StrArr(256)]
    ENDWHILE 
    DONE: On_IOERROR, NULL

    ;; analyze the header
    Header = Header[0:LinesRead-1]
    self->MimeType, Header, Class, Type, Length, Serial

    ; make temporary file name with PSUEDO_TMPNAM() in util.pro
    tmpFile_xml = GETENV('IDL_TMPDIR') + self->PSUEDO_TMPNAM() + '_' + STRTRIM(STRING(Serial),2) + '.xml'
    ;print, 'tmpFile_xml = ' + tmpFile_xml
    IF FILE_TEST(tmpFile_xml) THEN FILE_DELETE, tmpFile_xml

    ; Here is an Ore-Ore way!!
    
    ; reading length of XML part
    READF, unit, text
    ; is digit?
    IF STRMATCH(text,'*[!0123456789]*') NE 0 THEN BEGIN
        WHILE NOT EOF(unit) DO BEGIN
            READF, unit, text
        ENDWHILE
        MESSAGE, 'The server returns error!!', /CONTINUE
        RETURN, self.params     ; NULLPO
    ENDIF
    length_length_xml = LONG(STRLEN(text))
    length_xml = LONG(text)
    MESSAGE, 'length of xml part = ' + STRING(length_xml), /CONTINUE

    ; r/w XML part
    OPENW, wunit, tmpFile_xml, /GET_LUN, ERROR=error
    IF error NE 0 THEN MESSAGE, !ERROR_STATE.MSG
    buf = BYTARR(length_xml, /NOZERO)
    READU, unit, buf
    WRITEU, wunit, buf
    FREE_LUN, wunit
    
    ;SPAWN, "cat " + tmpFile_xml + " > tmpxml.txt"

    ; Parse XML-RPC response File
    self->ParseFile, tmpFile_xml
    IF KEYWORD_SET(KEEP) NE 1 THEN FILE_DELETE, tmpFile_xml

    IF self.isFault EQ 1 THEN BEGIN
        result = *self.fault
        MESSAGE, 'LDSRPC ERROR: ', /CONTINUE
        MESSAGE, 'faultCode:   ' + STRING(result.faultCode), /CONTINUE
        MESSAGE, 'faultString: ' + result.faultString, /CONTINUE
        PTR_FREE, self.params
        self.params = PTR_NEW()
        RETURN, self.params     ; NULLPO
    ENDIF
    
    IF NOT PTR_VALID(self.params) THEN BEGIN
        MESSAGE, 'The server returns no results.', /CONTINUE
        FREE_LUN, unit
        RETURN, self.params     ; returns POINTER!
    ENDIF
    IF N_TAGS(*self.params) LE 0 THEN BEGIN
        MESSAGE, 'The server returns no results.', /CONTINUE
        FREE_LUN, unit
        RETURN, self.params     ; returns POINTER!
    ENDIF

    ; get the size of binary part
    IF Length EQ length_length_xml + 1 + length_xml THEN BEGIN
        filesize = 0
    ENDIF ELSE BEGIN
        filesize = (*self.params).param_0.(0).filesize

        MESSAGE, 'length of binary part = ' + STRING(filesize), /CONTINUE
        IF NOT KEYWORD_SET(filename) THEN filename = (*self.params).param_0.(0).filename
        IF KEYWORD_SET(dirname) THEN BEGIN
            IF STRMATCH(STRTRIM(dirname,2),'*/') THEN BEGIN
                outfile = STRTRIM(dirname,2) + STRTRIM(filename,2)
            ENDIF ELSE BEGIN
                IF STRMATCH(STRTRIM(dirname,2),'*\') THEN BEGIN
                    outfile = STRTRIM(dirname,2) + STRTRIM(filename,2)
                ENDIF ELSE BEGIN
                    outfile = STRTRIM(dirname,2) + '/' + STRTRIM(filename,2)
                ENDELSE
            ENDELSE
        ENDIF ELSE BEGIN
            outfile = STRTRIM(filename,2)
        ENDELSE
        ; r/w binary part
        MESSAGE, 'writing... ' + outfile , /CONTINUE
        OPENW, wunit, outfile, /GET_LUN, ERROR=error
        IF error NE 0 THEN MESSAGE, !ERROR_STATE.MSG
        l = filesize
        WHILE l GT 0 DO BEGIN
            IF l GE 1048576 THEN BEGIN
                len_buf = 1048576
            ENDIF ELSE BEGIN
                len_buf = l
            ENDELSE
            buf = BYTARR(len_buf, /NOZERO)
            READU, unit, buf
            WRITEU, wunit, buf
            l = l - len_buf
            print, 100.0*l/filesize, FORMAT='($,%"%f%% left   \r")'
            flush, -1
        ENDWHILE
        FREE_LUN, wunit
        print, 'Done.                        '
    ENDELSE

    ; Close the HTTP socket
    FREE_LUN, unit

    RETURN, self.params         ; returns POINTER!
END

FUNCTION ldsrpc::params
    RETURN, self.params
END

FUNCTION ldsrpc::isFault
    RETURN, self.isFault
END

; Hook function just for debug
FUNCTION ldsrpc::PTR_NEW, InitExpr, ALLOCATE_HEAP=ah, NO_COPY=nc
    COMPILE_OPT hidden
    MESSAGE, 'Called', /INFORMATIONAL
    ptr = PTR_NEW(InitExpr, ALLOCATE_HEAP=ah, NO_COPY=nc)
    IF NOT !QUIET THEN HELP, ptr, InitExpr, /STRUCTURE
    RETURN, ptr
END

; Making an IDL Variable Undefined
; http://www.dfanning.com/tips/variable_undefine.html
; http://www.dfanning.com/programs/undefine.pro
PRO ldsrpc::undefine, varname
    IF (N_Elements(varname) NE 0) THEN tempvar = Temporary(varname)
END


;----------------------------------------------------------------------------
; Methods to define the class structure
;----------------------------------------------------------------------------

FUNCTION ldsrpc::INIT, DEFAULT_PROXY=default_proxy
    COMPILE_OPT hidden
    MESSAGE, 'Called', /INFORMATIONAL
    self.isFault = 0
    IF KEYWORD_SET(default_proxy) THEN BEGIN
        self.methodCall.proxy = default_proxy
    ENDIF
    RETURN, self->IDLffXMLSAX::Init()
END

; Free all heap variables recursively.
; See http://www.dfanning.com/fileio_tips/memleak.html for more detail.
PRO ldsrpc::CLEANUP
    COMPILE_OPT hidden
    MESSAGE, 'Called', /INFORMATIONAL
    PTR_FREE, self.params
    PTR_FREE, self.fault
    HEAP_FREE, self, /VERBOSE
    HEAP_GC ;, /VERBOSE
    self->IDLffXMLSAX::CLEANUP
END

PRO ldsrpc__define
    !QUIET = 1 ; Set 0 when you need verbose messages for debug.

    MESSAGE, 'Called', /INFORMATIONAL
    ; Not intended to be called directly by regular users
    COMPILE_OPT idl2, hidden
    ON_ERROR, 2                 ; Return to caller if an error occurs

    ; Number of maximum members / parameters
    nMAX = 256
    ; Number of maximum data
    ndMAX = 65536

    STRUCT_HIDE, { methodCall, proxy: '', methodName: '', params: PTR_NEW() }

;    STRUCT_HIDE, { BUF, $
;                   char: '', name: '', value: PTR_NEW(), type: '', $
;                   struct: PTR_NEW(), array: PTR_NEW() }
    STRUCT_HIDE, { NEST, struct: 0B, array: 0B }
    STRUCT_HIDE, { NUMBER, param: 0B, member: 0B, array: 0L }
    ; <!ELEMENT member           (name, value)>
    STRUCT_HIDE, { MEMBER, name: '', value: PTR_NEW(), type: '' }

    ; main part
    STRUCT_HIDE, { LDSRPC, $ ; methodResponse
                   INHERITS IDLffXMLSAX, $
                   methodCall: {methodCall}, $
;                   proxy:  '', $
;                   method: '', $
;                   args:   MAKE_ARRAY(nMAX, /STRING), $
                   ; buffer variables
                   buffer: '', $
                   name:   '', $
                   value:  PTR_NEW(), $
                   type:   '', $
                   struct: PTR_NEW(), $
                   array:  PTR_NEW(), $
;                   buf: {BUF}, $
                   ; status
                   nest: {NEST}, $
                   number: {NUMBER}, $
                   ; temporary fifo area
                   dataArray:   MAKE_ARRAY(ndMAX, VALUE={MEMBER}), $
                   memberArray: MAKE_ARRAY(nMAX, VALUE={MEMBER}), $
                   paramArray:  MAKE_ARRAY(nMAX, /PTR), $
                   ; return values
                   params: PTR_NEW(), $
                   fault:  PTR_NEW(), $ ; {faultCode: 0L, faultString: ''}
                   isFault: 0 }    

;    IF NOT !QUIET THEN $
;      HELP,{NEST},{NUMBER},{MEMBER},{LDSRPC},/STRUCTURE

END

;----------------------------------------------------------------------------
; End of file
;----------------------------------------------------------------------------
