PRO y_Eval_Ramp_Curve,        $
   name,                      $ ; key name for output files
   tsdfile,                   $ ; TSD filename
   range,                     $ ; data range to be extracted and evaluated
   RAMP=ramp,                 $ ; use this ramp data instead of TSD
   REF=ref,                   $ ; # of reference ramp (default: averaged ramp)
   SWIDTH=swidth,             $ ; smooth width (default=10 points)
   FITRANGE=fitrange,         $ ; X range for shift & scale processing
   YFITRANGE=yfitrange,       $ ; Y range for shift & scale processing (relative)
   NSFITRANGENS=nsfitrange,   $ ; range for shift only evaluation
   XPLOTRANGE=xplotrange,     $ ; plot range in X-axis
   GLITCHLIMIT=glitchlimit,   $ ; LIMIT option for y_flag_glitched_ramps()
   NOOUTOFRANGE=nooutofrange, $ ; no data range check
   ALLDATA=alldata,           $ ; plot all data in y_plot_cmp_ramps.pro
   NONSFIT=nonsfit,           $ ; skip shift only fit (for compatibility)
   DONSFIT=donsfit,           $ ; Do shift only fit (default: skip)
   NOSTAT=nostat,             $ ; No statistcs file (*_sig.txt)
   PRESCALE=prescale,         $ ; scale with ramps of smallest inclination
   PREMASK=premask,           $ ; call y_premask_ramp.pro
   EXTRA_X=extra_x,           $ ; give extrapolation length at y_poly_smooth_ramp.pro
   FEC5=fec5,                 $ ; FEC#5 specific correction for LW 7th row
   YRANGE=yrange,             $ ; plot range in Y-axis
   _EXTRA=extra                 ; anything else

; 04/12/21 I.Yamamura
; Evaluate ramp curve by overlay plots
;
; 04/12/31 do at least shift, nsfitrange inplemented
; 05/01/01 renamed with 'y_'
; 05/01/10 Can stard directly from ramp data
; 05/01/20 glitchlimit added
; 05/01/21 give message at each step
; 05/01/28 /NOOUTOFRANGE, /ALLDATA, /NONSFIT, /REF implemented
; 05/02/04 _EXTRA passed to y_plot_cmp_ramps.pro
; 05/02/06 create sigfile, log of data scatters
; 05/05/30 add XPLOTRANGE to specify the plot range (passed to y_plot_cmp_ramps.pro)
; 06/02/16 add YRANGE option
; 06/03/20 By default (without REF option), averaged ramp is calculated
;          and used for evaluation.
; 06/03/21 By default average ramp is used for the reference.
;          Iterative sigma cliping of bad ramps combined with average
;          is applied.
;          Simple shift remains but may not be useful.
; 06/08/17 set datatype when ramp is given directly
; 06/08/20 gb_set_flag_rstanom() included in the preparation process
; 06/10/11 significantly reconfigured. Average strategy changed.
; 06/10/17 bug fix: not properly working when REF was given
; 06/12/29 WIDTH implemented as and option
; 07/01/22 /PRESCALE option scales ramps to the smallest inclination
; 07/02/06 extra_x implemented
; 07/05/29 DONSFIT implemented (default: skip nsfit)
; 07/06/03 call Clear_All_Flag before GB processing
; 07/06/09 residual output referes to averaged ramp by default
;          /PREMASK implemented
; 07/06/19 loop process revisited
; 07/09/30 yfitrange implemented as an input parameter [0..1]
; 07/10/01 enable extra_x when ref is specified.
; 07/10/01 LIMIT for y_flag_abssaturation() = 0.7 <- 0.8
 
;
; Notes
; /FEC5: FEC5 data needs special care for 7th row of the LW detectors

; set color table emvironment
   DEVICE, get_decomposed=decompose_save
   if (decompose_save NE 0) then DEVICE, decompose = 0

; default glitch detection limit in sigma
   if (n_elements(glitchlimit) EQ 0) then glitchlimit = 10D

; define constant
   BAD = y_bad_val_def()

; smoothing width
   if (N_ELEMENTS(SWIDTH) GT 0) then WIDTH = swidth else WIDTH = 10

; X-axis extension 
   if (N_ELEMENTS(extra_x) EQ 0) then extra_x = [0, 0]

; set output file names
   savefile = name+'.sav'
   psfile = name+'.ps'
   avgpsfile = name+'_avg.ps'
   nosclpsfile = name+'_noscl.ps'
   rawpsfile = name+'_raw.ps'
   sigfile = name+'_sig.txt'

; Input ramp data check
   if (n_elements(RAMP) GT 0) then begin
; if RAMP data is directly given

      r_dim = size(ramp, /DIMENSIONS)
      if (n_elements(r_dim) NE 3) then begin
         lmessage, 1, "Incorrect ramp dimension (expect [nsmpl, nramp, ndet])."+STRING(r_dim,/PRINT)
         return
      endif
      v = ramp
      nsmpl = LONG(r_dim[0])
      nramp = LONG(r_dim[1])
      ndet  = LONG(r_dim[2])

      if (ndet EQ 100) then datatype = 'FIS_SW' else datatype = 'FIS_LW'

   endif else begin
; start from TSD file

; read input TSD file
      lmessage, 5, 'Reading input file ('+tsdfile+').'
      tsd=obj_new('tsd')
      ret=tsd->Read_TSD(tsdfile)
      if (ret NE !SUCCEED) then begin
         lmessage, 1, "TSD File not found."+tsdfile
         return
      endif

; preprocessing
      lmessage, 5, 'Processing TSD.'
      clear_all_flag, tsd
      ret = gb_flag_bad_data(tsd)
      ret = gb_set_flag_rstanom(tsd)
      ret = gb_conv_to_volt(tsd, FERR=0)
;      ret = gb_differentiation(tsd)
;      ret = gb_set_glitch_status_gp(tsd)
;      ret = gb_conv_to_volt(tsd, FERR=0)
      if (Keyword_set(FEC5)) then ret = y_lw_col7_fec5(tsd)   ; ignored for SW data

;   ret=tsd->write_tsd("x.fits")
;   stop

; extract ramps
      v= y_rconv(tsd, bp,  range=long(range), /MYRESET, /FLAGGED, /GLITCH)
      r_dim = size(v, /DIMENSIONS)
      nsmpl = r_dim[0]
      nramp = r_dim[1]
      ndet  = r_dim[2]

; check DataType
      datatype = tsd->Get_DataType()

; flagging bad data
;      v[0, *, *] = BAD ; reset anomaly

; clear up
      Obj_Destroy, tsd

   endelse

; set options
   if (NOT Keyword_Set(ALLDATA)) then alldata = 0
   if (n_elements(REF) NE 1) then ref = -1

   if (n_elements(XPLOTRANGE) LT 2) then xplotrange=[0,nsmpl]

; premasking
   if (KEYWORD_SET(PREMASK)) then begin
      if (FILE_TEST('premask_ramp.inc')) then begin
         lmessage, 5, 'Premask data with ./premask_ramp.inc', /CONTINUE
         RESOLVE_ROUTINE, 'y_premask_ramp'
         y_premask_ramp, v
      endif
   endif

; data qualification
   lmessage, 5, 'Flagging Ramp data.'
   if (NOT Keyword_Set(NOOUTOFRANGE)) then v=y_flag_outofrange(v)
   v=y_flag_abssaturation(v, datatype, LIMIT=0.7)
   v=y_flag_rstanom(v, LIMIT=glitchlimit)
; 07/06/03 implemented
   v=y_flag_glitched_ramps_int(v, LIMIT=3D)
   v=y_flag_glitched_ramps(v, LIMIT=glitchlimit)

; define x-axis data
   x = y_xidx(v)
;   lmessage, 5, 'Plotting raw scan data to '+rawpsfile
;   pson, rawpsfile
;   y_plot_cmp_ramps, x, v, psym=3, MEMO=rawpsfile, XRANGE=xplotrange, _EXTRA=extra
;   psoff

; Shift fit
   if (Keyword_Set(DONSFIT)) then begin
      lmessage, 5, 'Shift only fit.'
      if (n_elements(nsfitrange) NE 2) then nsfitrange = [1, fix(nsmpl*0.05)+2]
      ret=y_xsftscl(x, v, dx, fx, sig=s, REFRAMP=ref, range=nsfitrange, /NOSCALE, /PLOT)
;      a=y_average_ramps(x+dx, v, SIG=s, TSIG=t)
      a=y_poly_smooth_ramp(x+dx, v, WIDTH, SIG=s, TSIG=t)

      lmessage, 5, 'Plotting deviation data to '+nosclpsfile
      pson, nosclpsfile
      y_plot_cmp_ramps, x+dx, v, psym=3, REF=ref, yra=[-5e-4, +5e-4], ALLDATA=alldata, MEMO=nosclpsfile, XRANGE=xplotrange, _EXTRA=extra
      psoff
   endif

; Shift & Scale fit
   lmessage, 5, 'Shift & Scale fit.'
   if (n_elements(fitrange) NE 2) then fitrange = [1, fix(nsmpl*0.95)]

   xo = 0D & xr = 0D & yd = 0D & s = 0D & t = 0D

   if (ref GE 0) then begin

; Evaluation w.r.t. given reference ramp
      !P.MULTI=[0, 5, 5]
;print, ref
;      ret=y_xsftscl(x, v, dx, fx, sig=s, XOUT=xo, REFRAMP=ref, range=fitrange, /PLOT)
      ret=y_xsftscl(x, v, dx, fx, sig=s, XOUT=xo, REFRAMP=ref, YFITRANGE=yfitrange, /PLOT)
      !P.MULTI=[0, 1, 1]
;      a=y_average_ramps(xo, v, XREF=xr, SIG=s, TSIG=t)
      a=y_poly_smooth_ramp(xo, v, WIDTH, EXTRA_X=extra_x, XREF=xr, YRESD=yd, SIGMA=s, TSIGMA=t)

      save, filename=savefile, v, x, dx, fx, a, xr, s, t
   endif else begin

; evaluation with averaged ramp
; 06/03/20 I.Yamamura

      nflagged = -1
      xo = x
      xr = REPLICATE(BAD, nsmpl, 1, ndet)
      ye = REPLICATE(BAD, nsmpl, nramp, ndet)

; 07/01/22 pre-scale with smallest inclination ramps
      if (KEYWORD_SET(PRESCALE)) then begin 
;         ret=y_xsftscl(x, v, dx, fx, XOUT=xo, SIG=s, range=fitrange, /PLOT)
         ret=y_xsftscl(x, v, dx, fx, XOUT=xo, SIG=s, YFITRANGE=yfitrange, /HIGHREF, /PLOT)
;         a = y_average_ramps(xo, v, YRESD=yd, XREF=xr, SIG=s, TSIG=t)
         a=y_poly_smooth_ramp(xo, v, WIDTH, XREF=xr, YRESD=yd, SIGMA=s, TSIGMA=t)

      endif else begin

;         a = y_average_ramps(x, v, YRESD=yd, XREF=xr, SIG=s, TSIG=t)
         a=y_poly_smooth_ramp(x, v, WIDTH, XREF=xr, YRESD=yd, SIGMA=s, TSIGMA=t)

      endelse

      iloop = 0
      WHILE (nflagged NE 0) do begin

; plot residual (before flagging)
         y_plot_cmp_ramps, xo, yd, /ALLDATA, DIFF=0, psym=3, XRANGE=xplotrange, CHARSIZE=1.5, _EXTRA=extra

; calculate thresould
; local in a ramp
         ss = replicate(1D, LONG(nramp)) # reform(s, LONG(nsmpl) * ndet)
         ss = reform(ss, nramp, nsmpl, ndet, /OVERWRITE)
         ss = transpose(ss, [1, 0, 2])
         idx = WHERE (ss NE BAD, cnt)
         if (cnt GT 0) then ss[idx] *= 3D    ; 3-sigma clipping

; average over ramps
         st = replicate(1D, LONG(nsmpl)*LONG(nramp)) # t
         st = reform(ss, nsmpl, nramp, ndet, /OVERWRITE)
         idx = WHERE (st NE BAD, cnt)
         if (cnt GT 0) then st[idx] *= 5D    ; 5-sigma clipping

         nsn = replicate(1D, nsmpl, nramp, ndet)

; flag array to record 'outlier' data
         rbad = replicate(0B, nsmpl, nramp, ndet)

; record already bad data
         ibx = where ((yd EQ BAD) OR (ss EQ BAD) OR (ss EQ 0D), bnt)
; set bad data as 0D for convenience
         if (bnt GT 0) then begin
            yd[ibx] = 0D
            ss[ibx] = 100D    ; large enough
            nsn[ibx] = 0D
         endif

; pick bad data
         idx = where (yd^2 GT ss^2 or yd^2 GT st^2, cnt)
         if (cnt GT 0) then rbad[idx] = 1B

; search for ramps with many outliers (= probably glitched ramps)
; count up bad flags
         trbad = Total(rbad, 1, /INTEGER)

; if more than bad flags per ramp
         idx = where(trbad GT nsmpl/20+1, cnt)
         if (cnt GT 0) then begin
            rbad = REFORM(rbad, nsmpl, nramp*ndet, /OVERWRITE)
            rbad[*, idx] = BAD
            rbad = reform(rbad, nsmpl, nramp, ndet, /OVERWRITE)
         endif
         nflagged = cnt

; reject if total deviation is larger than the thresould
         tnsn = Total(nsn, 1)
         idx = where (tnsn EQ 0D, cnt)
         if (cnt GT 0) then tnsn[idx] = 1D

         tsn = Total((yd / ss)^2, 1) / tnsn
         ttn = Total((yd / st)^2, 1) / tnsn
         idx = where (tsn GT 1.0 OR ttn GT 1.0, cnt)
         if (cnt GT 0) then begin
            rbad = reform(rbad, nsmpl, nramp*ndet, /OVERWRITE)
            rbad[*, idx] = 1B
            rbad = reform(rbad, nsmpl, nramp, ndet, /OVERWRITE)
         endif
         nflagged += cnt

; recover BAD flag
         if (bnt GT 0) then rbad[ibx] = 1B

; print results
         print, 'loop=', iloop, ',  ', nflagged, '  ramps flagged out.'

; write thre results back to data
         idx = WHERE ( rbad NE 0B, cnt)
         if (cnt GT 0) then begin
            v[idx] = BAD
            yd[idx] = BAD
         endif

; plot residual (after flagging)
;         y_plot_cmp_ramps, xo, yd, /ALLDATA, DIFF=0, psym=3, XRANGE=xplotrange, CHARSIZE=1.5, _EXTRA=extra

; Shift and Scale with the array
         !P.MULTI=[0, 5, 5]
;         ret=y_xsftscl(x, v, dx, fx, XOUT=xo, SIG=s, EXTREFX=xr, EXTREFY=a, range=fitrange, /PLOT)
         ret=y_xsftscl(x, v, dx, fx, XOUT=xo, SIG=s, EXTREFX=xr, EXTREFY=a, YFITRANGE=yfitrange, /PLOT)
         !P.MULTI=[0, 1, 1]

; calcluate average ramp again
;         a = y_average_ramps(xo, v, YRESD=yd, XREF=xr, SIG=s, TSIG=t)
         a=y_poly_smooth_ramp(xo, v, WIDTH, XREF=xr, YRESD=yd, SIGMA=s, TSIGMA=t)

         iloop = iloop + 1
      END

; flag remaining small glitches
;      if (cnt GT 0) then v[idx] = BAD

; make averaged ramp with the final data set
;      a = y_average_ramps(xo, v, YRESD=yd, XREF=xr, SIG=s, TSIG=t)
      a=y_poly_smooth_ramp(xo, v, WIDTH, EXTRA_X=extra_x, XREF=xr, ERROR=ye, YRESD=yd, SIGMA=s, TSIGMA=t)

; plot residual (final data)
      y_plot_cmp_ramps, xo, yd, /ALLDATA, DIFF=0, psym=3, XRANGE=xplotrange, _EXTRA=extra
      save, filename=savefile, v, x, dx, fx, xo, yd, ye, xr, a, s, t
   endelse

; Set y-axis range
   if (n_elements(YRANGE) NE 2) then begin
      idx = where ((v NE BAD) AND (x NE BAD), cnt)
      if (cnt EQ 0) then begin
         lmessage, 1, 'No data to plot...'
         return
      endif
      ymin = min(v[idx], max=ymax)
      yrange=[ymin,ymax]
   endif

; plot raw ramps (but bad ramps flagged)
   lmessage, 5, 'Plotting raw scan data to '+rawpsfile
   pson, rawpsfile
      y_plot_cmp_ramps, x, v, /ALLDATA, DIFF=0, psym=3, YRMS=yrmsr, $
         MEMO=rawpsfile, XRANGE=xplotrange, YRANGE=yrange, _EXTRA=extra
   psoff

; plot the result
   lmessage, 5, 'Plotting deviation data to '+psfile
   pson, psfile
   if (alldata NE 0) then begin
      if (ref GE 0) then begin
         y_plot_cmp_ramps, xo, v, psym=3, REF=ref, $
            /DIFF, /ALLDATA, MEMO=psfile, $
            XRANGE=xplotrange, YRANGE=[-5e-4, +5e-4], $
            _EXTRA=extra, TAVG=tavg, TSIG=tsig
      endif else begin
         y_plot_cmp_ramps, xo, yd, psym=3, $
            DIFF=0, /ALLDATA, MEMO=psfile, $
            XRANGE=xplotrange, YRANGE=[-5e-4, +5e-4], $
            _EXTRA=extra, TAVG=tavg, TSIG=tsig
      endelse
   endif else begin
;      y_plot_cmp_ramps, xo, v, psym=3, AREF_X=xr, AREF_Y=a,$
      y_plot_cmp_ramps, xo, yd, psym=3, $
         DIFF=0, ALLDATA=0, MEMO=psfile, $
         XRANGE=xplotrange, YRANGE=[-5e-4, +5e-4], $
         _EXTRA=extra, TAVG=tavg, TSIG=tsig
   endelse
   psoff
   lmessage, 5, 'Plotting average ramp to '+avgpsfile
   pson, avgpsfile
      y_plot_all_ramps, xr, a, PSYM=-3, MEMO=psfile, $
         _EXTRA=extra
   psoff

; write data scatter records
; t: Stdev from averaged ramp averaged over 20--50 % part of ramp
; tavg: Avarage deviation from the refrerence in y_plot_cmp_ramps.pro
; tsig: Stdev of deviation from the refrerence in y_plot_cmp_ramps.pro
; yrms: Stdev of the reference ramp from its smoothed one.

   yrms = yrmsr      ; number have to come from raw plot
   if (NOT Keyword_Set(NOSTAT)) then begin
      OpenW, lun, /Get_LUN, sigfile
      printf, lun, 'Det#    RMS[uV]  AVGr[uV]  RMSr[uV] YRMS[uV]'
      i = where(t    EQ BAD, c) & t    *= 1D6 & if (c GT 0) then t[i]    = BAD
      i = where(tavg EQ BAD, c) & tavg *= 1D6 & if (c GT 0) then tavg[i] = BAD
      i = where(tsig EQ BAD, c) & tsig *= 1D6 & if (c GT 0) then tsig[i] = BAD
      i = where(yrms EQ BAD, c) & yrms *= 1D6 & if (c GT 0) then yrms[i] = BAD
      for i = 0, ndet-1 do begin
         printf, lun, FORMAT='(I5, F10.2, F10.2, F10.2, F10.2)', i+1, t[i], tavg[i], tsig[i], yrms[i]
      endfor
      Close, lun
   endif

   if (decompose_save NE 0) then DEVICE, decompose = decompose_save

   lmessage, 5, 'Finish.'
return
end
