FUNCTION y_Make_Corr_Tbl, $
   ramp,             $ ; array of ramp curve. ramp[samplings, ramp_id, det_id]
   RANGE=range,      $ ; data range in a ramp used for fit
   BOXCAR=boxcar,    $ ; boxcar smoothing size
   PLOTFIT=plotfit,  $ ; show fitting results
   WAITSEC=waitsec,  $ ; wait between plots. default = 0.0 sec
   CCF=ccf,          $ ; write results in CCF file.
   EXTYRNGFILE=extyrngfile,$ ; externally given Y-range file (.sav)
   EXTPCEFFILE=extpceffile,$ ; externally given polynomial coefficients file (.sav)
   PCEF=pcef           ; (returned) polynomial coefficients
; (returned)           ; correction factor table
;
; 02/10/31 I.Yamamura
; 03/03/25 treat numerical exception properly
;          boxcar implemented
; 05/01/11 renamed with 'y_'
;          Polynomial fit included in the process
;          Change table format. Add o (offset) field to avoid zero-cross.
;          /PLOTFIT implemented
; 06/10/15 Print detector number on plot
; 06/11/15 CCF option implemented.
; 07/01/05 Plotting result table is exported as y_plot_corr_tbl.pro
; 07/01/12 add WAITSEC option
; 07/10/01 accept external pcef
; 07/10/11 Y-range of data used for fitting recorded in corr_tbl_yrng.sav
; 07/10/11 YRNGFILE => EXTYRNGFILE, PCEFFILE => EXTPCEFFILE (avoid name conflict)
; 07/10/29 bug at saving yrange -> fixed.

BAD = y_bad_val_def()

OFFSET = 0.01D          ; to avoid dividing by 0

; always line fit
   degree = 1

; set waitsec
   if (N_ELEMENTS(waitsec) EQ 0) then waitsec = 0.0

; prepare for plot
   if (keyword_set(PLOTFIT)) then loadct, 39

; input data dimension check
   r_dim = size(ramp, /DIMENSIONS)
   if (n_elements(r_dim) NE 3 OR r_dim[1] NE 1) then begin
      MESSAGE, "Incorrect dimension of ramp (expect [nsmpl, 1, ndet])."+STRING(r_dim,/PRINT), /CONTINUE
      return, 0
   endif
   nsmpl = r_dim[0]
   nramp = r_dim[1]
   ndet  = r_dim[2]

; check dimension of external pcef
   extpcef = 0
   if (N_ELEMENTS(extpceffile) GT 0) then begin
      restore, extpceffile
      npcef = SIZE(pcef, /DIMENSION)
      if (npcef[0] NE 2 OR npcef[1] NE ndet) then begin
         msg = 'PCEF file contains data of inconsistent dimension: '
         msg = msg + STRING(FORMAT='(3I5)', ndet, npcef)
         MESSAGE, msg, /CONTINUE
         tbl = -1
         goto, EXIT
      endif
      extpcef = 1
   endif else begin
   ; initialize fit coeffciency data area
      pcef = replicate(BAD, degree+1, ndet)
   endelse

; external Y-range data
   extyrng = 0
   yrange = REPLICATE(0D, 2, r_dim[2])
   if (N_ELEMENTS(extyrngfile) GT 0) then begin
      restore, extyrngfile
      n_dim = SIZE(yrange, /DIMENSION)
      if (n_dim[0] EQ 2 AND n_dim[1] EQ r_dim[2]) then begin
         extyrng = 1
      endif
   endif

; fit range check
   if (n_elements(range) NE 2) then range = [4L, LONG(nsmpl*0.5)]

; prepare correction table
   t = { v: replicate(BAD, r_dim[0]), c: replicate(BAD, r_dim[0]), o: BAD}
   tbl = replicate(t, r_dim[2])

; prepare sampling index
   x  = indgen(nsmpl, /DOUBLE)
   xx = dblarr(nsmpl, degree+1)
   for i = 0, degree do xx[*, i] = x^i

; calculate correction factor for each pixel
   for idet = 0L, r_dim[2]-1 do begin
; linear fit to the ramp
      vv = ramp[*, 0, idet]
      idx = where (vv NE BAD, cnt)
      if (cnt GT 0) then begin
         vv = vv[idx]
         xf = x[idx]

         if (extyrng) then begin
            idx2 = where (vv GE yrange[0, idet] AND vv LE yrange[1, idet], cnt2)
         endif else begin
            idx2 = where (xf GE range[0] AND xf LE range[1], cnt2)
         endelse
         if (cnt2 GE 2) then begin
; we need at least two data points for line fit
            if (extpcef) then begin
; fit only offset
               db = TOTAL(vv[idx2] - (pcef[0, idet] + pcef[1, idet] * xf[idx2])) / DOUBLE(cnt2)
               pcef[0, idet] += db
            endif else begin
; full fitting
               pcef[*, idet] =  POLY_FIT(xf[idx2], vv[idx2], degree, /DOUBLE)
            endelse
; save yrange
            if (NOT extyrng) then yrange[*, idet] = [MIN(vv[idx2]), MAX(vv[idx2])]

; reset detector data and 
            cc = xx[idx, *] # pcef[*, idet]

            if (keyword_set(BOXCAR)) then vv = smooth(vv, boxcar)

            tbl[idet].v[0:(cnt-1)] = vv
            tbl[idet].o            = 0D
            tbl[idet].c[0:(cnt-1)] = cc - vv

; if plot fitting results
            if (Keyword_set(PLOTFIT)) then begin
               print, idet, pcef[*, idet]
               y_Plot_Corr_Tbl, tbl, idet
               wait, waitsec
            endif

         endif
      endif
   endfor


; -------- save fitting results
   if (NOT extpcef) then save, filename='corr_tbl_pcef.sav', pcef

; -------- save Y-range results
   if (NOT extyrng) then save, filename='corr_tbl_yrng.sav', yrange

; -------- output CCF file

   if (KEYWORD_SET(CCF)) then begin
         if (ndet EQ 100) then begin   ; SW
            basename = 'RCTBS'
         endif else if (ndet EQ 75) then begin
            basename = 'RCTBL'
         endif else begin
            msg = STRING('Invalid array size', ndet)
            lmessage, 1, msg
            goto, EXIT
         endelse

      if (SIZE(CCF, /TNAME) EQ 'STRING') then begin
         ccf_file = ccf
      endif else begin
         ccf_file = basename+'_999_999.ccf'
      endelse
      CALDAT, SYSTIME(1, /JULIAN), mo, dy, yr, hr, mn, sc
      date = STRING(FORMAT='(I4,A,I2,A,I2,A,I2,A,I2,A,F2.0)', $
         yr, '-', mo, '-', dy, 'T', hr, ':', mn, ':', sc)

      h = CCF_Create_New_Header()
      h.NAME      = basename
      h.SET_NO    = 999
      h.REV_NO    = 999
      h.STATUS    = 'TEST'
      h.RANGE     = [0.00000, 0.00000]
      h.CREATOR   = 'y_make_corr_tbl'
      h.CDATE     = date
      h.FMTVER    = 001
      h.COMMENT   = 'Ramp curve correction table (temporal)'

      r = CCF_Write_CCF_File(ccf_file, tbl, h, /OVER);
   endif

EXIT:

return, tbl
end
